
(* This generated code requires the following version of MenhirLib: *)

let () =
  MenhirLib.StaticVersion.require_20220210

module MenhirBasics = struct
  
  exception Error
  
  let _eRR =
    fun _s ->
      raise Error
  
  type token = Js_token.token
  
end

include MenhirBasics

# 18 "compiler/lib/js_parser.mly"
  
(*
 * src: ocamlyaccified from Marcel Laverdet 'fbjs2' via emacs macros, itself
 * extracted from the official ECMAscript specification at:
 *  http://www.ecma-international.org/publications/standards/ecma-262.htm
 *
 * see also http://en.wikipedia.org/wiki/ECMAScript_syntax
 *
 * related work:
 *  - http://marijnhaverbeke.nl/parse-js/, js parser in common lisp
 *    (which has been since ported to javascript by nodejs people)
 *  - jslint
 *)

open Javascript

let var pi name = ident ~loc:(Pi pi) name


# 42 "compiler/lib/js_parser.ml"

module Tables = struct
  
  include MenhirBasics
  
  let token2terminal : token -> int =
    fun _tok ->
      match _tok with
      | Js_token.EOF _ ->
          90
      | Js_token.TAnnot _ ->
          89
      | Js_token.TComment _ ->
          88
      | Js_token.TCommentLineDirective _ ->
          87
      | Js_token.TUnknown _ ->
          86
      | Js_token.T_AND _ ->
          85
      | Js_token.T_ASSIGN _ ->
          84
      | Js_token.T_BIT_AND _ ->
          83
      | Js_token.T_BIT_AND_ASSIGN _ ->
          82
      | Js_token.T_BIT_NOT _ ->
          81
      | Js_token.T_BIT_OR _ ->
          80
      | Js_token.T_BIT_OR_ASSIGN _ ->
          79
      | Js_token.T_BIT_XOR _ ->
          78
      | Js_token.T_BIT_XOR_ASSIGN _ ->
          77
      | Js_token.T_BREAK _ ->
          76
      | Js_token.T_CASE _ ->
          75
      | Js_token.T_CATCH _ ->
          74
      | Js_token.T_COLON _ ->
          73
      | Js_token.T_COMMA _ ->
          72
      | Js_token.T_CONTINUE _ ->
          71
      | Js_token.T_DEBUGGER _ ->
          70
      | Js_token.T_DECR _ ->
          69
      | Js_token.T_DECR_NB _ ->
          68
      | Js_token.T_DEFAULT _ ->
          67
      | Js_token.T_DELETE _ ->
          66
      | Js_token.T_DIV _ ->
          65
      | Js_token.T_DIV_ASSIGN _ ->
          64
      | Js_token.T_DO _ ->
          63
      | Js_token.T_ELSE _ ->
          62
      | Js_token.T_EQUAL _ ->
          61
      | Js_token.T_FALSE _ ->
          60
      | Js_token.T_FINALLY _ ->
          59
      | Js_token.T_FOR _ ->
          58
      | Js_token.T_FUNCTION _ ->
          57
      | Js_token.T_GREATER_THAN _ ->
          56
      | Js_token.T_GREATER_THAN_EQUAL _ ->
          55
      | Js_token.T_IDENTIFIER _ ->
          54
      | Js_token.T_IF _ ->
          53
      | Js_token.T_IN _ ->
          52
      | Js_token.T_INCR _ ->
          51
      | Js_token.T_INCR_NB _ ->
          50
      | Js_token.T_INSTANCEOF _ ->
          49
      | Js_token.T_LBRACKET _ ->
          48
      | Js_token.T_LCURLY _ ->
          47
      | Js_token.T_LESS_THAN _ ->
          46
      | Js_token.T_LESS_THAN_EQUAL _ ->
          45
      | Js_token.T_LPAREN _ ->
          44
      | Js_token.T_LSHIFT _ ->
          43
      | Js_token.T_LSHIFT_ASSIGN _ ->
          42
      | Js_token.T_MINUS _ ->
          41
      | Js_token.T_MINUS_ASSIGN _ ->
          40
      | Js_token.T_MOD _ ->
          39
      | Js_token.T_MOD_ASSIGN _ ->
          38
      | Js_token.T_MULT _ ->
          37
      | Js_token.T_MULT_ASSIGN _ ->
          36
      | Js_token.T_NEW _ ->
          35
      | Js_token.T_NOT _ ->
          34
      | Js_token.T_NOT_EQUAL _ ->
          33
      | Js_token.T_NULL _ ->
          32
      | Js_token.T_NUMBER _ ->
          31
      | Js_token.T_OR _ ->
          30
      | Js_token.T_PERIOD _ ->
          29
      | Js_token.T_PLING _ ->
          28
      | Js_token.T_PLUS _ ->
          27
      | Js_token.T_PLUS_ASSIGN _ ->
          26
      | Js_token.T_RBRACKET _ ->
          25
      | Js_token.T_RCURLY _ ->
          24
      | Js_token.T_REGEX _ ->
          23
      | Js_token.T_RETURN _ ->
          22
      | Js_token.T_RPAREN _ ->
          21
      | Js_token.T_RSHIFT _ ->
          20
      | Js_token.T_RSHIFT3 _ ->
          19
      | Js_token.T_RSHIFT3_ASSIGN _ ->
          18
      | Js_token.T_RSHIFT_ASSIGN _ ->
          17
      | Js_token.T_SEMICOLON _ ->
          16
      | Js_token.T_SPREAD _ ->
          15
      | Js_token.T_STRICT_EQUAL _ ->
          14
      | Js_token.T_STRICT_NOT_EQUAL _ ->
          13
      | Js_token.T_STRING _ ->
          12
      | Js_token.T_SWITCH _ ->
          11
      | Js_token.T_THIS _ ->
          10
      | Js_token.T_THROW _ ->
          9
      | Js_token.T_TRUE _ ->
          8
      | Js_token.T_TRY _ ->
          7
      | Js_token.T_TYPEOF _ ->
          6
      | Js_token.T_VAR _ ->
          5
      | Js_token.T_VIRTUAL_SEMICOLON _ ->
          4
      | Js_token.T_VOID _ ->
          3
      | Js_token.T_WHILE _ ->
          2
      | Js_token.T_WITH _ ->
          1
  
  and error_terminal =
    0
  
  and token2value : token -> Obj.t =
    fun _tok ->
      match _tok with
      | Js_token.EOF _v ->
          Obj.repr _v
      | Js_token.TAnnot _v ->
          Obj.repr _v
      | Js_token.TComment _v ->
          Obj.repr _v
      | Js_token.TCommentLineDirective _v ->
          Obj.repr _v
      | Js_token.TUnknown _v ->
          Obj.repr _v
      | Js_token.T_AND _v ->
          Obj.repr _v
      | Js_token.T_ASSIGN _v ->
          Obj.repr _v
      | Js_token.T_BIT_AND _v ->
          Obj.repr _v
      | Js_token.T_BIT_AND_ASSIGN _v ->
          Obj.repr _v
      | Js_token.T_BIT_NOT _v ->
          Obj.repr _v
      | Js_token.T_BIT_OR _v ->
          Obj.repr _v
      | Js_token.T_BIT_OR_ASSIGN _v ->
          Obj.repr _v
      | Js_token.T_BIT_XOR _v ->
          Obj.repr _v
      | Js_token.T_BIT_XOR_ASSIGN _v ->
          Obj.repr _v
      | Js_token.T_BREAK _v ->
          Obj.repr _v
      | Js_token.T_CASE _v ->
          Obj.repr _v
      | Js_token.T_CATCH _v ->
          Obj.repr _v
      | Js_token.T_COLON _v ->
          Obj.repr _v
      | Js_token.T_COMMA _v ->
          Obj.repr _v
      | Js_token.T_CONTINUE _v ->
          Obj.repr _v
      | Js_token.T_DEBUGGER _v ->
          Obj.repr _v
      | Js_token.T_DECR _v ->
          Obj.repr _v
      | Js_token.T_DECR_NB _v ->
          Obj.repr _v
      | Js_token.T_DEFAULT _v ->
          Obj.repr _v
      | Js_token.T_DELETE _v ->
          Obj.repr _v
      | Js_token.T_DIV _v ->
          Obj.repr _v
      | Js_token.T_DIV_ASSIGN _v ->
          Obj.repr _v
      | Js_token.T_DO _v ->
          Obj.repr _v
      | Js_token.T_ELSE _v ->
          Obj.repr _v
      | Js_token.T_EQUAL _v ->
          Obj.repr _v
      | Js_token.T_FALSE _v ->
          Obj.repr _v
      | Js_token.T_FINALLY _v ->
          Obj.repr _v
      | Js_token.T_FOR _v ->
          Obj.repr _v
      | Js_token.T_FUNCTION _v ->
          Obj.repr _v
      | Js_token.T_GREATER_THAN _v ->
          Obj.repr _v
      | Js_token.T_GREATER_THAN_EQUAL _v ->
          Obj.repr _v
      | Js_token.T_IDENTIFIER _v ->
          Obj.repr _v
      | Js_token.T_IF _v ->
          Obj.repr _v
      | Js_token.T_IN _v ->
          Obj.repr _v
      | Js_token.T_INCR _v ->
          Obj.repr _v
      | Js_token.T_INCR_NB _v ->
          Obj.repr _v
      | Js_token.T_INSTANCEOF _v ->
          Obj.repr _v
      | Js_token.T_LBRACKET _v ->
          Obj.repr _v
      | Js_token.T_LCURLY _v ->
          Obj.repr _v
      | Js_token.T_LESS_THAN _v ->
          Obj.repr _v
      | Js_token.T_LESS_THAN_EQUAL _v ->
          Obj.repr _v
      | Js_token.T_LPAREN _v ->
          Obj.repr _v
      | Js_token.T_LSHIFT _v ->
          Obj.repr _v
      | Js_token.T_LSHIFT_ASSIGN _v ->
          Obj.repr _v
      | Js_token.T_MINUS _v ->
          Obj.repr _v
      | Js_token.T_MINUS_ASSIGN _v ->
          Obj.repr _v
      | Js_token.T_MOD _v ->
          Obj.repr _v
      | Js_token.T_MOD_ASSIGN _v ->
          Obj.repr _v
      | Js_token.T_MULT _v ->
          Obj.repr _v
      | Js_token.T_MULT_ASSIGN _v ->
          Obj.repr _v
      | Js_token.T_NEW _v ->
          Obj.repr _v
      | Js_token.T_NOT _v ->
          Obj.repr _v
      | Js_token.T_NOT_EQUAL _v ->
          Obj.repr _v
      | Js_token.T_NULL _v ->
          Obj.repr _v
      | Js_token.T_NUMBER _v ->
          Obj.repr _v
      | Js_token.T_OR _v ->
          Obj.repr _v
      | Js_token.T_PERIOD _v ->
          Obj.repr _v
      | Js_token.T_PLING _v ->
          Obj.repr _v
      | Js_token.T_PLUS _v ->
          Obj.repr _v
      | Js_token.T_PLUS_ASSIGN _v ->
          Obj.repr _v
      | Js_token.T_RBRACKET _v ->
          Obj.repr _v
      | Js_token.T_RCURLY _v ->
          Obj.repr _v
      | Js_token.T_REGEX _v ->
          Obj.repr _v
      | Js_token.T_RETURN _v ->
          Obj.repr _v
      | Js_token.T_RPAREN _v ->
          Obj.repr _v
      | Js_token.T_RSHIFT _v ->
          Obj.repr _v
      | Js_token.T_RSHIFT3 _v ->
          Obj.repr _v
      | Js_token.T_RSHIFT3_ASSIGN _v ->
          Obj.repr _v
      | Js_token.T_RSHIFT_ASSIGN _v ->
          Obj.repr _v
      | Js_token.T_SEMICOLON _v ->
          Obj.repr _v
      | Js_token.T_SPREAD _v ->
          Obj.repr _v
      | Js_token.T_STRICT_EQUAL _v ->
          Obj.repr _v
      | Js_token.T_STRICT_NOT_EQUAL _v ->
          Obj.repr _v
      | Js_token.T_STRING _v ->
          Obj.repr _v
      | Js_token.T_SWITCH _v ->
          Obj.repr _v
      | Js_token.T_THIS _v ->
          Obj.repr _v
      | Js_token.T_THROW _v ->
          Obj.repr _v
      | Js_token.T_TRUE _v ->
          Obj.repr _v
      | Js_token.T_TRY _v ->
          Obj.repr _v
      | Js_token.T_TYPEOF _v ->
          Obj.repr _v
      | Js_token.T_VAR _v ->
          Obj.repr _v
      | Js_token.T_VIRTUAL_SEMICOLON _v ->
          Obj.repr _v
      | Js_token.T_VOID _v ->
          Obj.repr _v
      | Js_token.T_WHILE _v ->
          Obj.repr _v
      | Js_token.T_WITH _v ->
          Obj.repr _v
  
  and default_reduction =
    (16, "\000\000\000\003\000\000\000\129\000\000\000\000\000\000\000\000\000\000\000\030\001\007\001\012\001\020\000\000\000\149\000\148\000\000\000\000\000\000\000\000\000\000\000k\000j\000i\000h\000g\000f\000n\000e\000d\000c\001\018\000b\000l\000a\000V\000U\000`\000R\000_\000^\000T\000m\000W\000]\000\\\000[\000o\000Z\000S\000Y\000X\000\000\0005\000\000\000\000\000\000\000\b\000\000\000\000\001B\000\000\001@\000\168\000\000\000\000\000\000\000\000\001\029\000\135\000\000\000\000\000\000\000\000\000\000\000\031\000\000\000\000\000\000\000\000\001\b\001\r\001\004\000\136\000\000\000\216\000\217\000\219\001\005\001\011\001\t\000v\000\000\000\000\000\138\000\000\000\000\000,\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\218\001\006\000\152\000\151\000\000\000\000\000$\000\000\000\000\000\000\000\000\000\000\000\172\001\n\001\014\000\000\000\170\000\000\000\174\000\000\000\184\000\000\000\173\000\000\000\171\000\000\000\000\000\000\000\000\000\024\000\025\000\021\000\018\000\020\000\022\000\023\000\019\000\027\000\028\000\026\000\017\000\000\000\011\000\012\000\000\000\000\001:\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000#\000\000\000E\000D\000\"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\004\000\133\000\000\000\006\000\005\000\000\000\000\001\023\000\000\000\000\000\137\000!\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000r\000\000\000\000\001\025\000\160\001A\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000C\000\000\001\021\000\000\000\000\000\000\000\000\000\000\000\000\000\147\000\000\000\143\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000.\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\r\000\014\000\000\000\000\001;\000\000\000\000\000\000\000\000\000\000\000\188\000\000\000\186\000\000\000\190\000\000\000\189\000\000\000\187\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000s\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\0006\000\000\000t\0001\000\164\000\000\000 \000\000\000\000\0014\0011\001'\0016\001-\0000\0015\001%\000\000\000:\0009\001&\000\000\000\000\000\000\000\000\0008\001,\000\140\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\240\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001<\000\000\000\000\000\000\000\000\000\000\000\203\000\000\000\201\000\000\000\205\000\000\000\215\000\000\000\204\000\000\000\202\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000x\000\000\000\000\000\142\000\000\000\000\000\141\000%\000\000\000\000\000\016\0017\000\000\000\000\000u\0010\0012\0013\001(\000\000\000\000\000I\0018\001)\001.\001/\001*\000\015\000\000\000\000\000(\000\000\000\000\000'\000&\001+\000H\000M\000\000\000\000\000\000\000\162\001\027\000\000\000\000\000\000\000\000\000\000\000\000\000O\000\000\000\000\000\000\000\000\000\000\000\000\000N\000\000\000\000\000\000\000\000\000L\000\000\000\000\000G\000F\000\000\000\000\000p\000\000\000\131\000)\000\000\000\000\000\000\0007\000\000\0019\000\000\000\166\000\000\000}\000\000\0004\000\029\000\000\000\000\000K\000\000\000\000\000\000\000\000\000*\001?\000\000\001>\000\158\001C\000\000\000\000\000\000\000\000\000\000\000P\001!\000\000\000\127\001\"\000\000\0003\000Q\000\000\000\000\000@\000\000\000<\000\000\000\007\000A\000\000\000\000\000\n\000\000\000>\000\000\000\t\000=\000\150\000\000\000\000\001 \001\019\001\017\000\000\0002\000\000\000\000\001\015\000\145\000\000\000\139\000\000\000\000\000\000\000\000\000\000\000\000\001D\001#\000\000\000{\000\001\000\000\001\016\000\000\000\002\000\000\001$")
  
  and error =
    (91, "w\248\131\017\176I\182i/\b@`\000\000\000\000\000\000\000\000\000\000\001\223\226\012F\193&\217\164\188!\001\128\000\000\000\000\000\000\000\000\000\000\007\127\1361\027\004\155f\146\240\132\000\000\000\000\000\000\016\000\000\000\000\000\004\170\000Dl\018l\146\011\000\016\000\149@\b\141\130M\146A`\002\000\018\168\001\017\176I\178H,\000@\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000%P\002#`\147d\144X\000\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\018\168\001\017\176I\178H,\000@\000\021\000 2\0010I\000\000\000\000J\160\004F\193&\201 \176\001\000\tT\000\136\216$\217$\022\000 \007\127\128(\025\000\004\231\1793\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\016\000\000\149@\b\141\130M\146A`\002\000\018\168\001Q\176I\178H,\128@\000\000\000\000\000\000\000\000\000\000\000\000J\160\004F\193&\201 \176\001\000\tT\000\136\216$\217$\022\000 \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\016\004\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\b\000\000\000\000\000\000\000\000\128\000\000\000@\000\000\000\000\000\000\016\000\000\000\000\000\002\000\000\000\000\000\000\000\000\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\b\000\000\000\000\000w\248\131\145\176I\182i/\b@\000\000\000\000\000\001\000\000\000\000\000\000J\160\004F\193&\201 \176\001\000\000\000\000\000\000\000\000\000\000\000\000\001*\128\017\027\004\155$\130\192\004\000%P\002#`\147d\144X\000\128\004\170\000Dl\018l\146\011\000\016\000\149@\b\141\130M\146A`\002\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\128i\205\164Uf\152D\140)B\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\128o\207\228\255\238\152L\140{\194\239\224\004\001 \000\156\246fp\000\000\000\000\000\000\000\000\000\000\000\000\000\149H(\141\130M\146A`\002\000\018\168\001\017\176I\178H,\000@\000\000\000\000\000\000\000\000\000\000\000\000 \026si\021Y\166\017#\nP\137T\000\136\216$\217$\022\000 \000\128i\205\164Uf\152D\140)B%P\002#`\147d\144X\000\128\002\001\1676\145U\154a\0180\165\b\149@\b\141\130M\146A`\002\000\b\006\156\218EVi\132H\194\148 \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\016\r\249\252\159\253\211\t\145\143x]\252\000\128$\000\019\158\204\206\000\000\000\000\000\000\000\000\000\000\000\000\000\018\168\001\017\176I\178H,\000@\001\000\208\153H\000\2010\128\024R\132J\160\004F\193&\201 \176\001\000\004\003Be \003$\194\000aJ\017*\128\017\027\004\155$\130\192\004\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\018\168\001\017\176I\178H,\000@\000\000\000\000\000\000\000\000\000\000\000\000J\160\004F\193&\201 \176\001\000\000\000\000\000\000\000\000\000\000\000\000\001*\128\017\027\004\155$\130\192\004\000\000\000\000\000\000\000\000\000\000\000\000\004\170\000Dl\018l\146\011\000\016\000\000\000\000\000\000\000\000\000\000\000\000\018\168\001\017\176I\178H,\000@\000\000\000\000\000\000\000\000\000\000\000\000J\160\004F\193&\201 \176\001\000\004\003Be \003$\194\000aJ\017*\128\017\027\004\155$\130\192\004\000\016\r\249\244\159\236\211\t\145\143x@\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000J\160\004F\193&\201 \176\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\128\000\004\170\000Dl\018l\146\011\000\016\000\000\000\000\000\000\000\000\000\000\000\000\018\168\001\017\176I\178H,\000@\001\000\208\153H\000\2010\128\024R\132J\160\004F\193&\201 \176\001\000\004\003Be \003$\194\000aJ\017*\128\017\027\004\155$\130\192\004\000\016\r\t\148\128\012\147\b\001\133(D\170\000Dl\018l\146\011\000\016\000@4&R\0002L \006\020\161\018\168\001\017\176I\178H,\000@\001\000\208\153H\000\2010\128\024R\132J\160\004F\193&\201 \176\001\000\004\003Be \003$\194\000aJ\017*\128\017\027\004\155$\130\192\004\000\016\r\t\148\128\012\147\b\001\133(@\000\000\016\000\000\000\000\000 \000\000\000\000\000\000\000\000\000\000\000\000\000\018\168\001\017\176I\178H,\000@\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001*\128\017\027\004\155$\130\192\004\000\016\r9\180\138\172\211\b\145\133(D\170\000Dl\018l\146\011\000\016\000@4\230\210*\179L\"F\020\161\018\168\001\017\176I\178H,\000@\001\000\211\155H\170\2050\137\024R\132J\160\004F\193&\201 \176\001\000\004\003Nm\"\1714\194$aJ\017*\128\017\027\004\155$\130\192\004\000\016\r9\180\138\172\211\b\145\133(@\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\004\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\002\000\000\000\000\000\000@\000\001*\144\017\027\004\155$\130\192\004\000\000\000\000\000\000\000\000\000\000\000\000\004\170\000Dl\018l\146\011\000\016\000\000\000\002\000\000\000\000\000\004\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000 \026si\021Y\166\017#\nP\132\003Nm\"\1714\194$aJ\016\128i\205\164Uf\152D\140)B\000\000\b\000\000\000\000\000\001\000\000\029\254 \196l\018m\138K\194\016\000\149@\b\141\130M\146A`\002\000\b\006\152\026EVi\132H\130\148\000\000\000\000\000\000\000@\000\000\000\000 \002\000\000\000\000\000\000\002\000 \tT\000\136\216$\217$\022\000 \000\000\000\000\000\000\000\000\000\000\000\000\016\001\000\000\000\000\000\000\001\000\000\000\000\000\000\000\000\000\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000J\160\004F\193&\201 \176\001\000\004\003L\r\"\1714\194$AJ\000\000\000\000\000\000\016\000\000\000\000\000\239\241\007#`\147lR^\016\128\004\170\000Dl\018l\146\011\000\016\000@\004\000\000\000\000\000\000\004\000\000\000\000\000\000\000\b\000\000\000\000\000\002U\000\"6\t6I\005\128\b\000\000\000\016\000\000\000\000\000\002\000\000\000\000\000\000\000\000\128\000\000\000\000\000\000\000\b\000\000\000\000\001\001\000\000%P\002#`\147d\144X\000\128\000\000\000\000\000\000\000\000\0000\000\003\191\196\028\141\130M\177I\248\194\000\000\000\000\000\000\000\000\000\000\000\000\003U\016\"6\t6I\005\128\b\000\000\000\000\000\000\000\000\000\000\000\000\004\000B\000\000\000\000\000\000@\000\001*\128\017\027\004\155$\130\192\004\000\016\r04\138\172\211\b\145\005(\004\170\000Dl\018l\146\011\000\016\000@4\192\210*\179L\"D\020\160\000\168\001\001\144\t\130H\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000 \027\224\249?\251\166\019\"\030\240\000\000\000\000\000\000\000\000\000\000\000\001*\128\017\027\004\155$\130\192\004\000\016\r04\138\172\211\b\145\005(\004\170\000Dl\018l\146\011\000\016\000@4\192\210*\179L\"D\020\160\018\168\001\017\176I\178H,\000@\001\000\211\003H\170\2050\137\016R\128\000\000\000\000\000 \000\000\000\000\000\tT\000\136\216$\217$\022\000 \000\000\000@\000\000\000\000\000\b\000\000\239\241\006#`\147lR^\016\128\002\001\190\015\147\255\186a21\239\000\000\000\000\000\000@\000\000\000\000\000\022\168\129\017\176I\178H,\000@\000\000\000\000\000\000\000@\000\000\000\000\000\002\000\000\000\000 \000\002\000 \tT\000\136\216$\217$\022\000 \000\000\000\000\000\000\000\000\000\000\000\000\000\r04\138\172\211\b\145\133(\000\001\160\002\144\001\146a\0000\165\000\149@\b\141\130M\146A`\002\000\000\006\128\n@\006I\132\000\194\148\002U\000\"6\t6I\005\128\b\000\000\026\000)\000\025&\016\003\nP\tT\000\136\216$\217$\022\000 \000\000o\131\164\255f\152L\140{\192%P\002#`\147d\144X\000\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000@\000\002U\000\"6\t6I\005\128\b\000\000\000\000\000\000\000\000\000\000\000\000\tT\000\136\216$\217$\022\000 \000\000h\000\164\000d\152@\012)@%P\002#`\147d\144X\000\128\000\001\160\002\144\001\146a\0000\165\000\149@\b\141\130M\146A`\002\000\000\000\000\000\000\000\000\000\000\000\000\002U\000\"6\t6I\005\128\b\000\000\000\000\000\000\000\000\000\000\000\000\tT\000\136\216$\217$\022\000 \000\000\000\000\000\000\000\000\000\000\000\000%P\002#`\147d\144X\000\128\000\000\000\000\000\000\000\000\000\000\000\000\149@\b\141\130M\146A`\002\000\000\000\000\000\000\000\000\000\000\000\000\002U\000\"6\t6I\005\128\b\000\000\026\000)\000\025&\016\003\nP\tT\000\136\216$\217$\022\000 \000\000h\000\164\000d\152@\012)@%P\002#`\147d\144X\000\128\000\001\160\002\144\001\146a\0000\165\000\149@\b\141\130M\146A`\002\000\000\006\128\n@\006I\132\000\194\148\002U\000\"6\t6I\005\128\b\000\000\026\000)\000\025&\016\003\nP\000\000\000\000\000\000\000\000\000\000\000\000\000\b\000\000\000\000\128\000\b\000\000%P\002#`\147d\144X\000\128\000\000\001\000\000\000\000\000\000 \000\003\191\196\024\141\130M\177IxB\000w\248\131\017\176I\182)/\b@\002U\000\"6\t6I\005\128\b\000 \026`i\021Y\166\017\"\nP\tT\000\136\216$\217$\022\000 \000\128i\129\164Uf\152D\136)@%P\002#`\147d\144X\000\128\002\001\166\006\145U\154a\018 \165\000\000\000\000\000\000\000\000\000\000\000\000\b\000\128\000\000\000\002\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\128\b\000\000\000\000 \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\004\170\000Dl\018l\146\011\000\016\000@4\192\210*\179L\"D\020\160\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\b\000\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\002\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\016\000\000\000\000\000\004\170\000Dl\018l\146\011\000\016\000\000\000 \000\000\000\000\000\004\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\004\003L\r\"\1714\194$AJ\001*\128\017\027\004\155$\130\192\004\000\016\r04\138\172\211\b\145\005(\004\170\000Dl\018l\146\011\000\016\000@4\192\210*\179L\"D\020\160\018\168\001\017\176I\178H,\000@\001\000\211\003H\170\2050\137\016R\128J\160\004F\193&\201 \176\001\000\004\003L\r\"\1714\194$AJ\001*\128\017\027\004\155$\130\192\004\000\016\r04\138\172\211\b\145\005(\004\170\000Dl\018l\146\011\000\016\000@4\192\210*\179L\"D\020\160\018\168\001\017\176I\178H,\000@\001\000\211\003H\170\2050\137\016R\128J\160\004F\193&\201 \176\001\000\004\003L\r\"\1714\194$AJ\000\000\000\000\000\000\000\000\000\000\000\000\016\r\000\020\128\012\147\b\001\005(\004\170\000Dl\018l\146\011\000\016\000@4\000R\0002L \004\020\160\018\168\001\017\176I\178H,\000@\001\000\208\001H\000\2010\128\016R\128J\160\004F\193&\201 \176\001\000\000\000\000\000\000\000\000\000\000 \000\001*\128\017\027\004\155$\130\192\004\000\000\000\000\000\000\000\000\000\000\000\000\004\170\000Dl\018l\146\011\000\016\000@4\000R\0002L \004\020\160\018\168\001\017\176I\178H,\000@\001\000\208\001H\000\2010\128\016R\128J\160\004F\193&\201 \176\001\000\000\000\000\000\000\000\000\000\000\000\000\001*\128\017\027\004\155$\130\192\004\000\000\000\000\000\000\000\000\000\000\000\000\004\170\000Dl\018l\146\011\000\016\000\000\000\000\000\000\000\000\000\000\000\000\018\168\001\017\176I\178H,\000@\000\000\000\000\000\000\000\000\000\000\000\000J\160\004F\193&\201 \176\001\000\000\000\000\000\000\000\000\000\000\000\000\001*\128\017\027\004\155$\130\192\004\000\000\000\000\000\000\000\000\000\000\000\000\004\170\000Dl\018l\146\011\000\016\000@4\000R\0002L \004\020\160\018\168\001\017\176I\178H,\000@\001\000\208\001H\000\2010\128\016R\128J\160\004F\193&\201 \176\001\000\004\003@\005 \003$\194\000AJ\001*\128\017\027\004\155$\130\192\004\000\016\r\000\020\128\012\147\b\001\005(\004\170\000Dl\018l\146\011\000\016\000@4\000R\0002L \004\020\160\000\000\000\000\000\000\000\000\000\000\000\001\000\223\007\201\255\2210\153\016\247\129\223\192\b\002@\0019\236\204\224\000\000\000\000\000\000\000\000\000\000\000\000\001*\128\017\027\004\155$\130\192\004\000\000\000\000\128\000\000\000\000\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\b\006\248:O\246i\132\200\135\188\002U\000\"6\t6I\005\128\b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\004\000\000\239\241\006#`\147lR^\016\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\004\000@\000\000\000\000\000\000@\000\001*\128\017\027\004\155$\130\192\004\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\128o\131\228\255\238\152L\136{\192\239\224\004\001 \000\156\246fp\000\000\000\000\000\000\000\000\000\000\000\000\000\149@\b\141\130M\146A`\002\000\000\000\000@\000\000\000\000\000\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\128\000\000\000\000\000\004\000\000\000\000\000\000\004\000@\000\000\128\000\000\000\000\000\000\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000@\000\000\000\000\000\000\000\000\001*\136\017\027\004\155$\130\192\004\000\000\001\000\000\000\000\000\000\000\000\000\004\170\001Dl\018l\146\011\000\016\000\000\000 \000\000\000\000\000\000\000\000w\248\131\017\176I\182)/\b@\000\000\000\000\000\000\000\000\000\000\000\000\000\002\000\000\000\000\000\000\000\000\000\tT@\136\216$\217$\022\000 \000\000\b\000\000\000\000\000\000\000\000\000%P\n#`\147d\144X\000\128\000\000\001\000\000\000\000\000\000\000\000\003\191\196\024\141\130M\177IxB\000\000\000\000\000\000\000\000\000\000\000\000\000\000\223\007I\254\2050\153\016\247\128J\160\004F\193&\201 \176\001\000\000\000\002\000\000\000\000\000\000@\000\007\127\1361\027\004\155b\146\240\132\000\000\000\000\000\000\000\000\000\000\000\000\000\000 \000\000\000\000\000\000 \000\000\149@\b\141\130M\146A`\002\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001\223\226\014F\193&\217\172\252a\001\187\252A\136\216$\219\020\151\132 \000\000\000\000\000\000\000\000\000\000\000\000\239\241\007#`\147lR~0\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\128\000\000\000\000\016\000\000\000\000\000\000\000\000\000\000\000\b\000\001\223\226\014F\193&\216\164\188a\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000 \000\000\000\000\000\004\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\128\000\000\000\000\016\016\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\002\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000 \000@\000\000\000\000\000\000\000@\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\b\000\000\000\000\000\000\000\000\000\000\000\000@\000\000\000\000\000\000\016\000\000\000\000\000\000\000\000\000\000\000\000\000\000\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\029\254 \228l\018m\158\207\198\016\024\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\b\000\000\000\000\000\000\000\000\000\004\000\000\000\000\000\000\000\000@\000\000\000 \000\000\000\000\000\000\b\000\000\000\000\000\000\000\000\000\000\000\000\000\000@\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\014\255\016r6\t6\205%\225\b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000@4\230\210*\179L\"F\020\161\b\006\156\218EVi\132H\194\148 \000\000\000\000\000\000\000\000\000\000\000J\160\005F\193&\201 \178\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\004\000\000\000\000\000\b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\002\000\000\000\000\000\004\000\000\018\168\001Q\176I\178H,\128@\000\000\000\000\000\000\000\000\000\000\000\000J\160\004F\193&\201 \178\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\004\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\128\000\000\000\000\000\128\000\014\255\000P2\000\t\207fg\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000@4\230\210*\179L\"F\020\161\000\000\004\000\000\000\000\000\000\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\004\003~\127'\255t\194dc\222\016\000\000\000\000\000\000\000\000\000\000\000\016\r9\180\138\172\211\b\145\133(B\001\1676\145U\154a\0180\165\b@4\230\210*\179L\"F\020\161\b\006\156\218EVi\132H\194\148 \000\000\128\000\000\000\000\000\016\000\001\223\226\012F\193&\216\164\188!\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\239\241\006#`\147l\210^\016\128\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000 \000\000\000\000\000\000\000\000\000\000\000J\160\004F\193&\201 \176\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\b\000\002\000\000\000\000\000\000\000\000\000\000\000\000")
  
  and start =
    2
  
  and action =
    ((16, "\005\180\000\000\005\180\000\000\006h\000O \146 \146 \146\000\000\000\000\000\000\000\000 \146\000\000\000\000 \146 \146 \146 \146\019\162\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\000\000\015 \146 \146\000\000 \146 \146\000\000\000W\000\000\000\000\0006\000:\000%\000W\000\000\000\000\001\000\000\214\007\150\000\230 \146\000\000 \146 \146 \146 \146\000\000\000\000\000\000\000\000\b \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\t\019\162\000\000 \146 \146\000\000\b\178 \146\tD \146\t\214 \146\nh\000\000\000\000\000\000\000\000\000\156\019\162\000\000 \146\"6 \146\"\146 \146\000\000\000\000\000\000 \146\000\000 \146\000\000 \146\000\000 \146\000\000 \146\000\000 \146\"\238 \146\001@\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000 \146\000\000\000\000\000\194 \146\000\000 \146#J \146#\166 \146$\002 \146$^ \146$\186 \146%\022 \146%r\0008\000\000 \146\000\000\000\000\000\000 \146\n\250 \146\011\140 \146\012\030 \146\012\176 \146\rB\000\000\000\000\001T\000\000\000\000\000\027 \146\000\000 \146\000J\000\000\000\000\r\212\014f\014\248\0008\006h \146\020\186\000W\000\136 \146\000\000\000\030\000W\000\000\000\000\000\000 \146\021L\001B\006h \146\000\146\001N \146\000\138\001T\000) \146\000[\006h\000\000!\000\000\000\001\b \146\021\222 \146\022p \146\000\000\000\t\000\000 \146\023\002 \146\023\148 \146\024&\001\\ \146\000\144\006h\001\228\001^ \146\000W\001f \146\000\000\024\174%\206 \146&* \146&\134 \146\001@ \146\000\000\000\000\001& \146\000\000 \146&\226 \146'> \146\000\000 \146\000\000 \146\000\000 \146\000\000 \146\000\000 \146'\154 \146'\246 \146(R \146(\174 \146)\n\000\000\000\019 \146\000\220\006h\006h \146\025@ \146\025\210 \146\026d\000\000\000\004\000\000\000\000\000\000\000\004\000\000 \146\026\246\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\242\000\000\000\000\000\000\001\186\001n \146\000\222\000\000\000\000\000\000\027\136 \146\028\026 \146\028\172 \146\029> \146\029\208 \146\030b \146\030\244 \146\031\134 \146 \024\000\000)f \146)\194 \146*\030 \146\0018 \146\000\000 \146*z \146*\214 \146\000\000 \146\000\000 \146\000\000 \146\000\000 \146\000\000 \146\000\000 \146+2 \146+\142 \146+\234 \146,F \146,\162\000\000\003*\019\162\000\000 \146\000P\000\000\000\000\005\n \146\000\000\000\000\001:\006h\000\000\000\000\000\000\000\000\000\000\000\194 \146\000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\196\019\162\000\000 \146\000~\000\000\000\000\000\000\000\000\000\000\000W\000\029\000\019\000\000\000\000\001\180!\000\001\182!\000\001\174\006h\000\000\001\194!\000\001\206!\000\001\198\006h\000\000\006\236 \146\000\234\006h\000\000\000\188 \146\000\000\000\000\004n\006h\000\000\006h\000\000\000\000\000\210\001`\006h\000\000\001\200\000\000\000)\000\000\000)\000\000\001\202\000\000\000\000\001\n\001B\000\000\001\164\000W\001\216\001B\000\000\000\000\002\142\000\000\000\000\000\000\000W\001\184\000:\001\232\000\214\000\000\000\000\007\150\000\000\000\000\001\228\000\000\000\000\015\138\016\028\000\000!\000\000\000\000\244\000\000\000\000\001\028!j\000\000!\212\000\000\001\228\000\000\000\000\000\000\001\026\020:\000\000\000\000\000\000\001\242\000\000\016\174\001,\000\000\000\000\000\t\000\000\017@\017\210\018d\018\246\0014\006h\000\000\000\000\005\180\000\000\000\000\001p\000\000 \146\000\000\000\011\000\000"), (16, "\002=\002\153\002\129\004m\001\237\000\026\002\137\004e\004U\002=\002=\000\242\002=\002=\002=\002=\002=\002=\002\137\004]\002=\002=\002=\002=\002=\001v\002=\002\162\003~\002=\000\214\004]\002=\002=\002=\002=\002=\002=\002=\002=\001~\002=\002=\004\150\002\250\002=\002=\001\237\002=\003\026\002\021\002=\002=\002\158\001\014\003v\004\198\002=\002\129\002\242\002=\002=\003\002\006\154\002=\006\026\000\222\002\162\002=\002=\003\238\001\006\002y\002=\002=\002=\002=\004\241\002=\002=\002=\002=\001\217\000\242\002y\bw\002=\003:\006\134\004\241\003n\001\217\001\217\003\214\001\217\001\217\001\217\001\217\001\217\001\217\002\162\001%\001\217\001\217\001\217\001\217\001\217\001\186\001\217\002\162\002m\001\217\002\162\001%\001\217\001\217\001\217\001\217\001\217\001\217\001\217\001\217\001~\001\217\001\217\005\006\001\194\001\217\001\217\002\145\001\217\004\158\005\"\001\217\001\217\002\162\002e\005\n\006\242\001\217\002y\002\162\001\217\001\217\002\162\002\162\001\217\007\214\002e\001\030\001\217\001\217\003.\002e\001\"\001\217\001\217\001\217\001\217\001*\001\217\001\217\001\217\001\217\003e\004u\006\254\000\233\001\217\006R\002Z\b&\007&\003e\003e\bJ\003e\002\026\002\030\003e\003e\003e\002\162\002\162\003e\003e\002\"\003e\003e\002\162\003e\002\230\007Z\003e\007\218\002\129\002&\003e\002*\003e\002.\003e\0022\003e\002\162\003e\003e\007b\003V\003e\003e\003f\003e\b\006\007\198\003e\003e\003r\003\206\003\226\004&\003e\002\162\005\022\0026\003e\002\162\005\026\003e\005\146\0066\002\129\003e\003e\006\178\006\186\006\194\002:\003e\002>\003e\006\206\002B\003e\002F\003e\005\005\006\214\006\222\007*\003e\002\129\0076\007N\007f\005\005\005\005\007n\005\005\005\005\005\005\005\005\005\005\003\238\007\146\007\154\007\182\007\246\005\005\005\005\005\005\005\005\005\005\b\026\bg\005\005\000\000\000\000\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005\005\000\000\005\005\005\005\005\005\000\000\005\005\000\000\000\000\005\005\005\005\000\000\000\000\000\000\000\000\005\005\000\000\000\000\005\005\005\005\000\000\000\000\005\005\000\000\000\000\000\000\005\005\001\205\000\000\000\000\000\000\005\005\005\005\005\005\005\005\000\000\005\005\005\005\005\005\005\005\002q\002q\002q\000\000\002q\002q\002q\002q\002q\002q\002q\002q\000\000\000\000\000\000\002q\000\000\000\000\000\000\000\000\000\000\002q\002q\002q\000\000\000\000\002q\000\000\000\000\000\000\002q\002q\000\000\002q\002q\000\000\000\000\000\000\000\000\000\000\002q\000\000\000\000\002q\000\000\000\000\002q\002q\000\000\002q\002q\000\000\002q\002q\000\000\000\000\002q\002q\007Z\002q\000\000\002q\002q\000\000\000\000\002q\002q\002q\002q\002q\002q\000\000\000\000\000\000\002q\002q\000\000\000\000\000\000\000\000\002q\002E\000\000\000\000\000\000\000\000\000\000\000\000\002q\002q\002E\002E\000\000\002E\002E\002E\002E\002E\000\000\000\000\000\000\000\000\000\000\002E\002E\002E\006\n\002E\000\000\000\000\002E\000\000\000\000\002E\002E\002E\002E\002E\002E\002E\002E\001~\002E\002E\000\000\006\018\002E\002E\000\000\002E\000\000\000\000\002E\002E\000\000\000\000\000\000\000\000\002E\000\000\000\000\002E\002E\000\000\000\000\002E\000\000\000\000\000\000\002E\000\000\000\000\000\000\000\000\002E\002E\002E\002E\001\225\002E\002E\002E\002E\000\000\000\000\000\000\000\000\001\225\001\225\000\000\001\225\001\225\001\225\001\225\001\225\000\000\000\000\000\000\000\000\000\000\001\225\001\225\001\225\006v\001\225\000\000\000\000\001\225\000\000\000\000\001\225\001\225\001\225\001\225\001\225\001\225\001\225\001\225\001~\001\225\001\225\000\000\006~\001\225\001\225\000\000\001\225\000\000\000\000\001\225\001\225\000\000\000\000\000\000\000\000\001\225\000\000\000\000\001\225\001\225\000\000\000\000\001\225\000\000\000\000\000\000\001\225\000\000\000\000\000\000\000\000\001\225\001\225\001\225\001\225\000\000\001\225\001\225\001\225\001\225\001\193\001\193\001\193\000\000\001\193\001\193\001\193\001\193\001\193\001\193\001\193\001\193\000\000\000\000\000\000\001\193\000\000\000\000\000\000\000\000\000\000\001\193\001\193\001\193\000\000\000\000\001\193\000\000\000\000\000\000\001\193\001\193\000\000\001\193\001\193\000\000\000\000\000\000\000\000\000\000\001\193\000\000\000\000\001\193\000\000\000\000\001\193\001\193\000\000\001\193\001\193\000\000\001\193\001\193\000\000\000\000\001\193\001\193\000\000\001\193\000\000\007\014\001\193\000\000\000\000\001\193\001\193\001\193\001\193\001\193\001\193\000\000\000\000\000\000\001\193\001\193\000\000\000\000\000\000\000\000\001\193\003\185\000\000\000\000\000\000\000\000\000\000\000\000\001\193\001\193\003\185\003\185\000\000\003\185\002\026\002\030\003\185\003\185\000\000\000\000\000\000\000\000\000\000\002\"\003\185\003\185\000\000\003\185\000\000\000\000\003\185\000\000\000\000\002&\003\185\002*\003\185\002.\003\185\0022\003\185\000\000\003\185\003\185\000\000\000\000\003\185\003\185\000\000\003\185\000\000\000\000\003\185\003\185\000\000\000\000\000\000\000\000\003\185\000\000\000\000\0026\003\185\000\000\000\000\003\185\000\000\000\000\000\000\003\185\000\000\000\000\000\000\000\000\002:\003\185\002>\003\185\000\000\002B\003\185\002F\003\185\001\229\001\229\001\229\000\000\001\229\001\229\001\229\001\229\001\229\001\229\001\229\001\229\000\000\000\000\000\000\001\229\000\000\000\000\000\000\000\000\000\000\001\229\001\229\000\000\000\000\000\000\001\229\000\000\000\000\000\000\001\229\001\229\000\000\001\229\001\229\000\000\000\000\000\000\000\000\000\000\001\229\000\000\000\000\001\229\000\000\000\000\001\229\001\229\000\000\001\229\001\229\000\000\001\229\001\229\000\000\000\000\001\229\001\229\000\000\001\229\000\000\000\000\001\229\000\000\000\000\001\229\000\000\001\229\001\229\001\229\001\229\000\000\000\000\000\000\000\000\001\229\000\000\000\000\000\000\000\000\001\229\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\006\001\253\000\022\001&\003\030\000\000\003&\003J\003R\000&\003Z\000*\003b\000.\000\000\000\000\000\000\003\130\000\000\000\000\000\000\000\000\000\000\003\134\0002\002\005\000\000\000\000\003\146\000\000\000\000\000\000\000:\000>\000\000\003\154\003\162\000\000\000\000\000\000\000\000\000\000\003\178\000\000\000\000\000J\000\000\000\000\003V\000\226\000\000\003\186\003\194\000\000\003\202\003\218\000\000\000\000\007\138\003\222\000\000\001.\000\000\000\000\004\162\000\000\000\000\004\166\002\005\004\174\004\182\004\190\004\194\000\000\000\000\000\000\002\005\004\210\000\000\000\000\003e\003e\004\218\003e\002\026\002\030\003e\003e\000\000\000\000\000\000\000\000\000\000\002\"\003e\003e\000\000\003e\000\000\000\000\003e\000\000\000\000\002&\003e\002*\003e\002.\003e\0022\003e\000\000\003e\003e\000\000\000\000\003e\003e\000\000\006\234\000\000\000\000\003e\003e\000\000\000\000\000\000\000\000\003e\000\000\000\000\0026\003e\000\000\000\000\003e\000\000\000\000\000\000\003e\000\000\000\000\000\000\000\000\002:\003e\002>\003e\000\000\002B\003e\002F\003e\000\022\001&\003\030\000\000\003&\003J\003R\000&\003Z\000*\003b\000.\000\000\000\000\000\000\003\130\000\000\000\000\000\000\000\000\000\000\003\134\0002\001\245\000\000\000\000\003\146\000\000\000\000\000\000\000:\000>\000\000\003\154\003\162\000\000\000\000\000\000\000\000\000\000\003\178\000\000\000\000\000J\000\000\000\000\003V\000\226\000\000\003\186\003\194\000\000\003\202\003\218\000\000\000\000\007\138\003\222\000\000\001.\000\000\000\000\004\162\000\000\000\000\004\166\000\000\004\174\004\182\004\190\004\194\000\000\003\145\000\000\000\000\004\210\000\000\000\000\000\000\000\000\004\218\003\145\003\145\000\000\003\145\000\000\000\000\003\145\003\145\003\145\000\000\000\000\003\145\003\145\000\000\003\145\003\145\000\000\003\145\000\000\000\000\003\145\000\000\000\000\000\000\003\145\000\000\003\145\000\000\003\145\000\000\003\145\000\000\003\145\003\145\000\000\000\000\003\145\001V\000\000\003\145\000\000\000\000\003\145\003\145\000\000\000\000\000\000\000\000\003\145\000\000\000\000\000\000\003\145\000\000\000\000\001Z\000\000\000\000\000\000\003\145\003\145\000\000\000\000\000\000\002\161\003\145\000\000\003\145\000\000\000\000\003\145\000\000\003\145\002\161\002\161\000\000\002\161\003\145\000\000\001\142\002\202\002\161\000\000\000\000\002\161\002\161\000\000\001\150\002\161\000\000\002\161\000\000\000\000\002\161\000\000\000\000\000\000\001\158\000\000\002\178\000\000\002\194\000\000\002\210\000\000\002\161\002\161\000\000\000\000\002\161\001V\000\000\002\161\000\000\000\000\002\161\002\161\000\000\000\000\000\000\000\000\002\161\000\000\000\000\000\000\002\186\000\000\000\000\001Z\000\000\000\000\000\000\002\161\002\161\000\000\000\000\000\000\003\181\002\161\000\000\002\161\000\000\000\000\002\161\000\000\002\161\003\181\003\181\000\000\003\181\002\161\000\000\003\181\003\181\003\181\000\000\000\000\003\181\003\181\000\000\001\150\003\181\000\000\003\181\000\000\000\000\003\181\000\000\000\000\000\000\001\158\000\000\002\178\000\000\002\194\000\000\003\181\000\000\003\181\003\181\000\000\000\000\003\181\001V\000\000\003\181\000\000\000\000\003\181\003\181\000\000\000\000\000\000\000\000\003\181\000\000\000\000\000\000\002\186\000\000\000\000\001Z\000\000\000\000\000\000\003\181\003\181\000\000\000\000\000\000\003\165\003\181\000\000\003\181\000\000\000\000\003\181\000\000\003\181\003\165\003\165\000\000\003\165\003\181\000\000\003\165\003\165\003\165\000\000\000\000\003\165\003\165\000\000\003\165\003\165\000\000\003\165\000\000\000\000\003\165\000\000\000\000\000\000\001\158\000\000\002\178\000\000\003\165\000\000\003\165\000\000\003\165\003\165\000\000\000\000\003\165\001V\000\000\003\165\000\000\000\000\003\165\003\165\000\000\000\000\000\000\000\000\003\165\000\000\000\000\000\000\002\186\000\000\000\000\001Z\000\000\000\000\000\000\003\165\003\165\000\000\000\000\000\000\003\153\003\165\000\000\003\165\000\000\000\000\003\165\000\000\003\165\003\153\003\153\000\000\003\153\003\165\000\000\003\153\003\153\003\153\000\000\000\000\003\153\003\153\000\000\003\153\003\153\000\000\003\153\000\000\000\000\003\153\000\000\000\000\000\000\003\153\000\000\003\153\000\000\003\153\000\000\003\153\000\000\003\153\003\153\000\000\000\000\003\153\001V\000\000\003\153\000\000\000\000\003\153\003\153\000\000\000\000\000\000\000\000\003\153\000\000\000\000\000\000\003\153\000\000\000\000\001Z\000\000\000\000\000\000\003\153\003\153\000\000\000\000\000\000\003\161\003\153\000\000\003\153\000\000\000\000\003\153\000\000\003\153\003\161\003\161\000\000\003\161\003\153\000\000\003\161\003\161\003\161\000\000\000\000\003\161\003\161\000\000\003\161\003\161\000\000\003\161\000\000\000\000\003\161\000\000\000\000\000\000\003\161\000\000\003\161\000\000\003\161\000\000\003\161\000\000\003\161\003\161\000\000\000\000\003\161\001V\000\000\003\161\000\000\000\000\003\161\003\161\000\000\000\000\000\000\000\000\003\161\000\000\000\000\000\000\003\161\000\000\000\000\001Z\000\000\000\000\000\000\003\161\003\161\000\000\000\000\000\000\003\157\003\161\000\000\003\161\000\000\000\000\003\161\000\000\003\161\003\157\003\157\000\000\003\157\003\161\000\000\003\157\003\157\003\157\000\000\000\000\003\157\003\157\000\000\003\157\003\157\000\000\003\157\000\000\000\000\003\157\000\000\000\000\000\000\003\157\000\000\003\157\000\000\003\157\000\000\003\157\000\000\003\157\003\157\000\000\000\000\003\157\001V\000\000\003\157\000\000\000\000\003\157\003\157\000\000\000\000\000\000\000\000\003\157\000\000\000\000\000\000\003\157\000\000\000\000\001Z\000\000\000\000\000\000\003\157\003\157\000\000\000\000\000\000\003\169\003\157\000\000\003\157\000\000\000\000\003\157\000\000\003\157\003\169\003\169\000\000\003\169\003\157\000\000\003\169\003\169\003\169\000\000\000\000\003\169\003\169\000\000\003\169\003\169\000\000\003\169\000\000\000\000\003\169\000\000\000\000\000\000\001\158\000\000\002\178\000\000\003\169\000\000\003\169\000\000\003\169\003\169\000\000\000\000\003\169\001V\000\000\003\169\000\000\000\000\003\169\003\169\000\000\000\000\000\000\000\000\003\169\000\000\000\000\000\000\002\186\000\000\000\000\001Z\000\000\000\000\000\000\003\169\003\169\000\000\000\000\000\000\003\177\003\169\000\000\003\169\000\000\000\000\003\169\000\000\003\169\003\177\003\177\000\000\003\177\003\169\000\000\003\177\003\177\003\177\000\000\000\000\003\177\003\177\000\000\001\150\003\177\000\000\003\177\000\000\000\000\003\177\000\000\000\000\000\000\001\158\000\000\002\178\000\000\002\194\000\000\003\177\000\000\003\177\003\177\000\000\000\000\003\177\001V\000\000\003\177\000\000\000\000\003\177\003\177\000\000\000\000\000\000\000\000\003\177\000\000\000\000\000\000\002\186\000\000\000\000\001Z\000\000\000\000\000\000\003\177\003\177\000\000\000\000\000\000\003\173\003\177\000\000\003\177\000\000\000\000\003\177\000\000\003\177\003\173\003\173\000\000\003\173\003\177\000\000\003\173\003\173\003\173\000\000\000\000\003\173\003\173\000\000\001\150\003\173\000\000\003\173\000\000\000\000\003\173\000\000\000\000\000\000\001\158\000\000\002\178\000\000\002\194\000\000\003\173\000\000\003\173\003\173\000\000\000\000\003\173\001V\000\000\003\173\000\000\000\000\003\173\003\173\000\000\000\000\000\000\000\000\003\173\000\000\000\000\000\000\002\186\000\000\000\000\001Z\000\000\000\000\000\000\003\173\003\173\000\000\000\000\000\000\003\129\003\173\000\000\003\173\000\000\000\000\003\173\000\000\003\173\003\129\003\129\000\000\003\129\003\173\000\000\003\129\003\129\003\129\000\000\000\000\003\129\003\129\000\000\003\129\003\129\000\000\003\129\000\000\000\000\003\129\000\000\000\000\000\000\003\129\000\000\003\129\000\000\003\129\000\000\003\129\000\000\003\129\003\129\000\000\000\000\003\129\001V\000\000\003\129\000\000\000\000\003\129\003\129\000\000\000\000\000\000\000\000\003\129\000\000\000\000\000\000\003\129\000\000\000\000\001Z\000\000\000\000\000\000\003\129\003\129\000\000\000\000\000\000\003\133\003\129\000\000\003\129\000\000\000\000\003\129\000\000\003\129\003\133\003\133\000\000\003\133\003\129\000\000\003\133\003\133\003\133\000\000\000\000\003\133\003\133\000\000\003\133\003\133\000\000\003\133\000\000\000\000\003\133\000\000\000\000\000\000\003\133\000\000\003\133\000\000\003\133\000\000\003\133\000\000\003\133\003\133\000\000\000\000\003\133\001V\000\000\003\133\000\000\000\000\003\133\003\133\000\000\000\000\000\000\000\000\003\133\000\000\000\000\000\000\003\133\000\000\000\000\001Z\000\000\000\000\000\000\003\133\003\133\000\000\000\000\000\000\003m\003\133\000\000\003\133\000\000\000\000\003\133\000\000\003\133\003m\003m\000\000\003m\003\133\000\000\003m\003m\003m\000\000\000\000\003m\003m\000\000\003m\003m\000\000\003m\000\000\000\000\003m\000\000\000\000\000\000\003m\000\000\003m\000\000\003m\000\000\003m\000\000\003m\003m\000\000\000\000\003m\001V\000\000\003m\000\000\000\000\003m\003m\000\000\000\000\000\000\000\000\003m\000\000\000\000\000\000\003m\000\000\000\000\001Z\000\000\000\000\000\000\003m\003m\000\000\000\000\000\000\003y\003m\000\000\003m\000\000\000\000\003m\000\000\003m\003y\003y\000\000\003y\003m\000\000\003y\003y\003y\000\000\000\000\003y\003y\000\000\003y\003y\000\000\003y\000\000\000\000\003y\000\000\000\000\000\000\003y\000\000\003y\000\000\003y\000\000\003y\000\000\003y\003y\000\000\000\000\003y\001V\000\000\003y\000\000\000\000\003y\003y\000\000\000\000\000\000\000\000\003y\000\000\000\000\000\000\003y\000\000\000\000\001Z\000\000\000\000\000\000\003y\003y\000\000\000\000\000\000\003}\003y\000\000\003y\000\000\000\000\003y\000\000\003y\003}\003}\000\000\003}\003y\000\000\003}\003}\003}\000\000\000\000\003}\003}\000\000\003}\003}\000\000\003}\000\000\000\000\003}\000\000\000\000\000\000\003}\000\000\003}\000\000\003}\000\000\003}\000\000\003}\003}\000\000\000\000\003}\001V\000\000\003}\000\000\000\000\003}\003}\000\000\000\000\000\000\000\000\003}\000\000\000\000\000\000\003}\000\000\000\000\001Z\000\000\000\000\000\000\003}\003}\000\000\000\000\000\000\003\141\003}\000\000\003}\000\000\000\000\003}\000\000\003}\003\141\003\141\000\000\003\141\003}\000\000\003\141\003\141\003\141\000\000\000\000\003\141\003\141\000\000\003\141\003\141\000\000\003\141\000\000\000\000\003\141\000\000\000\000\000\000\001\158\000\000\002\178\000\000\003\141\000\000\003\141\000\000\003\141\003\141\000\000\000\000\003\141\001V\000\000\003\141\000\000\000\000\003\141\003\141\000\000\000\000\000\000\000\000\003\141\000\000\000\000\000\000\002\186\000\000\000\000\001Z\000\000\000\000\000\000\003\141\003\141\000\000\000\000\000\000\003\149\003\141\000\000\003\141\000\000\000\000\003\141\000\000\003\141\003\149\003\149\000\000\003\149\003\141\000\000\003\149\003\149\003\149\000\000\000\000\003\149\003\149\000\000\003\149\003\149\000\000\003\149\000\000\000\000\003\149\000\000\000\000\000\000\003\149\000\000\003\149\000\000\003\149\000\000\003\149\000\000\003\149\003\149\000\000\000\000\003\149\001V\000\000\003\149\000\000\000\000\003\149\003\149\000\000\000\000\000\000\000\000\003\149\000\000\000\000\000\000\003\149\000\000\000\000\001Z\000\000\000\000\000\000\003\149\003\149\000\000\000\000\000\000\003\137\003\149\000\000\003\149\000\000\000\000\003\149\000\000\003\149\003\137\003\137\000\000\003\137\003\149\000\000\003\137\003\137\003\137\000\000\000\000\003\137\003\137\000\000\003\137\003\137\000\000\003\137\000\000\000\000\003\137\000\000\000\000\000\000\001\158\000\000\002\178\000\000\003\137\000\000\003\137\000\000\003\137\003\137\000\000\000\000\003\137\001V\000\000\003\137\000\000\000\000\003\137\003\137\000\000\000\000\000\000\000\000\003\137\000\000\000\000\000\000\002\186\000\000\000\000\001Z\000\000\000\000\000\000\003\137\003\137\000\000\000\000\000\000\003u\003\137\000\000\003\137\000\000\000\000\003\137\000\000\003\137\003u\003u\000\000\003u\003\137\000\000\003u\003u\003u\000\000\000\000\003u\003u\000\000\003u\003u\000\000\003u\000\000\000\000\003u\000\000\000\000\000\000\003u\000\000\003u\000\000\003u\000\000\003u\000\000\003u\003u\000\000\000\000\003u\001V\000\000\003u\000\000\000\000\003u\003u\000\000\000\000\000\000\000\000\003u\000\000\000\000\000\000\003u\000\000\000\000\001Z\000\000\000\000\000\000\003u\003u\000\000\000\000\000\000\003q\003u\000\000\003u\000\000\000\000\003u\000\000\003u\003q\003q\000\000\003q\003u\000\000\003q\003q\003q\000\000\000\000\003q\003q\000\000\003q\003q\000\000\003q\000\000\000\000\003q\000\000\000\000\000\000\003q\000\000\003q\000\000\003q\000\000\003q\000\000\003q\003q\000\000\000\000\003q\001V\000\000\003q\000\000\000\000\003q\003q\000\000\000\000\000\000\000\000\003q\000\000\000\000\000\000\003q\000\000\000\000\001Z\000\000\000\000\000\000\003q\003q\000\000\000\000\000\000\000\000\003q\000\000\003q\000\000\000\000\003q\000\000\003q\000\000\000V\000Z\000^\003q\000b\000f\000j\000n\000r\000v\000z\000~\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\130\000\000\001\005\000\000\000\000\000\000\000\000\000\000\000\000\000:\000\134\000\000\000\000\000\138\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\142\000\000\000\000\000\146\000\150\000\154\000\000\000\000\000\158\000\162\000\166\000\170\000\000\000\174\000\178\000\000\000\000\000\182\000\186\000\000\000\000\000\190\000\194\000\000\000\000\000\198\000\202\000\206\000V\000Z\000^\000\000\000b\000f\000j\000n\000r\000v\000z\000~\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\130\000\000\004y\000\000\000\000\000\000\000\000\000\000\000\000\000:\000\134\000\000\000\000\000\138\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\142\000\000\000\000\000\146\000\150\000\154\000\000\000\000\000\158\000\162\000\166\000\170\000\000\000\174\000\178\000\000\000\000\000\182\000\186\003\197\000\000\000\190\000\194\000\000\000\000\000\198\000\202\000\206\003\197\003\197\000\000\003\197\000\000\000\000\003\197\003\197\000\000\000\000\000\000\000\000\000\000\000\000\003\197\003\197\000\000\003\197\000\000\000\000\003\197\000\000\000\000\000\000\003\197\000\000\003\197\000\000\003\197\000\000\003\197\000\000\003\197\003\197\000\000\000\000\003\197\001V\000\000\003\197\000\000\000\000\003\197\003\197\000\000\000\000\000\000\000\000\003\197\000\000\000\000\000\000\003\197\000\000\000\000\001Z\000\000\000\000\000\000\003\197\000\000\000\000\000\000\000\000\003\201\003\197\000\000\003\197\000\000\000\000\003\197\000\000\003\197\003\201\003\201\000\000\003\201\000\000\000\000\003\201\003\201\000\000\000\000\000\000\000\000\000\000\000\000\003\201\003\201\000\000\003\201\000\000\000\000\003\201\000\000\000\000\000\000\003\201\000\000\003\201\000\000\003\201\000\000\003\201\000\000\003\201\003\201\000\000\000\000\003\201\001V\000\000\003\201\000\000\000\000\003\201\003\201\000\000\000\000\000\000\000\000\003\201\000\000\000\000\000\000\003\201\000\000\000\000\001Z\000\000\000\000\000\000\003\201\000\000\000\000\000\000\000\000\003\221\003\201\000\000\003\201\000\000\000\000\003\201\000\000\003\201\003\221\003\221\000\000\003\221\000\000\000\000\003\221\003\221\000\000\000\000\000\000\000\000\000\000\000\000\003\221\003\221\000\000\003\221\000\000\000\000\003\221\000\000\000\000\000\000\001\158\000\000\002\178\000\000\003\221\000\000\003\221\000\000\003\221\003\221\000\000\000\000\003\221\001V\000\000\003\221\000\000\000\000\003\221\003\221\000\000\000\000\000\000\000\000\003\221\000\000\000\000\000\000\002\186\000\000\000\000\001Z\000\000\000\000\000\000\003\221\000\000\000\000\000\000\000\000\003\233\003\221\000\000\003\221\000\000\000\000\003\221\000\000\003\221\003\233\003\233\000\000\003\233\000\000\000\000\003\233\003\233\000\000\000\000\000\000\000\000\000\000\000\000\003\233\003\233\000\000\003\233\000\000\000\000\003\233\000\000\000\000\000\000\003\233\000\000\003\233\000\000\003\233\000\000\003\233\000\000\003\233\003\233\000\000\000\000\003\233\001V\000\000\003\233\000\000\000\000\003\233\003\233\000\000\000\000\000\000\000\000\003\233\000\000\000\000\000\000\003\233\000\000\000\000\001Z\000\000\000\000\000\000\003\233\000\000\000\000\000\000\000\000\003\225\003\233\000\000\003\233\000\000\000\000\003\233\000\000\003\233\003\225\003\225\000\000\003\225\000\000\000\000\003\225\003\225\000\000\000\000\000\000\000\000\000\000\000\000\003\225\003\225\000\000\003\225\000\000\000\000\003\225\000\000\000\000\000\000\001\158\000\000\002\178\000\000\003\225\000\000\003\225\000\000\003\225\003\225\000\000\000\000\003\225\001V\000\000\003\225\000\000\000\000\003\225\003\225\000\000\000\000\000\000\000\000\003\225\000\000\000\000\000\000\002\186\000\000\000\000\001Z\000\000\000\000\000\000\003\225\000\000\000\000\000\000\000\000\003\209\003\225\000\000\003\225\000\000\000\000\003\225\000\000\003\225\003\209\003\209\000\000\003\209\000\000\000\000\003\209\003\209\000\000\000\000\000\000\000\000\000\000\000\000\003\209\003\209\000\000\003\209\000\000\000\000\003\209\000\000\000\000\000\000\003\209\000\000\003\209\000\000\003\209\000\000\003\209\000\000\003\209\003\209\000\000\000\000\003\209\001V\000\000\003\209\000\000\000\000\003\209\003\209\000\000\000\000\000\000\000\000\003\209\000\000\000\000\000\000\003\209\000\000\000\000\001Z\000\000\000\000\000\000\003\209\000\000\000\000\000\000\000\000\003\205\003\209\000\000\003\209\000\000\000\000\003\209\000\000\003\209\003\205\003\205\000\000\003\205\000\000\000\000\003\205\003\205\000\000\000\000\000\000\000\000\000\000\000\000\003\205\003\205\000\000\003\205\000\000\000\000\003\205\000\000\000\000\000\000\003\205\000\000\003\205\000\000\003\205\000\000\003\205\000\000\003\205\003\205\000\000\000\000\003\205\001V\000\000\003\205\000\000\000\000\003\205\003\205\000\000\000\000\000\000\000\000\003\205\000\000\000\000\000\000\003\205\000\000\000\000\001Z\000\000\000\000\000\000\003\205\000\000\000\000\000\000\000\000\000\000\003\205\000\000\003\205\002\225\002\225\003\205\002\225\003\205\000\000\001\142\002\202\000\000\000\000\000\000\000\000\000\000\000\000\001\150\002\225\000\000\002\225\000\000\000\000\002\225\000\000\000\000\000\000\001\158\000\000\002\178\000\000\002\194\000\000\002\210\000\000\002\225\002\225\000\000\000\000\002\225\001V\000\000\002\225\000\000\000\000\002\225\002\225\000\000\000\000\000\000\000\000\002\225\000\000\000\000\000\000\002\186\000\000\000\000\001Z\000\000\000\000\000\000\002\225\002\225\000\000\000\000\000\000\003\193\002\225\000\000\002\225\000\000\000\000\002\225\000\000\002\225\003\193\003\193\000\000\003\193\000\000\000\000\003\193\003\193\000\000\000\000\000\000\000\000\000\000\000\000\003\193\003\193\000\000\003\193\000\000\000\000\003\193\000\000\000\000\000\000\003\193\000\000\003\193\000\000\003\193\000\000\003\193\000\000\003\193\003\193\000\000\000\000\003\193\001V\000\000\003\193\000\000\000\000\003\193\003\193\000\000\000\000\000\000\000\000\003\193\000\000\000\000\000\000\003\193\000\000\000\000\001Z\000\000\000\000\000\000\003\193\000\000\000\000\000\000\000\000\003\217\003\193\000\000\003\193\000\000\000\000\003\193\000\000\003\193\003\217\003\217\000\000\003\217\000\000\000\000\003\217\003\217\000\000\000\000\000\000\000\000\000\000\000\000\003\217\003\217\000\000\003\217\000\000\000\000\003\217\000\000\000\000\000\000\003\217\000\000\003\217\000\000\003\217\000\000\003\217\000\000\003\217\003\217\000\000\000\000\003\217\001V\000\000\003\217\000\000\000\000\003\217\003\217\000\000\000\000\000\000\000\000\003\217\000\000\000\000\000\000\003\217\000\000\000\000\001Z\000\000\000\000\000\000\003\217\000\000\000\000\000\000\000\000\003\213\003\217\000\000\003\217\000\000\000\000\003\217\000\000\003\217\003\213\003\213\000\000\003\213\000\000\000\000\003\213\003\213\000\000\000\000\000\000\000\000\000\000\000\000\003\213\003\213\000\000\003\213\000\000\000\000\003\213\000\000\000\000\000\000\003\213\000\000\003\213\000\000\003\213\000\000\003\213\000\000\003\213\003\213\000\000\000\000\003\213\001V\000\000\003\213\000\000\000\000\003\213\003\213\000\000\000\000\000\000\000\000\003\213\000\000\000\000\000\000\003\213\000\000\000\000\001Z\000\000\000\000\000\000\003\213\000\000\000\000\000\000\000\000\003\229\003\213\000\000\003\213\000\000\000\000\003\213\000\000\003\213\003\229\003\229\000\000\003\229\000\000\000\000\003\229\003\229\000\000\000\000\000\000\000\000\000\000\000\000\003\229\003\229\000\000\003\229\000\000\000\000\003\229\000\000\000\000\000\000\003\229\000\000\003\229\000\000\003\229\000\000\003\229\000\000\003\229\003\229\000\000\000\000\003\229\001V\000\000\003\229\000\000\000\000\003\229\003\229\000\000\000\000\000\000\000\000\003\229\000\000\000\000\000\000\003\229\000\000\000\000\001Z\000\000\000\000\000\000\003\229\000\000\000\000\000\000\000\000\003\029\003\229\000\000\003\229\000\000\000\000\003\229\000\000\003\229\003\029\003\029\000\000\003\029\000\000\000\000\0052\005:\000\000\000\000\000\000\000\000\000\000\000\000\005B\003\029\000\000\003\029\000\000\000\000\003\029\000\000\000\000\000\000\005J\000\000\005R\000\000\005Z\000\000\005b\000\000\003\029\003\029\000\000\000\000\003\029\001V\000\000\003\029\000\000\000\000\003\029\003\029\000\000\000\000\000\000\000\000\003\029\000\000\000\000\000\000\005j\000\000\000\000\001Z\000\000\000\000\000\000\003\029\000\000\000\000\000\000\000\000\004\t\003\029\000\000\003\029\000\000\000\000\003\029\000\000\003\029\004\t\004\t\000\000\004\t\000\000\000\000\004\t\004\t\000\000\000\000\000\000\000\000\000\000\000\000\001\150\004\t\000\000\004\t\000\000\000\000\004\t\000\000\000\000\000\000\001\158\000\000\002\178\000\000\002\194\000\000\004\t\000\000\004\t\004\t\000\000\000\000\004\t\001V\000\000\004\t\000\000\000\000\004\t\004\t\000\000\000\000\000\000\000\000\004\t\000\000\000\000\000\000\002\186\000\000\000\000\001Z\000\000\000\000\000\000\004\t\000\000\000\000\000\000\000\000\004\005\004\t\000\000\004\t\000\000\000\000\004\t\000\000\004\t\004\005\004\005\000\000\004\005\000\000\000\000\004\005\004\005\000\000\000\000\000\000\000\000\000\000\000\000\001\150\004\005\000\000\004\005\000\000\000\000\004\005\000\000\000\000\000\000\001\158\000\000\002\178\000\000\002\194\000\000\004\005\000\000\004\005\004\005\000\000\000\000\004\005\001V\000\000\004\005\000\000\000\000\004\005\004\005\000\000\000\000\000\000\000\000\004\005\000\000\000\000\000\000\002\186\000\000\000\000\001Z\000\000\000\000\000\000\004\005\000\000\000\000\000\000\000\000\003\249\004\005\000\000\004\005\000\000\000\000\004\005\000\000\004\005\003\249\003\249\000\000\003\249\000\000\000\000\003\249\003\249\000\000\000\000\000\000\000\000\000\000\000\000\003\249\003\249\000\000\003\249\000\000\000\000\003\249\000\000\000\000\000\000\001\158\000\000\002\178\000\000\003\249\000\000\003\249\000\000\003\249\003\249\000\000\000\000\003\249\001V\000\000\003\249\000\000\000\000\003\249\003\249\000\000\000\000\000\000\000\000\003\249\000\000\000\000\000\000\002\186\000\000\000\000\001Z\000\000\000\000\000\000\003\249\000\000\000\000\000\000\000\000\003\237\003\249\000\000\003\249\000\000\000\000\003\249\000\000\003\249\003\237\003\237\000\000\003\237\000\000\000\000\003\237\003\237\000\000\000\000\000\000\000\000\000\000\000\000\003\237\003\237\000\000\003\237\000\000\000\000\003\237\000\000\000\000\000\000\003\237\000\000\003\237\000\000\003\237\000\000\003\237\000\000\003\237\003\237\000\000\000\000\003\237\001V\000\000\003\237\000\000\000\000\003\237\003\237\000\000\000\000\000\000\000\000\003\237\000\000\000\000\000\000\003\237\000\000\000\000\001Z\000\000\000\000\000\000\003\237\000\000\000\000\000\000\000\000\003\245\003\237\000\000\003\237\000\000\000\000\003\237\000\000\003\237\003\245\003\245\000\000\003\245\000\000\000\000\003\245\003\245\000\000\000\000\000\000\000\000\000\000\000\000\003\245\003\245\000\000\003\245\000\000\000\000\003\245\000\000\000\000\000\000\003\245\000\000\003\245\000\000\003\245\000\000\003\245\000\000\003\245\003\245\000\000\000\000\003\245\001V\000\000\003\245\000\000\000\000\003\245\003\245\000\000\000\000\000\000\000\000\003\245\000\000\000\000\000\000\003\245\000\000\000\000\001Z\000\000\000\000\000\000\003\245\000\000\000\000\000\000\000\000\003\253\003\245\000\000\003\245\000\000\000\000\003\245\000\000\003\245\003\253\003\253\000\000\003\253\000\000\000\000\003\253\003\253\000\000\000\000\000\000\000\000\000\000\000\000\003\253\003\253\000\000\003\253\000\000\000\000\003\253\000\000\000\000\000\000\001\158\000\000\002\178\000\000\003\253\000\000\003\253\000\000\003\253\003\253\000\000\000\000\003\253\001V\000\000\003\253\000\000\000\000\003\253\003\253\000\000\000\000\000\000\000\000\003\253\000\000\000\000\000\000\002\186\000\000\000\000\001Z\000\000\000\000\000\000\003\253\000\000\000\000\000\000\000\000\004\001\003\253\000\000\003\253\000\000\000\000\003\253\000\000\003\253\004\001\004\001\000\000\004\001\000\000\000\000\004\001\004\001\000\000\000\000\000\000\000\000\000\000\000\000\001\150\004\001\000\000\004\001\000\000\000\000\004\001\000\000\000\000\000\000\001\158\000\000\002\178\000\000\002\194\000\000\004\001\000\000\004\001\004\001\000\000\000\000\004\001\001V\000\000\004\001\000\000\000\000\004\001\004\001\000\000\000\000\000\000\000\000\004\001\000\000\000\000\000\000\002\186\000\000\000\000\001Z\000\000\000\000\000\000\004\001\000\000\000\000\000\000\000\000\003\241\004\001\000\000\004\001\000\000\000\000\004\001\000\000\004\001\003\241\003\241\000\000\003\241\000\000\000\000\003\241\003\241\000\000\000\000\000\000\000\000\000\000\000\000\003\241\003\241\000\000\003\241\000\000\000\000\003\241\000\000\000\000\000\000\003\241\000\000\003\241\000\000\003\241\000\000\003\241\000\000\003\241\003\241\000\000\000\000\003\241\001V\000\000\003\241\000\000\000\000\003\241\003\241\000\000\000\000\000\000\000\000\003\241\000\000\000\000\000\030\003\241\003\230\000\"\001Z\000&\000\000\000*\003\241\000.\000\000\000\000\001\130\002i\003\241\000\000\003\241\000\000\002\r\003\241\0002\003\241\000\230\000\000\0006\000\000\000\000\000\000\000:\000>\000\000\000B\000F\000\000\000\000\000\000\000\000\000\000\000N\000\000\000\000\000J\000\000\000\000\000R\000\226\000\000\000\234\000\238\000\000\000\000\000\242\000\000\000\000\000\246\000\030\002a\001.\000\"\000\000\000&\000\000\000*\0012\000.\0016\001:\000\000\002a\007\198\000\000\000\000\000\000\002a\000\000\0002\000\000\000\249\001>\0006\000\000\000\000\000\000\000:\000>\000\000\000B\000F\000\000\000\000\000\000\000\000\000\000\000N\000\000\000\000\000J\000\000\000\000\000R\000\226\000\000\000\234\000\238\000\000\000\000\000\242\000\000\000\249\000\246\000\000\000\249\001.\000\249\000\000\000\249\000\000\000\249\0012\000\000\0016\001:\000\000\000\000\000\249\000\000\000\000\000\000\000\249\000\000\007\230\000\000\000\249\001>\000\000\000\000\000\249\000\249\000\000\000\249\000\249\000\000\000\000\000\000\000\000\000\000\000\249\000\000\000\000\000\249\000\000\000\000\000\249\000\249\000\000\000\249\000\249\000\000\000\000\000\249\000\000\000\030\000\249\000\000\000\"\000\249\000&\000\000\000*\000\000\000.\000\249\000\000\000\249\000\249\000\000\000\000\000\249\000\000\000\000\000\000\0002\000\000\000\000\000\000\0006\000\249\000\000\000\000\000:\000>\000\000\000B\000F\000\000\000\000\000\000\000\000\000\000\000N\000\000\000\000\000J\000\000\000\000\000R\000\226\000\000\000\234\000\238\000\000\000\169\000\242\000\000\000\000\000\246\000\000\000\000\001.\000\000\001\202\002\n\000\000\000\169\0012\000\000\0016\001:\000\169\000\000\007\218\000\169\000\169\000\000\000\000\002\018\000\000\002b\000\000\001>\002j\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001\210\001\226\000\000\000\000\001\234\002\197\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\002\197\002\197\002r\002\197\000\000\000\000\000\000\000\000\002\197\000\000\000\000\002\197\002\197\000\169\000\169\002\197\000\000\002\197\000\000\002z\002\197\002\138\000\000\000\000\002\130\000\000\002\146\000\000\000\000\000\000\000\000\000\169\001\210\001\226\000\000\000\000\001\234\002\193\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\002\193\002\193\002\197\002\193\000\000\000\000\000\000\000\000\002\193\000\000\000\000\002\193\002\193\002\197\002\197\002\193\000\000\002\193\000\000\002\197\002\193\002\197\000\000\000\000\002\197\000\000\002\197\000\000\000\000\000\000\000\000\002\197\001\210\001\226\000\000\000\000\001\234\002\217\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\001\202\002\n\002\193\002\217\000\000\000\000\000\000\000\000\002\217\000\000\000\000\002\217\002\217\002\193\002\193\002\217\000\000\002\217\000\000\002\193\002j\002\193\000\000\000\000\002\193\000\000\002\193\000\000\000\000\000\000\000\000\002\193\001\210\001\226\000\000\000\000\001\234\002\189\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\002\189\002\189\002r\002\189\000\000\000\000\000\000\000\000\002\189\000\000\000\000\002\189\002\189\002\217\002\217\002\189\000\000\002\189\000\000\002z\002\189\002\138\000\000\000\000\002\130\000\000\002\146\000\000\000\000\000\000\000\000\002\217\001\210\001\226\000\000\000\000\001\234\002\185\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\002\185\002\185\002\189\002\185\000\000\000\000\000\000\000\000\002\185\000\000\000\000\002\185\002\185\002\189\002\189\002\185\000\000\002\185\000\000\002\189\002\185\002\189\000\000\000\000\002\189\000\000\002\189\000\000\000\000\000\000\000\000\002\189\001\210\001\226\000\000\000\000\001\234\002\205\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\001\202\002\n\002\185\002\205\000\000\000\000\000\000\000\000\002\205\000\000\000\000\002\205\002\205\002\185\002\185\002\205\000\000\002\205\000\000\002\185\002j\002\185\000\000\000\000\002\185\000\000\002\185\000\000\000\000\000\000\000\000\002\185\001\210\001\226\000\000\000\000\001\234\002\201\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\001\202\002\n\002r\002\201\000\000\000\000\000\000\000\000\002\201\000\000\000\000\002\201\002\201\002\205\002\205\002\201\000\000\002\201\000\000\002\205\002j\002\205\000\000\000\000\002\130\000\000\002\205\000\000\000\000\000\000\000\000\002\205\001\210\001\226\000\000\000\000\001\234\002\209\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\001\202\002\n\002r\002\209\000\000\000\000\000\000\000\000\002\209\000\000\000\000\002\209\002\209\002\201\002\201\002\209\000\000\002\209\000\000\002\201\002j\002\201\000\000\000\000\002\201\000\000\002\201\000\000\000\000\000\000\000\000\002\201\001\210\001\226\000\000\000\000\001\234\002\213\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\001\202\002\n\002r\002\213\000\000\000\000\000\000\000\000\002\213\000\000\000\000\002\213\002\213\002\209\002\209\002\213\000\000\002\213\000\000\002z\002j\002\209\000\000\000\000\002\130\000\000\002\209\000\000\000\000\000\000\000\000\002\209\001\210\001\226\000\000\000\000\001\234\000\000\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\003\254\004\006\002r\000\177\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\002\213\002\213\004\014\000\000\004.\000\000\002z\0046\002\138\000\000\000\000\002\130\000\000\002\213\000\000\000\000\000\000\000\000\002\213\004>\004F\000\000\000\000\004N\000\000\000\000\000\177\000\000\000\000\004V\004^\000\000\000\000\003\005\003\005\004f\003\005\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\177\000\177\003\005\000\000\003\005\000\000\004n\003\005\004v\000\000\000\000\004~\000\000\004\134\000\000\000\000\000\000\000\000\000\000\001\210\001\226\000\000\000\000\001\234\000\000\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\003\001\003\001\003\005\003\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\005\003\005\003\001\000\000\003\001\000\000\003\005\003\001\003\005\000\000\000\000\003\005\000\000\003\005\000\000\000\000\000\000\000\000\000\000\001\210\001\226\000\000\000\000\001\234\000\000\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\001\202\002\n\003\001\003\025\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\001\003\001\003\025\000\000\003\025\000\000\003\001\002j\003\001\000\000\000\000\003\001\000\000\003\001\000\000\000\000\000\000\000\000\000\000\001\210\001\226\000\000\000\000\001\234\000\000\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\002\253\002\253\002r\002\253\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\025\003\025\002\253\000\000\002\253\000\000\002z\002\253\002\138\000\000\000\000\002\130\000\000\002\146\000\000\000\000\000\000\000\000\000\000\001\210\001\226\000\000\000\000\001\234\000\000\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\002\249\002\249\002\253\002\249\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\002\253\002\253\002\249\000\000\002\249\000\000\002\253\002\249\002\253\000\000\000\000\002\253\000\000\002\253\000\000\000\000\000\000\000\000\000\000\001\210\001\226\000\000\000\000\001\234\000\000\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\001\202\002\n\002\249\003\r\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\002\249\002\249\003\r\000\000\003\r\000\000\002\249\002j\002\249\000\000\000\000\002\249\000\000\002\249\000\000\000\000\000\000\000\000\000\000\001\210\001\226\000\000\000\000\001\234\000\000\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\001\202\002\n\002r\003\017\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\r\003\r\003\017\000\000\003\017\000\000\003\r\002j\003\r\000\000\000\000\002\130\000\000\003\r\000\000\000\000\000\000\000\000\000\000\001\210\001\226\000\000\000\000\001\234\000\000\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\001\202\002\n\002r\003\t\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\017\003\017\003\t\000\000\003\t\000\000\002z\002j\003\017\000\000\000\000\002\130\000\000\003\017\000\000\000\000\000\000\000\000\000\000\001\210\001\226\000\000\000\000\001\234\000\000\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\001\202\002\n\002r\003\021\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\t\003\t\003\021\000\000\003\021\000\000\003\t\002j\003\t\000\000\000\000\003\t\000\000\003\t\000\000\000\000\000\000\000\000\000\000\001\210\001\226\000\000\000\000\001\234\000\185\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\005z\005\130\002r\000\185\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\021\003\021\005\138\000\000\005\154\000\000\002z\005\162\002\138\000\000\000\000\002\130\000\000\003\021\000\000\000\000\000\000\000\000\000\000\005\170\005\178\000\000\000\000\005\186\003A\000\000\005\194\000\000\000\000\005\202\005\210\000\000\000\000\003A\003A\005\218\003A\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\185\000\000\003A\000\000\003A\000\000\005\226\003A\005\234\000\000\000\000\005\242\000\000\005\250\000\000\000\000\000\000\000\000\000\000\001\210\001\226\000\000\000\000\001\234\003=\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\003=\003=\003A\003=\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\003A\000\000\003=\000\000\003=\000\000\003A\003=\003A\000\000\000\000\003A\000\000\003A\000\000\000\000\000\000\000\000\000\000\001\210\001\226\000\000\000\000\001\234\003U\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\001\202\002\n\003=\003U\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\003=\000\000\003U\000\000\003U\000\000\003=\002j\003=\000\000\000\000\003=\000\000\003=\000\000\000\000\000\000\000\000\000\000\001\210\001\226\000\000\000\000\001\234\0039\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\0039\0039\002r\0039\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\003U\000\000\0039\000\000\0039\000\000\002z\0039\002\138\000\000\000\000\002\130\000\000\002\146\000\000\000\000\000\000\000\000\000\000\001\210\001\226\000\000\000\000\001\234\0035\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\0035\0035\0039\0035\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\0039\000\000\0035\000\000\0035\000\000\0039\0035\0039\000\000\000\000\0039\000\000\0039\000\000\000\000\000\000\000\000\000\000\001\210\001\226\000\000\000\000\001\234\003I\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\001\202\002\n\0035\003I\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\0035\000\000\003I\000\000\003I\000\000\0035\002j\0035\000\000\000\000\0035\000\000\0035\000\000\000\000\000\000\000\000\000\000\001\210\001\226\000\000\000\000\001\234\003M\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\001\202\002\n\002r\003M\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\003I\000\000\003M\000\000\003M\000\000\003I\002j\003I\000\000\000\000\002\130\000\000\003I\000\000\000\000\000\000\000\000\000\000\001\210\001\226\000\000\000\000\001\234\003E\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\001\202\002\n\002r\003E\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\003M\000\000\003E\000\000\003E\000\000\002z\002j\003M\000\000\000\000\002\130\000\000\003M\000\000\000\000\000\000\000\000\000\000\001\210\001\226\000\000\000\000\001\234\003Q\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\001\202\002\n\002r\003Q\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\003E\000\000\003Q\000\000\003Q\000\000\003E\002j\003E\000\000\000\000\003E\000\000\003E\000\000\000\000\000\000\000\000\000\000\001\210\001\226\000\000\000\000\001\234\000\000\000\000\001\242\000\000\000\000\001\250\002\002\000\000\000\000\000\000\000\000\002r\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\003Q\000\000\000\000\000\000\000\000\000\000\002z\000\000\002\138\000\000\000\000\002\130\000\000\003Q"))
  
  and lhs =
    (8, "\001\000lkkjiiiihhggffeeeeeeeeeeeedccbaaaa````_^]]\\\\[[ZYXWVUTSRRQPPPONNMLKKJJIIHGFFEEDCBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBAA@?>=<<;;::998877665544333322221100/.-,,++**))((''&&%%$$################\"\"\"\"\"\"\"\"\"\"\"\"\"\"\"!!!!!!!!!!!!!!!!  \031\031\031\031\031\031\031\031\031\031\031\031\031\031\031\031\031\031\031\031\031\030\030\030\030\030\030\030\030\030\030\030\030\030\030\030\030\030\030\030\030\030\029\029\029\028\028\028\028\028\028\028\028\028\027\026\026\026\025\024\023\023\022\022\021\021\020\020\019\019\019\018\018\017\016\015\015\014\014\014\014\014\014\014\014\r\r\r\r\r\r\r\r\r\r\012\011\n\t\b\007\007\006\005\004\003\002")
  
  and goto =
    ((16, "\000\000\000\000\000*\000\000\001\156\000\000\017\2080Z0\240\000\000\000\000\000\000\000\0001(\000\000\000\0001l\026\136\018\"1\164\000\011\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\023\184\r\146\000\0001\2322 \000\000\000\028\000\000\000\000\000\000\000\234\000\000\000:\000\000\000\000\000\000\000\161\000\005\000\000\018\168\000\0002d2\1562\2243\024\000\000\000\000\000\000\000\000\000\015\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\155\000\003\000\000\r\228\024\n\000\000\000\0153\\\000\0153\148\000\0153\216\000\015\000\000\000\000\000\000\000\000\000C\000$\000\000\018\250\000\000\0310\000\000\031\198\000\000\000\000\000\000\031\254\000\000 \148\000\000 \204\000\000!b\000\000!\154\000\000\"0\000\000\024\144\000\019\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\024\226\000\000\000\000\000\000\025h\000\000\"h\000\000\"\254\000\000#6\000\000#\204\000\000$\004\000\000$\154\000\000$\210\000\000\000\000\000\000\025\186\000\000\000\000\000\0004\016\000\0154T\000\0154\140\000\0154\208\000\0155\b\000\015\000\000\000\000\000\000\000\000\000\000\000\000\014j\000\000\019\128\000\000\000\000\000\000\000\015\000\015\000\015\000\000\005\1725L\000\015\000<\0003\026@\000\000\000\000\000^\000\000\000\000\000\0005\132\000\015\000\127\002l\019\210\000\000\000\000\020X\000\000\000\000\000k\020\170\000\000\003<\000\000\014\188\000\000\000\0005\200\000\0156\000\000\015\029\198\000\000\000\026\000\0006D\000\0156|\000\0156\192\000\015\000\000\0210\000\000\006|\000\000\000\000\015B\0004\000)\026\198\000\000\000\015\000\000%h\000\000%\160\000\000\027\030\000.\027\154\000\000\000\000\000\000\027\210\000\000&6\000\000&n\000\000'\004\000\000'<\000\000'\210\000\000(\n\000\000(\160\000\000(\216\000\000)n\000\000)\166\000\000*<\000\000*t\000\000\000\000\000\000\021\130\000\000\007L\b\0286\248\000\0157<\000\0157t\000\015\000\000\0000\000\000\000\000\000\000\000\148\000\0007\184\000\015\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000|\000\000\000\000\000\000\000\000\000\000\022\b\000\000\000\000\000\000\000\000\000\2267\240\000\01584\000\0158l\000\0158\176\000\0158\232\000\0159,\000\0159d\000\0159\168\000\015\000\000\000\000+\n\000\000+B\000\000\028p\000\000\028\168\000\000+\216\000\000,\016\000\000,\166\000\000,\222\000\000-t\000\000-\172\000\000.B\000\000.z\000\000/\016\000\000/H\000\000/\222\000\0000\022\000\0000\172\000\000\000\000\000P\000\174\000\000\022Z\000\000\000\000\000\000\000j\029J\000\000\000\000\000\000\b\236\000\000\000\000\000\000\000\000\000\000\000\000\029\130\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000b\000\180\000\000\022\224\000\000\000\000\000\000\000\000\000\000\000\000\000\148\000\158\000\000\000\000\000\000\000\000\015\154\000\000\016 \000\000\t\188\000\000\000\000\016r\000\000\016\248\000\000\n\140\000\000\000.\0232\000\000\011\\\000\000\000\000\030\030\000\000\000\000\000\000\012,\000\000\004\012\000\000\000\000\0009\000\000\004\220\000\000\000\000\000\000\000<\000\000\000B\000\000\000\000\000\000\000\000\000~\000/\000\000\000\000\000\174\000\000\000f\000\000\000\000\000\144\000\000\000\000\000\000\001\014\000\000\000\254\000\000\000\142\000\000\000\000\000\204\000\000\000\000\000\000\000\000\000\000\000\015\000\015\000\000\030V\000\000\000\000\000\000\000\000\000(\000\000\000\000\030\248\000\000\000\000\000\000\000\000\000\000\000\000\000p\000\000\000\000\000\000\000\000\000\000\000\015\000\000\000\000\000\000\000l\000\000\000\015\000\015\000\015\000\015\000\000\012\252\000\000\000\000\000\164\000\000\000\000\000\000\000\000\017J\000\000\000\000\000\000"), (16, "\0019\001:\000Q\001;\001\201\001<\001=\001>\001\239\001\205\001?\001@\001A\001\233\0005\000\206\001\234\000\003\000?\001%\000@\0007\001J\000R\001\213\000X\001K\002\024\001L\000\194\000?\001^\000\251\000?\000?\000C\000\203\001\214\000\003\000\209\002\001\002\004\001\209\001\170\000Z\000[\001\129\001\172\001\130\000E\000A\000?\000\210\000\203\001\237\002\025\001\207\001\137\000\005\001\140\001\141\002\005\0013\001\144\001\213\000_\001\236\001\145\001\146\000\208\001\147\001\148\002\006\000\172\001\151\002\003\000\004\001\216\000?\000\005\001\168\001\152\0007\001\153\000p\001\154\0014\001\208\001\155\001\156\001\210\000?\000\147\001\219\001\157\001\171\001\164\000w\001\249\001\251\001\165\002\001\002\004\000x\0019\001:\000Q\001;\002\022\001<\001=\001>\0016\001\224\001?\001@\001A\001\233\001\169\000\236\001\234\000?\002\005\000C\001\006\001\209\001J\000R\001\209\002\022\001K\000?\001L\000C\001\222\001^\0014\000F\001\213\000?\002\023\001\228\001\170\001\225\001D\001]\001\136\000F\000Z\000[\001\129\001\221\001\130\001\132\001\138\001\163\001\159\001\223\001\235\001\232\002\r\001\137\000\000\001\140\001\141\000\000\000\000\001\144\000\000\000G\001\236\001\145\001\146\000\000\001\147\001\148\000\000\000\000\001\151\001\230\000\000\000\000\000\000\000\000\000\000\001\152\000\000\001\153\000\000\001\154\002\022\000\000\001\155\001\156\000\000\000\000\000\000\000\000\001\157\000\000\001\164\000w\000\000\000\000\001\165\000\000\000\000\000x\0019\001:\000Q\001;\000\000\001<\001=\001>\000\000\000\000\001?\001@\001A\001\233\000\000\000\000\002\021\000\000\000\000\000\000\000\000\000\000\001J\000R\000\000\000\000\001K\000\000\001L\000\000\000\000\001^\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Z\000[\001\129\000\000\001\130\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001\137\000\000\001\140\001\141\000\000\000\000\001\144\000\000\000\000\001\236\001\145\001\146\000\000\001\147\001\148\000\000\000\000\001\151\000\000\000\000\000\000\000\000\000\000\000\000\001\152\000\000\001\153\000\000\001\154\000\000\000\000\001\155\001\156\000\000\000\000\000\000\000\000\001\157\000\000\001\164\000w\000\000\000\000\001\165\000\000\000\000\000x\0019\001:\000Q\001;\000\000\001<\001=\001>\000\000\000\000\001?\001@\001A\001\198\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001J\000R\000\000\000\000\001K\000\000\001L\000\000\000\000\001^\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Z\000[\001\129\000\000\001\130\000\000\000\000\000\000\001\211\000\000\000\000\000\000\000\000\001\137\000\000\001\140\001\141\000\000\000\000\001\144\000\000\000\000\000\000\001\145\001\146\000\000\001\147\001\148\000\000\000\000\001\151\000\000\000\000\000\000\000\000\000\000\000\000\001\152\000\000\001\153\000\000\001\154\000\000\000\000\001\155\001\156\000\000\000\000\000\000\000\000\001\157\000\000\001\164\000w\000\000\000\000\001\165\000\000\000\000\000x\0019\001:\000Q\001;\000\000\001<\001=\001>\000\000\000\000\001?\001@\001A\001\198\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001J\000R\000\000\000\000\001K\000\000\001L\000\000\000\000\001^\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Z\000[\001\129\000\000\001\130\000\000\000\000\000\000\001\200\000\000\000\000\000\000\000\000\001\137\000\000\001\140\001\141\000\000\000\000\001\144\000\000\000\000\000\000\001\145\001\146\000\000\001\147\001\148\000\000\000\000\001\151\000\000\000\000\000\000\000\000\000\000\000\000\001\152\000\000\001\153\000\000\001\154\000\000\000\000\001\155\001\156\000\000\000\000\000\000\000\000\001\157\000\000\001\164\000w\000\000\000\000\001\165\000\000\000\000\000x\0019\001:\000Q\001;\000\000\001<\001=\001>\000\000\000\000\001?\001@\001A\001\198\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001J\000R\000\000\000\000\001K\000\000\001L\000\000\000\000\001^\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Z\000[\001\129\000\000\001\130\000\000\000\000\000\000\001\199\000\000\000\000\000\000\000\000\001\137\000\000\001\140\001\141\000\000\000\000\001\144\000\000\000\000\000\000\001\145\001\146\000\000\001\147\001\148\000\000\000\000\001\151\000\000\000\000\000\000\000\000\000\000\000\000\001\152\000\000\001\153\000\000\001\154\000\000\000\000\001\155\001\156\000\000\000\000\000\000\000\000\001\157\000\000\001\164\000w\000\000\000\000\001\165\000\000\000\000\000x\0019\001:\000Q\001;\000\000\001<\001=\001>\000\000\000\000\001?\001@\001A\001\198\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001J\000R\000\000\000\000\001K\000\000\001L\000\000\000\000\001^\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Z\000[\001\129\000\000\001\130\000\000\000\000\000\000\001\204\000\000\000\000\000\000\000\000\001\137\000\000\001\140\001\141\000\000\000\000\001\144\000\000\000\000\000\000\001\145\001\146\000\000\001\147\001\148\000\000\000\000\001\151\000\000\000\000\000\000\000\000\000\000\000\000\001\152\000\000\001\153\000\000\001\154\000\000\000\000\001\155\001\156\000\000\000\000\000\000\000\000\001\157\000\000\001\164\000w\000\000\000\000\001\165\000\000\000\000\000x\0019\001:\000Q\001;\000\000\001<\001=\001>\000\000\000\000\001?\001@\001A\001\226\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001J\000R\000\000\000\000\001K\000\000\001L\000\000\000\000\001^\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Z\000[\001\129\000\000\001\130\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001\137\000\000\001\140\001\141\000\000\000\000\001\144\000\000\000\000\000\000\001\145\001\146\000\000\001\147\001\148\000\000\000\000\001\151\000\000\000\000\000\000\000\000\000\000\000\000\001\152\000\000\001\153\000\000\001\154\000\000\000\000\001\155\001\156\000\000\000\000\000\000\000\000\001\157\000\000\001\164\000w\000\000\000\000\001\165\000\000\000\000\000x\0019\001:\000Q\001;\000\000\001<\001=\001>\000\000\000\000\001?\001@\001A\001\195\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001J\000R\000\000\000\000\001K\000\000\001L\000\000\000\000\001^\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Z\000[\001\129\000\000\001\130\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001\137\000\000\001\140\001\141\000\000\000\000\001\144\000\000\000\000\000\000\001\145\001\146\000\000\001\147\001\148\000\000\000\000\001\151\000\000\000\000\000\000\000\000\000\000\000\000\001\152\000\000\001\153\000\000\001\154\000\000\000\000\001\155\001\156\000\000\000\000\000\000\000\000\001\157\000\000\001\164\000w\000\000\000\000\001\165\000\000\000\000\000x\0019\001:\000Q\001;\000\000\001<\001=\001>\000\000\000\000\001?\001@\001A\001\166\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001J\000R\000\000\000\000\001K\000\000\001L\000\000\000\000\001^\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Z\000[\001\129\000\000\001\130\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001\137\000\000\001\140\001\141\000\000\000\000\001\144\000\000\000\000\000\000\001\145\001\146\000\000\001\147\001\148\000\000\000\000\001\151\000\000\000\000\000\000\000\000\000\000\000\000\001\152\000\000\001\153\000\000\001\154\000\000\000\000\001\155\001\156\000\000\000\000\000\000\000\000\001\157\000\000\001\164\000w\000\000\000\000\001\165\000\000\000\000\000x\0019\001:\000Q\001;\000\000\001<\001=\001>\000\000\000\000\001?\001@\001A\001E\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001J\000R\000\000\000\000\001K\000\000\001L\000\000\000\000\001^\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Z\000[\001\129\000\000\001\130\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001\137\000\000\001\140\001\141\000\000\000\000\001\144\000\000\000\000\000\000\001\145\001\146\000\000\001\147\001\148\000\000\000\000\001\151\000\000\000\000\000\000\000\000\000\000\000\000\001\152\000\000\001\153\000\000\001\154\000\000\000\000\001\155\001\156\000\000\000\000\000\000\000\000\001\157\000\000\001\164\000w\000\000\000\000\001\165\000\000\000\000\000x\0019\001:\000Q\001;\000\000\001<\001=\001>\000\000\000\000\001?\001@\001A\001\143\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001J\000R\000\000\000\000\001K\000\000\001L\000\000\000\000\001^\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Z\000[\001\129\000\000\001\130\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001\137\000\000\001\140\001\141\000\000\000\000\001\144\000\000\000\000\000\000\001\145\001\146\000\000\001\147\001\148\000\000\000\000\001\151\000\000\000\000\000\000\000\000\000\000\000\000\001\152\000\000\001\153\000\000\001\154\000\000\000\000\001\155\001\156\000\000\000\000\000\000\000\000\001\157\000\000\001\164\000w\000\000\000\000\001\165\000\000\000\000\000x\0019\001:\000Q\001;\000\000\001<\001=\001>\000\000\000\000\001?\001@\001A\001\178\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001J\000R\000\000\000\000\001K\000\000\001L\000\000\000\000\001^\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Z\000[\001\129\000\000\001\130\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001\137\000\000\001\140\001\141\000\000\000\000\001\144\000\000\000\000\000\000\001\145\001\146\000\000\001\147\001\148\000\000\000\000\001\151\000\000\000\000\000\000\000\000\000\000\000\000\001\152\000\000\001\153\000\000\001\154\000\000\000\000\001\155\001\156\000\000\000\000\000\000\000\000\001\157\000\000\001\164\000w\000\000\000\000\001\165\000\000\000\000\000x\0019\001:\000Q\001;\000\000\001<\001=\001>\000\000\000\000\001?\001@\001A\001\185\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001J\000R\000\000\000\000\001K\000\000\001L\000\000\000\000\001^\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Z\000[\001\129\000\000\001\130\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001\137\000\000\001\140\001\141\000\000\000\000\001\144\000\000\000\000\000\000\001\145\001\146\000\000\001\147\001\148\000\000\000\000\001\151\000\000\000\000\000\000\000\000\000\000\000\000\001\152\000\000\001\153\000\000\001\154\000\000\000\000\001\155\001\156\000\000\000\000\000\000\000\000\001\157\000\000\001\164\000w\000\000\000\000\001\165\000\000\000\000\000x\0019\001:\000Q\001;\000\000\001<\001=\001>\000\000\000\000\001?\001@\001A\001\190\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001J\000R\000\000\000\000\001K\000\000\001L\000\000\000\000\001^\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Z\000[\001\129\000\000\001\130\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001\137\000\000\001\140\001\141\000\000\000\000\001\144\000\000\000\000\000\000\001\145\001\146\000\000\001\147\001\148\000\000\000\000\001\151\000\000\000\000\000\000\000\000\000\000\000\000\001\152\000\000\001\153\000\000\001\154\000\000\000\000\001\155\001\156\000\000\000\000\000\000\000\000\001\157\000\000\001\164\000w\000\000\000\000\001\165\000\000\000\000\000x\0019\001:\000Q\001;\000\000\001<\001=\001>\000\000\000\000\001?\001@\001A\001\197\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001J\000R\000\000\000\000\001K\000\000\001L\000\000\000\000\001^\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Z\000[\001\129\000\000\001\130\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001\137\000\000\001\140\001\141\000\000\000\000\001\144\000\000\000\000\000\000\001\145\001\146\000\000\001\147\001\148\000\000\000\000\001\151\000\000\000\000\000\000\000\000\000\000\000\000\001\152\000\000\001\153\000\000\001\154\000\000\000\000\001\155\001\156\000\000\000\000\000\000\000\000\001\157\000\000\001\164\000w\000\000\000\000\001\165\000\000\000\000\000x\0019\001:\000Q\001;\000\000\001<\001=\001>\000\000\000\000\001?\001@\001A\002\020\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001J\000R\000\000\000\000\001K\000\000\001L\000\000\000\000\001^\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Z\000[\001\129\000\000\001\130\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001\137\000\000\001\140\001\141\000\000\000\000\001\144\000\000\000\000\000\000\001\145\001\146\000\000\001\147\001\148\000\000\000\000\001\151\000\000\000\000\000Q\000\000\000\000\000\000\001\152\000\000\001\153\000b\001\154\000\000\000\000\001\155\001\156\000\000\000\000\000\000\000\000\001\157\000\000\001\164\000w\000R\000\000\001\165\000S\000T\000x\000c\000\000\000\000\000\000\000r\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000Q\000Z\000[\000\000\000\\\000\000\000]\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\134\000\000\000\000\000\000\000\184\000\000\000R\000k\000\000\000S\000T\000\000\000c\000\000\000\000\000\000\000r\000\000\001\243\001\245\001\248\001\253\000\000\000\000\000\000\000Y\000l\000Z\000[\000m\000\\\000\000\000]\000\148\000\185\000\000\000\000\000n\000\000\000w\000\000\000\134\000\000\000\000\001\255\000x\000\000\000\000\000k\000\000\000\000\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\190\000\000\000R\000\148\000\000\000S\000T\000n\000c\000w\000\000\000\000\000r\000\000\000\187\000x\000\000\000\188\000\000\000\000\000\000\000Y\000Q\000Z\000[\000\000\000\\\000\000\000]\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\134\000\000\000\000\000\000\000\000\000\000\000R\000k\000\000\000S\000T\000\000\000c\000\000\000\000\000\000\000r\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\227\000Y\000l\000Z\000[\000m\000\\\000\000\000]\000\148\000\000\000\000\000\000\000n\000\000\000w\000\000\000\134\000\000\000\000\000\187\000x\000\000\000\188\000k\000\000\000\000\000\000\000Q\000\000\000\000\000\000\000\228\000\000\000\253\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000R\000\148\000\000\000S\000T\000n\000\254\000w\000\000\000\255\000\000\000\000\000\171\000x\000\000\000\000\000\000\001\179\000\000\000Y\000\000\000Z\000[\000Q\000\\\000\000\000]\000\000\000\000\000\000\000b\000\000\000\000\000\000\000\000\001\186\000\000\000\000\000\000\000\000\000\000\000\000\000k\000\000\000R\000\000\000\000\000S\000T\001\191\000c\000\000\000\000\000\000\000r\000\000\000\000\000\000\000\000\000\000\000\000\000l\001\174\000Y\000m\000Z\000[\001\007\000\\\000\000\000]\000\000\000n\000\000\000w\000\000\000\000\000\000\001\194\000\134\000x\000\000\000\000\000\000\000\000\000\000\000k\000\000\000\000\000\000\000Q\000\000\000\000\000\000\000\228\000\000\000\000\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000R\000\148\000\000\000S\000T\000n\000c\000w\000\000\000\000\000r\000\000\000\171\000x\000\000\000\000\000\000\000\000\001\176\000Y\000Q\000Z\000[\000\000\000\\\000\000\000]\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\134\000\000\000\000\000\000\000\000\000\000\000R\000k\000\000\000S\000T\000\000\000c\000\000\000\000\000\228\000r\000\000\000\000\000\000\000\000\000\000\000\000\000\000\001\181\000Y\000l\000Z\000[\000m\000\\\000\000\000]\000\148\000\000\000\000\000\000\000n\000\000\000w\000\000\000\134\000\000\000\000\000\171\000x\000\000\000\000\000k\000\000\000\000\000\000\000Q\000\000\000\000\000\000\000\228\000\000\000\000\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000R\000\148\000\000\000S\000T\000n\000c\000w\000\000\000\000\000r\000\000\000\171\000x\000\000\000\000\000\000\000\000\001\183\000Y\000Q\000Z\000[\000\000\000\\\000\000\000]\000b\000\000\000\000\000\000\000\000\002\028\000\000\000\000\000\134\000\000\000\000\000\000\000\000\000\000\000R\000k\000\000\000S\000T\000\000\000c\000\000\000\000\000\228\000r\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000l\000Z\000[\000m\000\\\000\000\000]\000\148\000\000\000\000\000\000\000n\000\000\000w\000\000\000\134\000\000\000\000\000\171\000x\000\000\000\000\000k\000\000\000\000\000\000\000Q\000\000\000\000\000\000\002\029\000\000\000\000\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000R\000\148\000\000\000S\000T\000n\000c\000w\000\000\000\000\000r\000\000\000\171\000x\000\000\000\000\000\000\000\000\000\000\000Y\000Q\000Z\000[\000\000\000\\\000\000\000]\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\134\000\000\000\000\000\000\000\000\000\000\000R\000k\000\000\000S\000T\000\000\000c\000\000\000\000\002\018\000r\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000l\000Z\000[\000m\000\\\000\000\000]\000\148\000\000\000\000\000\000\000n\000\000\000w\000\000\000\134\000\000\000\000\000\171\000x\000\000\000\000\000k\000\000\000\000\000\000\000Q\000\000\000\000\000\000\002\t\000\000\000\000\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000R\000\148\000\000\000S\000T\000n\000c\000w\000\000\000\000\000r\000\000\000\171\000x\000\000\000\000\000\000\000\000\000\000\000Y\000Q\000Z\000[\000\000\000\\\000\000\000]\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\134\000\000\000\000\000\000\000\000\000\000\000R\000k\000\000\000S\000T\000\000\000c\000\000\000\000\000\198\000r\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000l\000Z\000[\000m\000\\\000\000\000]\000\148\000\000\000\000\000\000\000n\000\000\000w\000\000\000\134\000\000\000\000\000\171\000x\000\000\000\000\000k\000\000\000\000\000\000\000Q\000\000\000\000\000\000\000\167\000\000\000\000\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000R\000\148\000\000\000S\000T\000n\000c\000w\000\000\000\000\000r\000\000\000\171\000x\000\000\000\000\000\000\000\000\000\000\000Y\000Q\000Z\000[\000\000\000\\\000\000\000]\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\134\000\000\000\000\000\000\000\000\000\000\000R\000k\000\000\000S\000T\000\000\000c\000\000\000\000\000\192\000r\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000l\000Z\000[\000m\000\\\000\000\000]\000\148\000\000\000\000\000\000\000n\000\000\000w\000\000\000\134\000\000\000\000\000\171\000x\000\000\000\000\000k\000\000\000\000\000\000\000Q\000\000\000\000\000\000\000\216\000\000\000\000\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000R\000\148\000\000\000S\000T\000n\000c\000w\000\000\000\000\000r\000\000\000\171\000x\000\000\000\000\000\000\000\000\000\000\000Y\000Q\000Z\000[\000\000\000\\\000\000\000]\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\134\000\000\000\000\000\000\000\000\000\000\000R\000k\000\000\000S\000T\000\000\000c\000\000\000\000\000\219\000r\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000l\000Z\000[\000m\000\\\000\000\000]\000\148\000\000\000\000\000\000\000n\000\000\000w\000\000\000\134\000\000\000\000\000\171\000x\000\000\000\000\000k\000\000\000\000\000\000\000Q\000\000\000\000\000\000\000\223\000\000\000\000\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000R\000\148\000\000\000S\000T\000n\000c\000w\000\000\000\000\000r\000\000\000\171\000x\000\000\000\000\000\000\000\000\000\000\000Y\000Q\000Z\000[\000\000\000\\\000\000\000]\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\134\000\000\000\000\000\000\000\000\000\000\000R\000k\000\000\000S\000T\000\000\000c\000\000\000\000\000\245\000r\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000l\000Z\000[\000m\000\\\000\000\000]\000\148\000\000\000\000\000\000\000n\000\000\000w\000\000\000\134\000\000\000\000\000\171\000x\000\000\000\000\000k\000\000\000\000\000\000\000Q\000\000\000\000\000\000\001'\000\000\000\000\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000R\000\148\000\000\000S\000T\000n\000c\000w\000\000\000\000\000r\000\000\000\171\000x\000\000\000\000\000\000\000\000\000\000\000Y\000Q\000Z\000[\000\000\000\\\000\000\000]\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\134\000\000\000\000\000\000\000\000\000\000\000R\000k\000\000\000S\000T\000\000\000c\000\000\000\000\001H\000r\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000l\000Z\000[\000m\000\\\000\000\000]\000\148\000\000\000\000\000\000\000n\000\000\000w\000\000\000\134\000\000\000\000\000\171\000x\000\000\000\000\000k\000\000\000\000\000\000\000Q\000\000\000\000\000\000\001\134\000\000\000\000\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000R\000\148\000\000\000S\000T\000n\000c\000w\000\000\000\000\000r\000\000\000\171\000x\000\000\000\000\000\000\000\000\000\000\000Y\000Q\000Z\000[\000\000\000\\\000\000\000]\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\134\000\000\000\000\000\000\000\000\000\000\000R\000k\000\000\000S\000T\000\000\000c\000\000\000\000\001\161\000r\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000l\000Z\000[\000m\000\\\000\000\000]\000\148\000\000\000\000\000\000\000n\000\000\000w\000\000\000\134\000\000\000\000\000\171\000x\000\000\000\000\000k\000\000\000\000\000\000\000Q\000\000\000\000\000\000\001\188\000\000\000\000\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000R\000\148\000\000\000S\000T\000n\000c\000w\000\000\000\000\000r\000\000\000\171\000x\000\000\000\000\000\000\000\000\000\000\000Y\000Q\000Z\000[\000\000\000\\\000\000\000]\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\134\000\000\000\000\000\000\000\000\000\000\000R\000k\000\000\000S\000T\000\000\000c\000\000\000\000\000\000\000r\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000l\000Z\000[\000m\000\\\000\000\000]\000\148\000\000\000\000\000\000\000n\000\000\000w\000\000\000\134\000\000\000\000\002\000\000x\000\000\000\000\000k\000\000\000\000\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000R\000\148\000\000\000S\000T\000n\000c\000w\000\000\000\000\000r\000\000\000\183\000x\000\000\000\000\000\000\000\000\000\000\000Y\000Q\000Z\000[\000\000\000\\\000\000\000]\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\134\000\000\000\000\000\000\000\000\000\000\000R\000k\000\000\000S\000T\000\000\000c\000\000\000\000\000\000\000r\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000l\000Z\000[\000m\000\\\000\000\000]\000\148\000\000\000\000\000\000\000n\000\000\000w\000\000\000\134\000\000\000\000\000\150\000x\000\000\000\000\000k\000\000\000\000\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000R\000\148\000\000\000S\000T\000n\000c\000w\000\000\000\000\000r\000\000\000\149\000x\000\000\000\000\000\000\000\000\000\000\000Y\000Q\000Z\000[\000\000\000\\\000\000\000]\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\134\000\000\000\000\000\000\000\000\000\000\000R\000k\000\000\000S\000T\000\000\000c\000\000\000\000\000\000\000r\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000l\000Z\000[\000m\000\\\000\000\000]\000\148\000\000\000\000\000\000\000n\000\000\000w\000\000\000\134\000\000\000\000\000\152\000x\000\000\000\000\000k\000\000\000\000\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000b\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000R\000\148\000\000\000S\000T\000n\000c\000w\000\000\000\000\000r\000\000\000\170\000x\000\000\000Q\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\134\000R\000\000\000\000\000S\000T\000\000\000k\000\000\000\000\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\253\000\000\000\000\000Y\000\000\000Z\000[\000\000\002\011\000l\002\012\000\000\000m\000\000\000\000\000R\000\148\000\000\000S\000T\000n\000\254\000w\000\000\000\255\000\000\000k\000\205\000x\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000Q\000\\\000\000\000]\000\000\000\000\000\253\000l\000\000\000\000\000m\000\000\001\005\000\000\000\000\000\000\000\000\000\000\000\000\000k\000w\000R\000\000\000\000\000S\000T\000x\000\254\000\000\000\000\000\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000Y\000m\000Z\000[\001\007\000\\\000\000\000]\000\000\000n\000\000\000w\000\000\000\000\000\000\001$\001\005\000x\000\000\000\000\000\000\000\000\000Q\000k\000\000\000\000\000\000\000\000\000\253\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000l\000R\000\000\000m\000S\000T\001\007\000\254\000Q\000\000\000\255\000n\000\000\000w\000\253\000\000\000\000\001\t\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000\254\001\005\000\000\000\255\000\000\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\001\005\000m\000\000\000\000\001\007\000\000\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\001\b\000\000\000x\000\000\000\000\000\000\000\000\000\000\000Q\000\000\000\000\000l\000\000\000\000\000m\000b\000\000\001\007\000\000\000\000\000\000\000\000\000n\000\000\000w\000\000\000\000\000\000\001\011\000R\000x\000\000\000S\000T\000\000\000c\000Q\000\000\000\000\000r\000\000\000\000\000\000\000b\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000\134\000\000\000\000\000r\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\134\000m\000\000\000\000\000\000\000\148\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\001d\000x\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Q\000l\000\000\000\000\000m\000\000\000\000\000b\000\148\000\000\000\000\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\001f\000x\000R\000\000\000\000\000S\000T\000\000\000c\000Q\000\000\000\000\000r\000\000\000\000\000\000\000b\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000\134\000\000\000\000\000r\000\000\000\000\000Q\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000R\000\134\000m\000S\000T\000\000\000\148\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\001\139\000x\000Y\000\000\000Z\000[\000Q\000\234\000\000\000\235\000l\000\000\000\253\000m\000\000\000\000\000\000\000\148\000\000\000\000\000\000\000n\000\000\000w\000\000\000k\000\000\000R\001\150\000x\000S\000T\000\000\000\254\000Q\000\000\000\255\000\000\000\000\000\000\000\000\000b\000\000\000\000\000l\000\000\000Y\000m\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000w\000S\000T\000\000\000c\001\005\000x\000\000\000r\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\134\000m\000\000\000\000\001\007\000\000\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\001\193\000\000\000x\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Q\000l\000\000\000\000\000m\000\000\000\000\000b\000\148\000\000\000\000\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\001\244\000x\000R\000\000\000\000\000S\000T\000\000\000c\000Q\000\000\000\000\000r\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000\134\000\000\000\000\000t\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000j\000m\000\000\000\000\000\000\000\148\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\001\252\000x\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000n\000\000\000w\000R\000\000\000\000\000S\000T\000x\000c\000Q\000\000\000\000\000v\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000j\000\000\000\000\000z\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000j\000m\000\000\000\000\000\000\000\000\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000n\000\000\000w\000R\000\000\000\000\000S\000T\000x\000c\000Q\000\000\000\000\000|\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000j\000\000\000\000\000~\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000j\000m\000\000\000\000\000\000\000\000\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000n\000\000\000w\000R\000\000\000\000\000S\000T\000x\000c\000Q\000\000\000\000\000\128\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000j\000\000\000\000\000\130\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000j\000m\000\000\000\000\000\000\000\000\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000n\000\000\000w\000R\000\000\000\000\000S\000T\000x\000c\000Q\000\000\000\000\000\132\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000j\000\000\000\000\000\154\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000j\000m\000\000\000\000\000\000\000\000\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000n\000\000\000w\000R\000\000\000\000\000S\000T\000x\000c\000Q\000\000\000\000\000\156\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000j\000\000\000\000\000\158\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000j\000m\000\000\000\000\000\000\000\000\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000n\000\000\000w\000R\000\000\000\000\000S\000T\000x\000c\000Q\000\000\000\000\000\160\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000j\000\000\000\000\000\162\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000j\000m\000\000\000\000\000\000\000\000\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000n\000\000\000w\000R\000\000\000\000\000S\000T\000x\000c\000Q\000\000\000\000\000\164\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000j\000\000\000\000\000\166\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000j\000m\000\000\000\000\000\000\000\000\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000n\000\000\000w\000R\000\000\000\000\000S\000T\000x\000c\000Q\000\000\000\000\001\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000j\000\000\000\000\001\003\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000j\000m\000\000\000\000\000\000\000\000\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000n\000\000\000w\000R\000\000\000\000\000S\000T\000x\000c\000Q\000\000\000\000\001\r\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000j\000\000\000\000\001\015\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000j\000m\000\000\000\000\000\000\000\000\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000n\000\000\000w\000R\000\000\000\000\000S\000T\000x\000c\000Q\000\000\000\000\001\017\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000j\000\000\000\000\001\019\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000j\000m\000\000\000\000\000\000\000\000\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000n\000\000\000w\000R\000\000\000\000\000S\000T\000x\000c\000Q\000\000\000\000\001\021\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000j\000\000\000\000\001\023\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000j\000m\000\000\000\000\000\000\000\000\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000n\000\000\000w\000R\000\000\000\000\000S\000T\000x\000c\000Q\000\000\000\000\001\025\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000j\000\000\000\000\001\027\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000j\000m\000\000\000\000\000\000\000\000\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000n\000\000\000w\000R\000\000\000\000\000S\000T\000x\000c\000Q\000\000\000\000\001\029\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000j\000\000\000\000\001\031\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000j\000m\000\000\000\000\000\000\000\000\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000n\000\000\000w\000R\000\000\000\000\000S\000T\000x\000c\000Q\000\000\000\000\001!\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000j\000\000\000\000\001#\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000j\000m\000\000\000\000\000\000\000\000\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000n\000\000\000w\000R\000\000\000\000\000S\000T\000x\000c\000Q\000\000\000\000\001`\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000j\000\000\000\000\001b\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000j\000m\000\000\000\000\000\000\000\000\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000n\000\000\000w\000R\000\000\000\000\000S\000T\000x\000c\000Q\000\000\000\000\001h\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000j\000\000\000\000\001j\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000j\000m\000\000\000\000\000\000\000\000\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000n\000\000\000w\000R\000\000\000\000\000S\000T\000x\000c\000Q\000\000\000\000\001l\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000j\000\000\000\000\001n\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000j\000m\000\000\000\000\000\000\000\000\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000n\000\000\000w\000R\000\000\000\000\000S\000T\000x\000c\000Q\000\000\000\000\001p\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000j\000\000\000\000\001r\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000j\000m\000\000\000\000\000\000\000\000\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000n\000\000\000w\000R\000\000\000\000\000S\000T\000x\000c\000Q\000\000\000\000\001t\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000j\000\000\000\000\001v\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000j\000m\000\000\000\000\000\000\000\000\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000n\000\000\000w\000R\000\000\000\000\000S\000T\000x\000c\000Q\000\000\000\000\001x\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000j\000\000\000\000\001z\000\000\000\000\000\000\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000j\000m\000\000\000\000\000\000\000\000\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000n\000\000\000w\000R\000\000\000\000\000S\000T\000x\000c\000Q\000\000\000\000\001|\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000R\000\000\000\000\000S\000T\000\000\000c\000j\000\000\000\000\001~\000\000\000\000\000Q\000k\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000R\000j\000m\000S\000T\000\000\002\017\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000Y\000Q\000Z\000[\000\000\000\\\000\000\000]\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000j\000\000\000\000\000n\000\000\000w\000R\000k\000\000\000S\000T\000x\000c\000\000\000\000\000\000\001\128\000\000\000\000\000Q\000\000\000\000\000\000\000\000\000\000\000Y\000l\000Z\000[\000m\000\\\000\000\000]\000\000\000\000\000\000\000\000\000n\000\000\000w\000R\000j\000\000\000S\000T\000x\002\016\000Q\000k\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000l\000R\000\000\000m\000S\000T\000\000\002\015\000j\000\000\000\000\000n\000\000\000w\000Q\000k\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000R\000j\000m\000S\000T\000\000\002\014\000Q\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000l\000R\000\000\000m\000S\000T\000\000\002\b\000j\000\000\000\000\000n\000\000\000w\000Q\000k\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000R\000j\000m\000S\000T\000\000\001\241\000Q\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000l\000R\000\000\000m\000S\000T\000\000\001\240\000j\000\000\000\000\000n\000\000\000w\000Q\000k\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000R\000j\000m\000S\000T\000\000\000\197\000Q\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000l\000R\000\000\000m\000S\000T\000\000\000\196\000j\000\000\000\000\000n\000\000\000w\000Q\000k\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000R\000j\000m\000S\000T\000\000\000\195\000Q\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000l\000R\000\000\000m\000S\000T\000\000\000U\000j\000\000\000\000\000n\000\000\000w\000Q\000k\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000R\000j\000m\000S\000T\000\000\000e\000Q\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000l\000R\000\000\000m\000S\000T\000\000\000g\000j\000\000\000\000\000n\000\000\000w\000Q\000k\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000R\000j\000m\000S\000T\000\000\000i\000Q\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000l\000R\000\000\000m\000S\000T\000\000\000\174\000j\000\000\000\000\000n\000\000\000w\000Q\000k\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000R\000j\000m\000S\000T\000\000\000\176\000Q\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000l\000R\000\000\000m\000S\000T\000\000\000\178\000j\000\000\000\000\000n\000\000\000w\000Q\000k\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000R\000j\000m\000S\000T\000\000\000\180\000Q\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000l\000R\000\000\000m\000S\000T\000\000\000\182\000j\000\000\000\000\000n\000\000\000w\000Q\000k\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000R\000j\000m\000S\000T\000\000\000\201\000Q\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000l\000R\000\000\000m\000S\000T\000\000\000\212\000j\000\000\000\000\000n\000\000\000w\000Q\000k\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000R\000j\000m\000S\000T\000\000\000\230\000Q\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000l\000R\000\000\000m\000S\000T\000\000\000\232\000j\000\000\000\000\000n\000\000\000w\000Q\000k\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000R\000j\000m\000S\000T\000\000\000\238\000Q\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000l\000R\000\000\000m\000S\000T\000\000\000\240\000j\000\000\000\000\000n\000\000\000w\000Q\000k\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000R\000j\000m\000S\000T\000\000\000\242\000Q\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000l\000R\000\000\000m\000S\000T\000\000\001+\000j\000\000\000\000\000n\000\000\000w\000Q\000k\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000R\000j\000m\000S\000T\000\000\001-\000Q\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000l\000R\000\000\000m\000S\000T\000\000\001/\000j\000\000\000\000\000n\000\000\000w\000Q\000k\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000R\000j\000m\000S\000T\000\000\0018\000Q\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000l\000R\000\000\000m\000S\000T\000\000\001N\000j\000\000\000\000\000n\000\000\000w\000Q\000k\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000R\000j\000m\000S\000T\000\000\001P\000Q\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000l\000R\000\000\000m\000S\000T\000\000\001R\000j\000\000\000\000\000n\000\000\000w\000Q\000k\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000R\000j\000m\000S\000T\000\000\001T\000Q\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000l\000R\000\000\000m\000S\000T\000\000\001V\000j\000\000\000\000\000n\000\000\000w\000Q\000k\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000R\000j\000m\000S\000T\000\000\001X\000Q\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000l\000R\000\000\000m\000S\000T\000\000\001Z\000j\000\000\000\000\000n\000\000\000w\000Q\000k\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000\000\000\000\000\000\000\000\000\000\000\000\000l\000R\000j\000m\000S\000T\000\000\001\\\000\000\000k\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x\000Y\000\000\000Z\000[\000\000\000\\\000\000\000]\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000j\000\000\000\000\000n\000\000\000w\000\000\000k\000\000\000\000\000\000\000x\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000l\000\000\000\000\000m\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000n\000\000\000w\000\000\000\000\000\000\000\000\000\000\000x"))
  
  and semantic_action =
    [|
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = a;
          MenhirLib.EngineTypes.startp = _startpos_a_;
          MenhirLib.EngineTypes.endp = _endpos_a_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let a : (
# 99 "compiler/lib/js_parser.mly"
       (Js_token.Annot.t)
# 451 "compiler/lib/js_parser.ml"
        ) = Obj.magic a in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_a_ in
        let _endpos = _endpos_a_ in
        let _v : (Js_token.Annot.t) = 
# 150 "compiler/lib/js_parser.mly"
             ( a )
# 459 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = arg;
          MenhirLib.EngineTypes.startp = _startpos_arg_;
          MenhirLib.EngineTypes.endp = _endpos_arg_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let arg : (Javascript.expression) = Obj.magic arg in
        let _1 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 487 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_arg_ in
        let _v : (Javascript.expression * Javascript.spread) = 
# 551 "compiler/lib/js_parser.mly"
                                      ( arg, `Spread )
# 495 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = arg;
          MenhirLib.EngineTypes.startp = _startpos_arg_;
          MenhirLib.EngineTypes.endp = _endpos_arg_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let arg : (Javascript.expression) = Obj.magic arg in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_arg_ in
        let _endpos = _endpos_arg_ in
        let _v : (Javascript.expression * Javascript.spread) = 
# 552 "compiler/lib/js_parser.mly"
                             ( arg, `Not_spread )
# 520 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _3;
          MenhirLib.EngineTypes.startp = _startpos__3_;
          MenhirLib.EngineTypes.endp = _endpos__3_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = xs;
            MenhirLib.EngineTypes.startp = _startpos_xs_;
            MenhirLib.EngineTypes.endp = _endpos_xs_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = _1;
              MenhirLib.EngineTypes.startp = _startpos__1_;
              MenhirLib.EngineTypes.endp = _endpos__1_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let _3 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 553 "compiler/lib/js_parser.ml"
        ) = Obj.magic _3 in
        let xs : (Javascript.arguments) = Obj.magic xs in
        let _1 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 559 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v =
          let args =
            let item = 
# 229 "<standard.mly>"
    ( xs )
# 569 "compiler/lib/js_parser.ml"
             in
            
# 697 "compiler/lib/js_parser.mly"
                                                         ( item )
# 574 "compiler/lib/js_parser.ml"
            
          in
          (
# 555 "compiler/lib/js_parser.mly"
                                                    ( args )
# 580 "compiler/lib/js_parser.ml"
           : (Javascript.arguments))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _3;
          MenhirLib.EngineTypes.startp = _startpos__3_;
          MenhirLib.EngineTypes.endp = _endpos__3_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = pi;
              MenhirLib.EngineTypes.startp = _startpos_pi_;
              MenhirLib.EngineTypes.endp = _endpos_pi_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let _3 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 614 "compiler/lib/js_parser.ml"
        ) = Obj.magic _3 in
        let _2 : (Javascript.element_list) = Obj.magic _2 in
        let pi : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 620 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos__3_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 517 "compiler/lib/js_parser.mly"
     ( (pi, EArr _2) )
# 628 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _2;
          MenhirLib.EngineTypes.startp = _startpos__2_;
          MenhirLib.EngineTypes.endp = _endpos__2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = pi;
            MenhirLib.EngineTypes.startp = _startpos_pi_;
            MenhirLib.EngineTypes.endp = _endpos_pi_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 655 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let pi : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 660 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos__2_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 519 "compiler/lib/js_parser.mly"
     ( (pi, EArr []) )
# 668 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _3;
          MenhirLib.EngineTypes.startp = _startpos__3_;
          MenhirLib.EngineTypes.endp = _endpos__3_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = pi;
              MenhirLib.EngineTypes.startp = _startpos_pi_;
              MenhirLib.EngineTypes.endp = _endpos_pi_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let _3 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 701 "compiler/lib/js_parser.ml"
        ) = Obj.magic _3 in
        let _2 : (Javascript.element_list) = Obj.magic _2 in
        let pi : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 707 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos__3_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 521 "compiler/lib/js_parser.mly"
     ( (pi, EArr _2) )
# 715 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _4;
          MenhirLib.EngineTypes.startp = _startpos__4_;
          MenhirLib.EngineTypes.endp = _endpos__4_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _3;
            MenhirLib.EngineTypes.startp = _startpos__3_;
            MenhirLib.EngineTypes.endp = _endpos__3_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = _2;
              MenhirLib.EngineTypes.startp = _startpos__2_;
              MenhirLib.EngineTypes.endp = _endpos__2_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _menhir_s;
                MenhirLib.EngineTypes.semv = pi;
                MenhirLib.EngineTypes.startp = _startpos_pi_;
                MenhirLib.EngineTypes.endp = _endpos_pi_;
                MenhirLib.EngineTypes.next = _menhir_stack;
              };
            };
          };
        } = _menhir_stack in
        let _4 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 754 "compiler/lib/js_parser.ml"
        ) = Obj.magic _4 in
        let _3 : (Javascript.element_list) = Obj.magic _3 in
        let _2 : (Javascript.expression option list) = Obj.magic _2 in
        let pi : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 761 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos__4_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 523 "compiler/lib/js_parser.mly"
     ( (pi, EArr (List.rev_append _2 (List.rev _3))) )
# 769 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.expression) = 
# 322 "compiler/lib/js_parser.mly"
                          ( _1 )
# 794 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _3;
          MenhirLib.EngineTypes.startp = _startpos__3_;
          MenhirLib.EngineTypes.endp = _endpos__3_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = _1;
              MenhirLib.EngineTypes.startp = _startpos__1_;
              MenhirLib.EngineTypes.endp = _endpos__1_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let _3 : (Javascript.expression) = Obj.magic _3 in
        let _2 : (Javascript.binop) = Obj.magic _2 in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : (Javascript.expression) = 
# 324 "compiler/lib/js_parser.mly"
   ( EBin (_2, _1, _3) )
# 833 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.expression) = 
# 406 "compiler/lib/js_parser.mly"
                                ( _1 )
# 858 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _3;
          MenhirLib.EngineTypes.startp = _startpos__3_;
          MenhirLib.EngineTypes.endp = _endpos__3_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = _1;
              MenhirLib.EngineTypes.startp = _startpos__1_;
              MenhirLib.EngineTypes.endp = _endpos__1_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let _3 : (Javascript.expression) = Obj.magic _3 in
        let _2 : (Javascript.binop) = Obj.magic _2 in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : (Javascript.expression) = 
# 408 "compiler/lib/js_parser.mly"
     ( EBin(_2,_1,_3) )
# 897 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.expression) = 
# 430 "compiler/lib/js_parser.mly"
                                       ( _1 )
# 922 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _3;
          MenhirLib.EngineTypes.startp = _startpos__3_;
          MenhirLib.EngineTypes.endp = _endpos__3_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = _1;
              MenhirLib.EngineTypes.startp = _startpos__1_;
              MenhirLib.EngineTypes.endp = _endpos__1_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let _3 : (Javascript.expression) = Obj.magic _3 in
        let _2 : (Javascript.binop) = Obj.magic _2 in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : (Javascript.expression) = 
# 432 "compiler/lib/js_parser.mly"
   ( EBin (_2,_1,_3) )
# 961 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 74 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 982 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.binop) = 
# 678 "compiler/lib/js_parser.mly"
                    ( Eq )
# 990 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 74 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1011 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.binop) = 
# 679 "compiler/lib/js_parser.mly"
                    ( StarEq )
# 1019 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 74 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1040 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.binop) = 
# 680 "compiler/lib/js_parser.mly"
                    ( SlashEq )
# 1048 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 74 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1069 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.binop) = 
# 681 "compiler/lib/js_parser.mly"
                    ( ModEq )
# 1077 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 74 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1098 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.binop) = 
# 682 "compiler/lib/js_parser.mly"
                    ( PlusEq )
# 1106 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 74 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1127 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.binop) = 
# 683 "compiler/lib/js_parser.mly"
                    ( MinusEq )
# 1135 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 74 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1156 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.binop) = 
# 684 "compiler/lib/js_parser.mly"
                    ( LslEq )
# 1164 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 74 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1185 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.binop) = 
# 685 "compiler/lib/js_parser.mly"
                    ( AsrEq )
# 1193 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 74 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1214 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.binop) = 
# 686 "compiler/lib/js_parser.mly"
                    ( LsrEq )
# 1222 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 74 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1243 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.binop) = 
# 687 "compiler/lib/js_parser.mly"
                    ( BandEq )
# 1251 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 74 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1272 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.binop) = 
# 688 "compiler/lib/js_parser.mly"
                    ( BxorEq )
# 1280 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 74 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1301 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.binop) = 
# 689 "compiler/lib/js_parser.mly"
                    ( BorEq )
# 1309 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = block;
          MenhirLib.EngineTypes.startp = _startpos_block_;
          MenhirLib.EngineTypes.endp = _endpos_block_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let block : (Javascript.statement_list * Parse_info.t * Parse_info.t) = Obj.magic block in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_block_ in
        let _endpos = _endpos_block_ in
        let _v : (Javascript.statement_list) = 
# 198 "compiler/lib/js_parser.mly"
   ( let statements, _, _ = block in statements )
# 1334 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = pi;
          MenhirLib.EngineTypes.startp = _startpos_pi_;
          MenhirLib.EngineTypes.endp = _endpos_pi_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1355 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos_pi_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 492 "compiler/lib/js_parser.mly"
              ( (pi, EBool true) )
# 1363 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = pi;
          MenhirLib.EngineTypes.startp = _startpos_pi_;
          MenhirLib.EngineTypes.endp = _endpos_pi_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1384 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos_pi_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 493 "compiler/lib/js_parser.mly"
              ( (pi, EBool false) )
# 1392 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _2;
          MenhirLib.EngineTypes.startp = _startpos__2_;
          MenhirLib.EngineTypes.endp = _endpos__2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = pi;
            MenhirLib.EngineTypes.startp = _startpos_pi_;
            MenhirLib.EngineTypes.endp = _endpos_pi_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let _2 : ((Javascript.Label.t * Parse_info.t) option) = Obj.magic _2 in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1420 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos__2_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 256 "compiler/lib/js_parser.mly"
                     ( (Break_statement (Stdlib.Option.map ~f:fst _2), Pi pi) )
# 1428 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _2;
          MenhirLib.EngineTypes.startp = _startpos__2_;
          MenhirLib.EngineTypes.endp = _endpos__2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let _2 : (Javascript.arguments) = Obj.magic _2 in
        let _1 : (Parse_info.t * Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__2_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 359 "compiler/lib/js_parser.mly"
     ( let (start, e) = _1 in (start, ECall(e, _2, Pi start)) )
# 1460 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _2;
          MenhirLib.EngineTypes.startp = _startpos__2_;
          MenhirLib.EngineTypes.endp = _endpos__2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let _2 : (Javascript.arguments) = Obj.magic _2 in
        let _1 : (Parse_info.t * Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__2_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 361 "compiler/lib/js_parser.mly"
     ( let (start, e) = _1 in (start, ECall(e, _2, Pi start)) )
# 1492 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _4;
          MenhirLib.EngineTypes.startp = _startpos__4_;
          MenhirLib.EngineTypes.endp = _endpos__4_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _3;
            MenhirLib.EngineTypes.startp = _startpos__3_;
            MenhirLib.EngineTypes.endp = _endpos__3_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = _2;
              MenhirLib.EngineTypes.startp = _startpos__2_;
              MenhirLib.EngineTypes.endp = _endpos__2_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _menhir_s;
                MenhirLib.EngineTypes.semv = _1;
                MenhirLib.EngineTypes.startp = _startpos__1_;
                MenhirLib.EngineTypes.endp = _endpos__1_;
                MenhirLib.EngineTypes.next = _menhir_stack;
              };
            };
          };
        } = _menhir_stack in
        let _4 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1531 "compiler/lib/js_parser.ml"
        ) = Obj.magic _4 in
        let _3 : (Javascript.expression) = Obj.magic _3 in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1537 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let _1 : (Parse_info.t * Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__4_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 363 "compiler/lib/js_parser.mly"
     ( let (start, e) = _1 in (start, EAccess (e, _3)) )
# 1546 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _3;
          MenhirLib.EngineTypes.startp = _startpos__3_;
          MenhirLib.EngineTypes.endp = _endpos__3_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = _1;
              MenhirLib.EngineTypes.startp = _startpos__1_;
              MenhirLib.EngineTypes.endp = _endpos__1_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let _3 : (string) = Obj.magic _3 in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1580 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let _1 : (Parse_info.t * Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 365 "compiler/lib/js_parser.mly"
     ( let (start, e) = _1 in (start, EDot (e, _3)) )
# 1589 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _2;
          MenhirLib.EngineTypes.startp = _startpos__2_;
          MenhirLib.EngineTypes.endp = _endpos__2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let _2 : (Javascript.arguments) = Obj.magic _2 in
        let _1 : (Parse_info.t * Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__2_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 466 "compiler/lib/js_parser.mly"
   ( let (start, e) = _1 in (start, ECall(e, _2, Pi start)) )
# 1621 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _2;
          MenhirLib.EngineTypes.startp = _startpos__2_;
          MenhirLib.EngineTypes.endp = _endpos__2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let _2 : (Javascript.arguments) = Obj.magic _2 in
        let _1 : (Parse_info.t * Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__2_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 468 "compiler/lib/js_parser.mly"
   ( let (start, e) = _1 in (start, ECall(e, _2, Pi start)) )
# 1653 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _4;
          MenhirLib.EngineTypes.startp = _startpos__4_;
          MenhirLib.EngineTypes.endp = _endpos__4_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _3;
            MenhirLib.EngineTypes.startp = _startpos__3_;
            MenhirLib.EngineTypes.endp = _endpos__3_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = _2;
              MenhirLib.EngineTypes.startp = _startpos__2_;
              MenhirLib.EngineTypes.endp = _endpos__2_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _menhir_s;
                MenhirLib.EngineTypes.semv = _1;
                MenhirLib.EngineTypes.startp = _startpos__1_;
                MenhirLib.EngineTypes.endp = _endpos__1_;
                MenhirLib.EngineTypes.next = _menhir_stack;
              };
            };
          };
        } = _menhir_stack in
        let _4 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1692 "compiler/lib/js_parser.ml"
        ) = Obj.magic _4 in
        let _3 : (Javascript.expression) = Obj.magic _3 in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1698 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let _1 : (Parse_info.t * Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__4_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 470 "compiler/lib/js_parser.mly"
   ( let (start, e) = _1 in (start, EAccess(e, _3)) )
# 1707 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _3;
          MenhirLib.EngineTypes.startp = _startpos__3_;
          MenhirLib.EngineTypes.endp = _endpos__3_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = _1;
              MenhirLib.EngineTypes.startp = _startpos__1_;
              MenhirLib.EngineTypes.endp = _endpos__1_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let _3 : (string) = Obj.magic _3 in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1741 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let _1 : (Parse_info.t * Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 472 "compiler/lib/js_parser.mly"
   ( let (start, e) = _1 in (start, EDot(e,_3)) )
# 1750 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = y;
          MenhirLib.EngineTypes.startp = _startpos_y_;
          MenhirLib.EngineTypes.endp = _endpos_y_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = x;
              MenhirLib.EngineTypes.startp = _startpos_x_;
              MenhirLib.EngineTypes.endp = _endpos_x_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _menhir_s;
                MenhirLib.EngineTypes.semv = _1;
                MenhirLib.EngineTypes.startp = _startpos__1_;
                MenhirLib.EngineTypes.endp = _endpos__1_;
                MenhirLib.EngineTypes.next = _menhir_stack;
              };
            };
          };
        } = _menhir_stack in
        let y : (Javascript.statement_list) = Obj.magic y in
        let _2 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1790 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let x : (Javascript.expression) = Obj.magic x in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1796 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_y_ in
        let _v =
          let pair = 
# 172 "<standard.mly>"
    ( (x, y) )
# 1805 "compiler/lib/js_parser.ml"
           in
          (
# 292 "compiler/lib/js_parser.mly"
                                                               ( pair )
# 1810 "compiler/lib/js_parser.ml"
           : (Javascript.case_clause))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = y;
          MenhirLib.EngineTypes.startp = _startpos_y_;
          MenhirLib.EngineTypes.endp = _endpos_y_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _3;
            MenhirLib.EngineTypes.startp = _startpos__3_;
            MenhirLib.EngineTypes.endp = _endpos__3_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = item;
              MenhirLib.EngineTypes.startp = _startpos_item_;
              MenhirLib.EngineTypes.endp = _endpos_item_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _;
                MenhirLib.EngineTypes.semv = _1_inlined1;
                MenhirLib.EngineTypes.startp = _startpos__1_inlined1_;
                MenhirLib.EngineTypes.endp = _endpos__1_inlined1_;
                MenhirLib.EngineTypes.next = {
                  MenhirLib.EngineTypes.state = _menhir_s;
                  MenhirLib.EngineTypes.semv = _1;
                  MenhirLib.EngineTypes.startp = _startpos__1_;
                  MenhirLib.EngineTypes.endp = _endpos__1_;
                  MenhirLib.EngineTypes.next = _menhir_stack;
                };
              };
            };
          };
        } = _menhir_stack in
        let y : (Javascript.statement_list) = Obj.magic y in
        let _3 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1857 "compiler/lib/js_parser.ml"
        ) = Obj.magic _3 in
        let item : (Javascript.ident) = Obj.magic item in
        let _1_inlined1 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1863 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1_inlined1 in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 1868 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_y_ in
        let _v =
          let pair =
            let x = 
# 697 "compiler/lib/js_parser.mly"
                                                         ( item )
# 1878 "compiler/lib/js_parser.ml"
             in
            
# 164 "<standard.mly>"
    ( (x, y) )
# 1883 "compiler/lib/js_parser.ml"
            
          in
          (
# 282 "compiler/lib/js_parser.mly"
                                                     ( pair )
# 1889 "compiler/lib/js_parser.ml"
           : (Javascript.ident * Javascript.statement_list))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.expression) = 
# 331 "compiler/lib/js_parser.mly"
                      ( _1 )
# 1915 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.expression) = 
# 332 "compiler/lib/js_parser.mly"
                                                      ( _1 )
# 1940 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.expression) = 
# 411 "compiler/lib/js_parser.mly"
                            ( _1 )
# 1965 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.expression) = 
# 412 "compiler/lib/js_parser.mly"
                                                                  ( _1 )
# 1990 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.expression) = 
# 435 "compiler/lib/js_parser.mly"
                                   ( _1 )
# 2015 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.expression) = 
# 436 "compiler/lib/js_parser.mly"
                                                                   ( _1 )
# 2040 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _2;
          MenhirLib.EngineTypes.startp = _startpos__2_;
          MenhirLib.EngineTypes.endp = _endpos__2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = pi;
            MenhirLib.EngineTypes.startp = _startpos_pi_;
            MenhirLib.EngineTypes.endp = _endpos_pi_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let _2 : ((Javascript.Label.t * Parse_info.t) option) = Obj.magic _2 in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2068 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos__2_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 253 "compiler/lib/js_parser.mly"
                        ( (Continue_statement (Stdlib.Option.map ~f:fst _2),Pi pi) )
# 2076 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = pi2;
          MenhirLib.EngineTypes.startp = _startpos_pi2_;
          MenhirLib.EngineTypes.endp = _endpos_pi2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = x;
            MenhirLib.EngineTypes.startp = _startpos_x_;
            MenhirLib.EngineTypes.endp = _endpos_x_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = pi1;
              MenhirLib.EngineTypes.startp = _startpos_pi1_;
              MenhirLib.EngineTypes.endp = _endpos_pi1_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let pi2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2109 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi2 in
        let x : (unit) = Obj.magic x in
        let pi1 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2115 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi1_ in
        let _endpos = _endpos_pi2_ in
        let _v : (unit * Parse_info.t * Parse_info.t) = 
# 624 "compiler/lib/js_parser.mly"
                                 ( (x, pi1, pi2) )
# 2123 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = pi2;
          MenhirLib.EngineTypes.startp = _startpos_pi2_;
          MenhirLib.EngineTypes.endp = _endpos_pi2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = x;
            MenhirLib.EngineTypes.startp = _startpos_x_;
            MenhirLib.EngineTypes.endp = _endpos_x_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = pi1;
              MenhirLib.EngineTypes.startp = _startpos_pi1_;
              MenhirLib.EngineTypes.endp = _endpos_pi1_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let pi2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2156 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi2 in
        let x : (Javascript.source_elements) = Obj.magic x in
        let pi1 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2162 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi1_ in
        let _endpos = _endpos_pi2_ in
        let _v : (Javascript.source_elements * Parse_info.t * Parse_info.t) = 
# 624 "compiler/lib/js_parser.mly"
                                 ( (x, pi1, pi2) )
# 2170 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = pi2;
          MenhirLib.EngineTypes.startp = _startpos_pi2_;
          MenhirLib.EngineTypes.endp = _endpos_pi2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = x;
            MenhirLib.EngineTypes.startp = _startpos_x_;
            MenhirLib.EngineTypes.endp = _endpos_x_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = pi1;
              MenhirLib.EngineTypes.startp = _startpos_pi1_;
              MenhirLib.EngineTypes.endp = _endpos_pi1_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let pi2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2203 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi2 in
        let x : (Javascript.statement_list) = Obj.magic x in
        let pi1 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2209 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi1_ in
        let _endpos = _endpos_pi2_ in
        let _v : (Javascript.statement_list * Parse_info.t * Parse_info.t) = 
# 624 "compiler/lib/js_parser.mly"
                                 ( (x, pi1, pi2) )
# 2217 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = pi2;
          MenhirLib.EngineTypes.startp = _startpos_pi2_;
          MenhirLib.EngineTypes.endp = _endpos_pi2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = x;
            MenhirLib.EngineTypes.startp = _startpos_x_;
            MenhirLib.EngineTypes.endp = _endpos_x_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = pi1;
              MenhirLib.EngineTypes.startp = _startpos_pi1_;
              MenhirLib.EngineTypes.endp = _endpos_pi1_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let pi2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2250 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi2 in
        let x : (Javascript.property_name_and_value_list) = Obj.magic x in
        let pi1 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2256 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi1_ in
        let _endpos = _endpos_pi2_ in
        let _v : (Javascript.property_name_and_value_list * Parse_info.t * Parse_info.t) = 
# 624 "compiler/lib/js_parser.mly"
                                 ( (x, pi1, pi2) )
# 2264 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = pi;
          MenhirLib.EngineTypes.startp = _startpos_pi_;
          MenhirLib.EngineTypes.endp = _endpos_pi_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2285 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos_pi_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 211 "compiler/lib/js_parser.mly"
                 ( Debugger_statement, Pi pi )
# 2293 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = list;
          MenhirLib.EngineTypes.startp = _startpos_list_;
          MenhirLib.EngineTypes.endp = _endpos_list_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = _1;
              MenhirLib.EngineTypes.startp = _startpos__1_;
              MenhirLib.EngineTypes.endp = _endpos__1_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let list : (Javascript.statement_list) = Obj.magic list in
        let _2 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2327 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2332 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_list_ in
        let _v : (Javascript.statement_list) = 
# 295 "compiler/lib/js_parser.mly"
                                     ( list )
# 2340 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _3_inlined1;
          MenhirLib.EngineTypes.startp = _startpos__3_inlined1_;
          MenhirLib.EngineTypes.endp = _endpos__3_inlined1_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = item;
            MenhirLib.EngineTypes.startp = _startpos_item_;
            MenhirLib.EngineTypes.endp = _endpos_item_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = _1;
              MenhirLib.EngineTypes.startp = _startpos__1_;
              MenhirLib.EngineTypes.endp = _endpos__1_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _;
                MenhirLib.EngineTypes.semv = _3;
                MenhirLib.EngineTypes.startp = _startpos__3_;
                MenhirLib.EngineTypes.endp = _endpos__3_;
                MenhirLib.EngineTypes.next = {
                  MenhirLib.EngineTypes.state = _;
                  MenhirLib.EngineTypes.semv = body;
                  MenhirLib.EngineTypes.startp = _startpos_body_;
                  MenhirLib.EngineTypes.endp = _endpos_body_;
                  MenhirLib.EngineTypes.next = {
                    MenhirLib.EngineTypes.state = _menhir_s;
                    MenhirLib.EngineTypes.semv = pi;
                    MenhirLib.EngineTypes.startp = _startpos_pi_;
                    MenhirLib.EngineTypes.endp = _endpos_pi_;
                    MenhirLib.EngineTypes.next = _menhir_stack;
                  };
                };
              };
            };
          };
        } = _menhir_stack in
        let _3_inlined1 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2391 "compiler/lib/js_parser.ml"
        ) = Obj.magic _3_inlined1 in
        let item : (Javascript.expression) = Obj.magic item in
        let _1 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2397 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _3 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2402 "compiler/lib/js_parser.ml"
        ) = Obj.magic _3 in
        let body : (Javascript.statement * Javascript.location) = Obj.magic body in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2408 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos__3_inlined1_ in
        let _v =
          let condition = 
# 697 "compiler/lib/js_parser.mly"
                                                         ( item )
# 2417 "compiler/lib/js_parser.ml"
           in
          (
# 224 "compiler/lib/js_parser.mly"
    ( (Do_while_statement (body, condition), Pi pi) )
# 2422 "compiler/lib/js_parser.ml"
           : (Javascript.statement * Javascript.location))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2444 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Parse_info.t) = 
# 693 "compiler/lib/js_parser.mly"
                ( _1 )
# 2452 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 98 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2473 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Parse_info.t) = 
# 693 "compiler/lib/js_parser.mly"
                           ( _1 )
# 2481 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.expression option list) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.element_list) = 
# 526 "compiler/lib/js_parser.mly"
                    ( List.rev _1 )
# 2506 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _2;
          MenhirLib.EngineTypes.startp = _startpos__2_;
          MenhirLib.EngineTypes.endp = _endpos__2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let _2 : (Javascript.expression) = Obj.magic _2 in
        let _1 : (Javascript.element_list) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__2_ in
        let _v : (Javascript.expression option list) = 
# 529 "compiler/lib/js_parser.mly"
                                    ( (Some _2)::_1 )
# 2538 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.expression option list) = 
# 530 "compiler/lib/js_parser.mly"
                                    ( [Some _1] )
# 2563 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _3;
          MenhirLib.EngineTypes.startp = _startpos__3_;
          MenhirLib.EngineTypes.endp = _endpos__3_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = _1;
              MenhirLib.EngineTypes.startp = _startpos__1_;
              MenhirLib.EngineTypes.endp = _endpos__1_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let _3 : (Javascript.expression) = Obj.magic _3 in
        let _2 : (Javascript.element_list) = Obj.magic _2 in
        let _1 : (Javascript.expression option list) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : (Javascript.expression option list) = 
# 531 "compiler/lib/js_parser.mly"
                                                 ( (Some _3) :: (List.rev_append _2 _1) )
# 2602 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.element_list) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.element_list) = 
# 620 "compiler/lib/js_parser.mly"
                   (_1)
# 2627 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2648 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.element_list) = 
# 617 "compiler/lib/js_parser.mly"
           ( [] )
# 2656 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _2;
          MenhirLib.EngineTypes.startp = _startpos__2_;
          MenhirLib.EngineTypes.endp = _endpos__2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2683 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let _1 : (Javascript.element_list) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__2_ in
        let _v : (Javascript.element_list) = 
# 618 "compiler/lib/js_parser.mly"
                  ( None :: _1 )
# 2692 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let _menhir_s = _menhir_env.MenhirLib.EngineTypes.current in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _endpos = _startpos in
        let _v : (unit) = 
# 695 "compiler/lib/js_parser.mly"
       ()
# 2710 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = pi;
          MenhirLib.EngineTypes.startp = _startpos_pi_;
          MenhirLib.EngineTypes.endp = _endpos_pi_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let pi : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2731 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos_pi_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 208 "compiler/lib/js_parser.mly"
                  ( Empty_statement, Pi pi )
# 2739 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.expression) = 
# 318 "compiler/lib/js_parser.mly"
                         ( _1 )
# 2764 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _3;
          MenhirLib.EngineTypes.startp = _startpos__3_;
          MenhirLib.EngineTypes.endp = _endpos__3_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = _1;
              MenhirLib.EngineTypes.startp = _startpos__1_;
              MenhirLib.EngineTypes.endp = _endpos__1_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let _3 : (Javascript.expression) = Obj.magic _3 in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2798 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : (Javascript.expression) = 
# 319 "compiler/lib/js_parser.mly"
                                            ( ESeq (_1, _3) )
# 2807 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.expression) = 
# 402 "compiler/lib/js_parser.mly"
                               ( _1 )
# 2832 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _3;
          MenhirLib.EngineTypes.startp = _startpos__3_;
          MenhirLib.EngineTypes.endp = _endpos__3_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = _1;
              MenhirLib.EngineTypes.startp = _startpos__1_;
              MenhirLib.EngineTypes.endp = _endpos__1_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let _3 : (Javascript.expression) = Obj.magic _3 in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2866 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : (Javascript.expression) = 
# 403 "compiler/lib/js_parser.mly"
                                                        ( ESeq (_1, _3) )
# 2875 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.expression) = 
# 426 "compiler/lib/js_parser.mly"
                                      ( _1 )
# 2900 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _3;
          MenhirLib.EngineTypes.startp = _startpos__3_;
          MenhirLib.EngineTypes.endp = _endpos__3_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = _1;
              MenhirLib.EngineTypes.startp = _startpos__1_;
              MenhirLib.EngineTypes.endp = _endpos__1_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let _3 : (Javascript.expression) = Obj.magic _3 in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2934 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : (Javascript.expression) = 
# 427 "compiler/lib/js_parser.mly"
                                                         ( ESeq(_1,_3) )
# 2943 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 214 "compiler/lib/js_parser.mly"
                           ( Expression_statement _1, N )
# 2968 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _2;
          MenhirLib.EngineTypes.startp = _startpos__2_;
          MenhirLib.EngineTypes.endp = _endpos__2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let _2 : (Javascript.statement_list) = Obj.magic _2 in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 2996 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__2_ in
        let _v : (Javascript.statement_list) = 
# 285 "compiler/lib/js_parser.mly"
                   ( _2 )
# 3004 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = body;
          MenhirLib.EngineTypes.startp = _startpos_body_;
          MenhirLib.EngineTypes.endp = _endpos_body_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _6;
            MenhirLib.EngineTypes.startp = _startpos__6_;
            MenhirLib.EngineTypes.endp = _endpos__6_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = right;
              MenhirLib.EngineTypes.startp = _startpos_right_;
              MenhirLib.EngineTypes.endp = _endpos_right_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _;
                MenhirLib.EngineTypes.semv = _4;
                MenhirLib.EngineTypes.startp = _startpos__4_;
                MenhirLib.EngineTypes.endp = _endpos__4_;
                MenhirLib.EngineTypes.next = {
                  MenhirLib.EngineTypes.state = _;
                  MenhirLib.EngineTypes.semv = left;
                  MenhirLib.EngineTypes.startp = _startpos_left_;
                  MenhirLib.EngineTypes.endp = _endpos_left_;
                  MenhirLib.EngineTypes.next = {
                    MenhirLib.EngineTypes.state = _;
                    MenhirLib.EngineTypes.semv = _2;
                    MenhirLib.EngineTypes.startp = _startpos__2_;
                    MenhirLib.EngineTypes.endp = _endpos__2_;
                    MenhirLib.EngineTypes.next = {
                      MenhirLib.EngineTypes.state = _menhir_s;
                      MenhirLib.EngineTypes.semv = pi;
                      MenhirLib.EngineTypes.startp = _startpos_pi_;
                      MenhirLib.EngineTypes.endp = _endpos_pi_;
                      MenhirLib.EngineTypes.next = _menhir_stack;
                    };
                  };
                };
              };
            };
          };
        } = _menhir_stack in
        let body : (Javascript.statement * Javascript.location) = Obj.magic body in
        let _6 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3062 "compiler/lib/js_parser.ml"
        ) = Obj.magic _6 in
        let right : (Javascript.expression) = Obj.magic right in
        let _4 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3068 "compiler/lib/js_parser.ml"
        ) = Obj.magic _4 in
        let left : (Javascript.expression) = Obj.magic left in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3074 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3079 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos_body_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 244 "compiler/lib/js_parser.mly"
   ( ForIn_statement (Left left, right, body), Pi pi )
# 3087 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = body;
          MenhirLib.EngineTypes.startp = _startpos_body_;
          MenhirLib.EngineTypes.endp = _endpos_body_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _7;
            MenhirLib.EngineTypes.startp = _startpos__7_;
            MenhirLib.EngineTypes.endp = _endpos__7_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = right;
              MenhirLib.EngineTypes.startp = _startpos_right_;
              MenhirLib.EngineTypes.endp = _endpos_right_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _;
                MenhirLib.EngineTypes.semv = _5;
                MenhirLib.EngineTypes.startp = _startpos__5_;
                MenhirLib.EngineTypes.endp = _endpos__5_;
                MenhirLib.EngineTypes.next = {
                  MenhirLib.EngineTypes.state = _;
                  MenhirLib.EngineTypes.semv = y;
                  MenhirLib.EngineTypes.startp = _startpos_y_;
                  MenhirLib.EngineTypes.endp = _endpos_y_;
                  MenhirLib.EngineTypes.next = {
                    MenhirLib.EngineTypes.state = _;
                    MenhirLib.EngineTypes.semv = x;
                    MenhirLib.EngineTypes.startp = _startpos_x_;
                    MenhirLib.EngineTypes.endp = _endpos_x_;
                    MenhirLib.EngineTypes.next = {
                      MenhirLib.EngineTypes.state = _;
                      MenhirLib.EngineTypes.semv = _3;
                      MenhirLib.EngineTypes.startp = _startpos__3_;
                      MenhirLib.EngineTypes.endp = _endpos__3_;
                      MenhirLib.EngineTypes.next = {
                        MenhirLib.EngineTypes.state = _;
                        MenhirLib.EngineTypes.semv = _2;
                        MenhirLib.EngineTypes.startp = _startpos__2_;
                        MenhirLib.EngineTypes.endp = _endpos__2_;
                        MenhirLib.EngineTypes.next = {
                          MenhirLib.EngineTypes.state = _menhir_s;
                          MenhirLib.EngineTypes.semv = pi;
                          MenhirLib.EngineTypes.startp = _startpos_pi_;
                          MenhirLib.EngineTypes.endp = _endpos_pi_;
                          MenhirLib.EngineTypes.next = _menhir_stack;
                        };
                      };
                    };
                  };
                };
              };
            };
          };
        } = _menhir_stack in
        let body : (Javascript.statement * Javascript.location) = Obj.magic body in
        let _7 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3157 "compiler/lib/js_parser.ml"
        ) = Obj.magic _7 in
        let right : (Javascript.expression) = Obj.magic right in
        let _5 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3163 "compiler/lib/js_parser.ml"
        ) = Obj.magic _5 in
        let y : (Javascript.initialiser option) = Obj.magic y in
        let x : (Javascript.ident) = Obj.magic x in
        let _3 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3170 "compiler/lib/js_parser.ml"
        ) = Obj.magic _3 in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3175 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3180 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos_body_ in
        let _v =
          let left = 
# 164 "<standard.mly>"
    ( (x, y) )
# 3189 "compiler/lib/js_parser.ml"
           in
          (
# 247 "compiler/lib/js_parser.mly"
   ( ForIn_statement (Right left, right, body), Pi pi )
# 3194 "compiler/lib/js_parser.ml"
           : (Javascript.statement * Javascript.location))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = statement;
          MenhirLib.EngineTypes.startp = _startpos_statement_;
          MenhirLib.EngineTypes.endp = _endpos_statement_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _8;
            MenhirLib.EngineTypes.startp = _startpos__8_;
            MenhirLib.EngineTypes.endp = _endpos__8_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = increment;
              MenhirLib.EngineTypes.startp = _startpos_increment_;
              MenhirLib.EngineTypes.endp = _endpos_increment_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _;
                MenhirLib.EngineTypes.semv = _6;
                MenhirLib.EngineTypes.startp = _startpos__6_;
                MenhirLib.EngineTypes.endp = _endpos__6_;
                MenhirLib.EngineTypes.next = {
                  MenhirLib.EngineTypes.state = _;
                  MenhirLib.EngineTypes.semv = condition;
                  MenhirLib.EngineTypes.startp = _startpos_condition_;
                  MenhirLib.EngineTypes.endp = _endpos_condition_;
                  MenhirLib.EngineTypes.next = {
                    MenhirLib.EngineTypes.state = _;
                    MenhirLib.EngineTypes.semv = _4;
                    MenhirLib.EngineTypes.startp = _startpos__4_;
                    MenhirLib.EngineTypes.endp = _endpos__4_;
                    MenhirLib.EngineTypes.next = {
                      MenhirLib.EngineTypes.state = _;
                      MenhirLib.EngineTypes.semv = initial;
                      MenhirLib.EngineTypes.startp = _startpos_initial_;
                      MenhirLib.EngineTypes.endp = _endpos_initial_;
                      MenhirLib.EngineTypes.next = {
                        MenhirLib.EngineTypes.state = _;
                        MenhirLib.EngineTypes.semv = _2;
                        MenhirLib.EngineTypes.startp = _startpos__2_;
                        MenhirLib.EngineTypes.endp = _endpos__2_;
                        MenhirLib.EngineTypes.next = {
                          MenhirLib.EngineTypes.state = _menhir_s;
                          MenhirLib.EngineTypes.semv = pi;
                          MenhirLib.EngineTypes.startp = _startpos_pi_;
                          MenhirLib.EngineTypes.endp = _endpos_pi_;
                          MenhirLib.EngineTypes.next = _menhir_stack;
                        };
                      };
                    };
                  };
                };
              };
            };
          };
        } = _menhir_stack in
        let statement : (Javascript.statement * Javascript.location) = Obj.magic statement in
        let _8 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3265 "compiler/lib/js_parser.ml"
        ) = Obj.magic _8 in
        let increment : (Javascript.expression option) = Obj.magic increment in
        let _6 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3271 "compiler/lib/js_parser.ml"
        ) = Obj.magic _6 in
        let condition : (Javascript.expression option) = Obj.magic condition in
        let _4 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3277 "compiler/lib/js_parser.ml"
        ) = Obj.magic _4 in
        let initial : (Javascript.expression option) = Obj.magic initial in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3283 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3288 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos_statement_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 234 "compiler/lib/js_parser.mly"
   ( For_statement (Left initial, condition, increment, statement), Pi pi )
# 3296 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = statement;
          MenhirLib.EngineTypes.startp = _startpos_statement_;
          MenhirLib.EngineTypes.endp = _endpos_statement_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _9;
            MenhirLib.EngineTypes.startp = _startpos__9_;
            MenhirLib.EngineTypes.endp = _endpos__9_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = increment;
              MenhirLib.EngineTypes.startp = _startpos_increment_;
              MenhirLib.EngineTypes.endp = _endpos_increment_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _;
                MenhirLib.EngineTypes.semv = _7;
                MenhirLib.EngineTypes.startp = _startpos__7_;
                MenhirLib.EngineTypes.endp = _endpos__7_;
                MenhirLib.EngineTypes.next = {
                  MenhirLib.EngineTypes.state = _;
                  MenhirLib.EngineTypes.semv = condition;
                  MenhirLib.EngineTypes.startp = _startpos_condition_;
                  MenhirLib.EngineTypes.endp = _endpos_condition_;
                  MenhirLib.EngineTypes.next = {
                    MenhirLib.EngineTypes.state = _;
                    MenhirLib.EngineTypes.semv = _5;
                    MenhirLib.EngineTypes.startp = _startpos__5_;
                    MenhirLib.EngineTypes.endp = _endpos__5_;
                    MenhirLib.EngineTypes.next = {
                      MenhirLib.EngineTypes.state = _;
                      MenhirLib.EngineTypes.semv = initial;
                      MenhirLib.EngineTypes.startp = _startpos_initial_;
                      MenhirLib.EngineTypes.endp = _endpos_initial_;
                      MenhirLib.EngineTypes.next = {
                        MenhirLib.EngineTypes.state = _;
                        MenhirLib.EngineTypes.semv = _3;
                        MenhirLib.EngineTypes.startp = _startpos__3_;
                        MenhirLib.EngineTypes.endp = _endpos__3_;
                        MenhirLib.EngineTypes.next = {
                          MenhirLib.EngineTypes.state = _;
                          MenhirLib.EngineTypes.semv = _2;
                          MenhirLib.EngineTypes.startp = _startpos__2_;
                          MenhirLib.EngineTypes.endp = _endpos__2_;
                          MenhirLib.EngineTypes.next = {
                            MenhirLib.EngineTypes.state = _menhir_s;
                            MenhirLib.EngineTypes.semv = pi;
                            MenhirLib.EngineTypes.startp = _startpos_pi_;
                            MenhirLib.EngineTypes.endp = _endpos_pi_;
                            MenhirLib.EngineTypes.next = _menhir_stack;
                          };
                        };
                      };
                    };
                  };
                };
              };
            };
          };
        } = _menhir_stack in
        let statement : (Javascript.statement * Javascript.location) = Obj.magic statement in
        let _9 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3372 "compiler/lib/js_parser.ml"
        ) = Obj.magic _9 in
        let increment : (Javascript.expression option) = Obj.magic increment in
        let _7 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3378 "compiler/lib/js_parser.ml"
        ) = Obj.magic _7 in
        let condition : (Javascript.expression option) = Obj.magic condition in
        let _5 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3384 "compiler/lib/js_parser.ml"
        ) = Obj.magic _5 in
        let initial : (Javascript.variable_declaration list) = Obj.magic initial in
        let _3 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3390 "compiler/lib/js_parser.ml"
        ) = Obj.magic _3 in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3395 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3400 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos_statement_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 239 "compiler/lib/js_parser.mly"
   ( For_statement (Right initial, condition, increment, statement), Pi pi )
# 3408 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = block;
          MenhirLib.EngineTypes.startp = _startpos_block_;
          MenhirLib.EngineTypes.endp = _endpos_block_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _3;
            MenhirLib.EngineTypes.startp = _startpos__3_;
            MenhirLib.EngineTypes.endp = _endpos__3_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = xs;
              MenhirLib.EngineTypes.startp = _startpos_xs_;
              MenhirLib.EngineTypes.endp = _endpos_xs_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _;
                MenhirLib.EngineTypes.semv = _1;
                MenhirLib.EngineTypes.startp = _startpos__1_;
                MenhirLib.EngineTypes.endp = _endpos__1_;
                MenhirLib.EngineTypes.next = {
                  MenhirLib.EngineTypes.state = _;
                  MenhirLib.EngineTypes.semv = name;
                  MenhirLib.EngineTypes.startp = _startpos_name_;
                  MenhirLib.EngineTypes.endp = _endpos_name_;
                  MenhirLib.EngineTypes.next = {
                    MenhirLib.EngineTypes.state = _menhir_s;
                    MenhirLib.EngineTypes.semv = pi;
                    MenhirLib.EngineTypes.startp = _startpos_pi_;
                    MenhirLib.EngineTypes.endp = _endpos_pi_;
                    MenhirLib.EngineTypes.next = _menhir_stack;
                  };
                };
              };
            };
          };
        } = _menhir_stack in
        let block : (Javascript.source_elements * Parse_info.t * Parse_info.t) = Obj.magic block in
        let _3 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3460 "compiler/lib/js_parser.ml"
        ) = Obj.magic _3 in
        let xs : (Javascript.formal_parameter_list) = Obj.magic xs in
        let _1 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3466 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let name : (Javascript.ident) = Obj.magic name in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3472 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos_block_ in
        let _v =
          let args =
            let item = 
# 229 "<standard.mly>"
    ( xs )
# 3482 "compiler/lib/js_parser.ml"
             in
            
# 697 "compiler/lib/js_parser.mly"
                                                         ( item )
# 3487 "compiler/lib/js_parser.ml"
            
          in
          (
# 304 "compiler/lib/js_parser.mly"
   ( let elements, _pi_start, pi_end = block in
     (name, args, elements, Pi pi_end), Pi pi )
# 3494 "compiler/lib/js_parser.ml"
           : (Javascript.function_declaration * Javascript.location))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = block;
          MenhirLib.EngineTypes.startp = _startpos_block_;
          MenhirLib.EngineTypes.endp = _endpos_block_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _3;
            MenhirLib.EngineTypes.startp = _startpos__3_;
            MenhirLib.EngineTypes.endp = _endpos__3_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = xs;
              MenhirLib.EngineTypes.startp = _startpos_xs_;
              MenhirLib.EngineTypes.endp = _endpos_xs_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _;
                MenhirLib.EngineTypes.semv = _1;
                MenhirLib.EngineTypes.startp = _startpos__1_;
                MenhirLib.EngineTypes.endp = _endpos__1_;
                MenhirLib.EngineTypes.next = {
                  MenhirLib.EngineTypes.state = _;
                  MenhirLib.EngineTypes.semv = name;
                  MenhirLib.EngineTypes.startp = _startpos_name_;
                  MenhirLib.EngineTypes.endp = _endpos_name_;
                  MenhirLib.EngineTypes.next = {
                    MenhirLib.EngineTypes.state = _menhir_s;
                    MenhirLib.EngineTypes.semv = pi;
                    MenhirLib.EngineTypes.startp = _startpos_pi_;
                    MenhirLib.EngineTypes.endp = _endpos_pi_;
                    MenhirLib.EngineTypes.next = _menhir_stack;
                  };
                };
              };
            };
          };
        } = _menhir_stack in
        let block : (Javascript.source_elements * Parse_info.t * Parse_info.t) = Obj.magic block in
        let _3 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3547 "compiler/lib/js_parser.ml"
        ) = Obj.magic _3 in
        let xs : (Javascript.formal_parameter_list) = Obj.magic xs in
        let _1 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3553 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let name : (Javascript.ident option) = Obj.magic name in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3559 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos_block_ in
        let _v =
          let args =
            let item = 
# 229 "<standard.mly>"
    ( xs )
# 3569 "compiler/lib/js_parser.ml"
             in
            
# 697 "compiler/lib/js_parser.mly"
                                                         ( item )
# 3574 "compiler/lib/js_parser.ml"
            
          in
          (
# 310 "compiler/lib/js_parser.mly"
   ( let elements, _pi_start, _pi_end = block in
     pi, EFun (name, args, elements, Pi pi) )
# 3581 "compiler/lib/js_parser.ml"
           : (Parse_info.t * Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 48 "compiler/lib/js_parser.mly"
      (string * Parse_info.t)
# 3603 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 566 "compiler/lib/js_parser.mly"
                  ( fst _1 )
# 3611 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3632 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 567 "compiler/lib/js_parser.mly"
             ( "catch" )
# 3640 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3661 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 568 "compiler/lib/js_parser.mly"
               ( "finally" )
# 3669 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3690 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 569 "compiler/lib/js_parser.mly"
          ( "in" )
# 3698 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3719 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 570 "compiler/lib/js_parser.mly"
                  ( "instanceof" )
# 3727 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 59 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3748 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 571 "compiler/lib/js_parser.mly"
            ( "else" )
# 3756 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3777 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 572 "compiler/lib/js_parser.mly"
             ( "break" )
# 3785 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3806 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 573 "compiler/lib/js_parser.mly"
            ( "case" )
# 3814 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3835 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 574 "compiler/lib/js_parser.mly"
                ( "continue" )
# 3843 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3864 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 575 "compiler/lib/js_parser.mly"
               ( "default" )
# 3872 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3893 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 576 "compiler/lib/js_parser.mly"
              ( "delete" )
# 3901 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3922 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 577 "compiler/lib/js_parser.mly"
          ( "do" )
# 3930 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3951 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 578 "compiler/lib/js_parser.mly"
           ( "for" )
# 3959 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 3980 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 579 "compiler/lib/js_parser.mly"
                ( "function" )
# 3988 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4009 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 580 "compiler/lib/js_parser.mly"
          ( "if" )
# 4017 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 61 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4038 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 581 "compiler/lib/js_parser.mly"
           ( "new" )
# 4046 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4067 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 582 "compiler/lib/js_parser.mly"
              ( "return" )
# 4075 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4096 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 583 "compiler/lib/js_parser.mly"
              ( "switch" )
# 4104 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4125 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 584 "compiler/lib/js_parser.mly"
            ( "this" )
# 4133 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4154 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 585 "compiler/lib/js_parser.mly"
             ( "throw" )
# 4162 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4183 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 586 "compiler/lib/js_parser.mly"
           ( "try" )
# 4191 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4212 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 587 "compiler/lib/js_parser.mly"
              ( "typeof" )
# 4220 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4241 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 588 "compiler/lib/js_parser.mly"
           ( "var" )
# 4249 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4270 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 589 "compiler/lib/js_parser.mly"
            ( "void" )
# 4278 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4299 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 590 "compiler/lib/js_parser.mly"
             ( "while" )
# 4307 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4328 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 591 "compiler/lib/js_parser.mly"
            ( "with" )
# 4336 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4357 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 592 "compiler/lib/js_parser.mly"
            ( "null" )
# 4365 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4386 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 593 "compiler/lib/js_parser.mly"
             ( "false" )
# 4394 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4415 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 594 "compiler/lib/js_parser.mly"
            ( "true" )
# 4423 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4444 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (string) = 
# 595 "compiler/lib/js_parser.mly"
                ( "debugger" )
# 4452 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _4;
            MenhirLib.EngineTypes.startp = _startpos__4_;
            MenhirLib.EngineTypes.endp = _endpos__4_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = t;
              MenhirLib.EngineTypes.startp = _startpos_t_;
              MenhirLib.EngineTypes.endp = _endpos_t_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _;
                MenhirLib.EngineTypes.semv = _3;
                MenhirLib.EngineTypes.startp = _startpos__3_;
                MenhirLib.EngineTypes.endp = _endpos__3_;
                MenhirLib.EngineTypes.next = {
                  MenhirLib.EngineTypes.state = _;
                  MenhirLib.EngineTypes.semv = item;
                  MenhirLib.EngineTypes.startp = _startpos_item_;
                  MenhirLib.EngineTypes.endp = _endpos_item_;
                  MenhirLib.EngineTypes.next = {
                    MenhirLib.EngineTypes.state = _;
                    MenhirLib.EngineTypes.semv = _1;
                    MenhirLib.EngineTypes.startp = _startpos__1_;
                    MenhirLib.EngineTypes.endp = _endpos__1_;
                    MenhirLib.EngineTypes.next = {
                      MenhirLib.EngineTypes.state = _menhir_s;
                      MenhirLib.EngineTypes.semv = pi;
                      MenhirLib.EngineTypes.startp = _startpos_pi_;
                      MenhirLib.EngineTypes.endp = _endpos_pi_;
                      MenhirLib.EngineTypes.next = _menhir_stack;
                    };
                  };
                };
              };
            };
          };
        } = _menhir_stack in
        let e : (Javascript.statement * Javascript.location) = Obj.magic e in
        let _4 : (
# 59 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4510 "compiler/lib/js_parser.ml"
        ) = Obj.magic _4 in
        let t : (Javascript.statement * Javascript.location) = Obj.magic t in
        let _3 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4516 "compiler/lib/js_parser.ml"
        ) = Obj.magic _3 in
        let item : (Javascript.expression) = Obj.magic item in
        let _1 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4522 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4527 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos_e_ in
        let _v =
          let condition = 
# 697 "compiler/lib/js_parser.mly"
                                                         ( item )
# 4536 "compiler/lib/js_parser.ml"
           in
          (
# 218 "compiler/lib/js_parser.mly"
     ( (If_statement (condition, t, Some e), Pi pi) )
# 4541 "compiler/lib/js_parser.ml"
           : (Javascript.statement * Javascript.location))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = t;
          MenhirLib.EngineTypes.startp = _startpos_t_;
          MenhirLib.EngineTypes.endp = _endpos_t_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _3;
            MenhirLib.EngineTypes.startp = _startpos__3_;
            MenhirLib.EngineTypes.endp = _endpos__3_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = item;
              MenhirLib.EngineTypes.startp = _startpos_item_;
              MenhirLib.EngineTypes.endp = _endpos_item_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _;
                MenhirLib.EngineTypes.semv = _1;
                MenhirLib.EngineTypes.startp = _startpos__1_;
                MenhirLib.EngineTypes.endp = _endpos__1_;
                MenhirLib.EngineTypes.next = {
                  MenhirLib.EngineTypes.state = _menhir_s;
                  MenhirLib.EngineTypes.semv = pi;
                  MenhirLib.EngineTypes.startp = _startpos_pi_;
                  MenhirLib.EngineTypes.endp = _endpos_pi_;
                  MenhirLib.EngineTypes.next = _menhir_stack;
                };
              };
            };
          };
        } = _menhir_stack in
        let t : (Javascript.statement * Javascript.location) = Obj.magic t in
        let _3 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4588 "compiler/lib/js_parser.ml"
        ) = Obj.magic _3 in
        let item : (Javascript.expression) = Obj.magic item in
        let _1 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4594 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4599 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos_t_ in
        let _v =
          let condition = 
# 697 "compiler/lib/js_parser.mly"
                                                         ( item )
# 4608 "compiler/lib/js_parser.ml"
           in
          (
# 220 "compiler/lib/js_parser.mly"
     ( (If_statement (condition, t, None), Pi pi) )
# 4613 "compiler/lib/js_parser.ml"
           : (Javascript.statement * Javascript.location))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = pi;
            MenhirLib.EngineTypes.startp = _startpos_pi_;
            MenhirLib.EngineTypes.endp = _endpos_pi_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let pi : (
# 74 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4642 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos_e_ in
        let _v : (Javascript.initialiser) = 
# 205 "compiler/lib/js_parser.mly"
                                       ( e, Pi pi )
# 4650 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _2;
          MenhirLib.EngineTypes.startp = _startpos__2_;
          MenhirLib.EngineTypes.endp = _endpos__2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let _2 : (Javascript.expression) = Obj.magic _2 in
        let _1 : (
# 74 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4678 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__2_ in
        let _v : (Javascript.initialiser) = 
# 250 "compiler/lib/js_parser.mly"
                                        ( _2, Pi _1 )
# 4686 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 48 "compiler/lib/js_parser.mly"
      (string * Parse_info.t)
# 4707 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.Label.t * Parse_info.t) = 
# 604 "compiler/lib/js_parser.mly"
                ( Label.of_string (fst _1), snd _1 )
# 4715 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = s;
          MenhirLib.EngineTypes.startp = _startpos_s_;
          MenhirLib.EngineTypes.endp = _endpos_s_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = l;
              MenhirLib.EngineTypes.startp = _startpos_l_;
              MenhirLib.EngineTypes.endp = _endpos_l_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let s : (Javascript.statement * Javascript.location) = Obj.magic s in
        let _2 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 4749 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let l : (Javascript.Label.t * Parse_info.t) = Obj.magic l in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_l_ in
        let _endpos = _endpos_s_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 194 "compiler/lib/js_parser.mly"
                              ( Labelled_statement (fst l, s), Pi (snd l))
# 4758 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Parse_info.t * Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.expression) = 
# 327 "compiler/lib/js_parser.mly"
                   ( snd _1 )
# 4783 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Parse_info.t * Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.expression) = 
# 328 "compiler/lib/js_parser.mly"
                   ( snd _1 )
# 4808 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Parse_info.t * Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.expression) = 
# 457 "compiler/lib/js_parser.mly"
                               ( snd _1 )
# 4833 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Parse_info.t * Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.expression) = 
# 458 "compiler/lib/js_parser.mly"
                                ( snd _1 )
# 4858 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let _menhir_s = _menhir_env.MenhirLib.EngineTypes.current in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _endpos = _startpos in
        let _v : (Js_token.Annot.t list) = 
# 208 "<standard.mly>"
    ( [] )
# 4876 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = xs;
          MenhirLib.EngineTypes.startp = _startpos_xs_;
          MenhirLib.EngineTypes.endp = _endpos_xs_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = x;
            MenhirLib.EngineTypes.startp = _startpos_x_;
            MenhirLib.EngineTypes.endp = _endpos_x_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let xs : (Js_token.Annot.t list) = Obj.magic xs in
        let x : (Js_token.Annot.t) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_xs_ in
        let _v : (Js_token.Annot.t list) = 
# 210 "<standard.mly>"
    ( x :: xs )
# 4908 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let _menhir_s = _menhir_env.MenhirLib.EngineTypes.current in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _endpos = _startpos in
        let _v : (Javascript.case_clause list) = 
# 208 "<standard.mly>"
    ( [] )
# 4926 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = xs;
          MenhirLib.EngineTypes.startp = _startpos_xs_;
          MenhirLib.EngineTypes.endp = _endpos_xs_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = x;
            MenhirLib.EngineTypes.startp = _startpos_x_;
            MenhirLib.EngineTypes.endp = _endpos_x_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let xs : (Javascript.case_clause list) = Obj.magic xs in
        let x : (Javascript.case_clause) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_xs_ in
        let _v : (Javascript.case_clause list) = 
# 210 "<standard.mly>"
    ( x :: xs )
# 4958 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let _menhir_s = _menhir_env.MenhirLib.EngineTypes.current in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _endpos = _startpos in
        let _v : (Javascript.source_elements) = 
# 208 "<standard.mly>"
    ( [] )
# 4976 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = xs;
          MenhirLib.EngineTypes.startp = _startpos_xs_;
          MenhirLib.EngineTypes.endp = _endpos_xs_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = x;
            MenhirLib.EngineTypes.startp = _startpos_x_;
            MenhirLib.EngineTypes.endp = _endpos_x_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let xs : (Javascript.source_elements) = Obj.magic xs in
        let x : (Javascript.source_element * Javascript.location) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_xs_ in
        let _v : (Javascript.source_elements) = 
# 210 "<standard.mly>"
    ( x :: xs )
# 5008 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let _menhir_s = _menhir_env.MenhirLib.EngineTypes.current in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _endpos = _startpos in
        let _v : (Javascript.program_with_annots) = 
# 208 "<standard.mly>"
    ( [] )
# 5026 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = xs;
          MenhirLib.EngineTypes.startp = _startpos_xs_;
          MenhirLib.EngineTypes.endp = _endpos_xs_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = x;
            MenhirLib.EngineTypes.startp = _startpos_x_;
            MenhirLib.EngineTypes.endp = _endpos_x_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let xs : (Javascript.program_with_annots) = Obj.magic xs in
        let x : ((Javascript.source_element * Javascript.location) * Js_token.Annot.t list) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_xs_ in
        let _v : (Javascript.program_with_annots) = 
# 210 "<standard.mly>"
    ( x :: xs )
# 5058 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let _menhir_s = _menhir_env.MenhirLib.EngineTypes.current in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _endpos = _startpos in
        let _v : (Javascript.statement_list) = 
# 208 "<standard.mly>"
    ( [] )
# 5076 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = xs;
          MenhirLib.EngineTypes.startp = _startpos_xs_;
          MenhirLib.EngineTypes.endp = _endpos_xs_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = x;
            MenhirLib.EngineTypes.startp = _startpos_x_;
            MenhirLib.EngineTypes.endp = _endpos_x_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let xs : (Javascript.statement_list) = Obj.magic xs in
        let x : (Javascript.statement * Javascript.location) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_xs_ in
        let _v : (Javascript.statement_list) = 
# 210 "<standard.mly>"
    ( x :: xs )
# 5108 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let _menhir_s = _menhir_env.MenhirLib.EngineTypes.current in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _endpos = _startpos in
        let _v : (Javascript.arguments) = 
# 139 "<standard.mly>"
    ( [] )
# 5126 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = x;
          MenhirLib.EngineTypes.startp = _startpos_x_;
          MenhirLib.EngineTypes.endp = _endpos_x_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let x : (Javascript.arguments) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_x_ in
        let _v : (Javascript.arguments) = 
# 141 "<standard.mly>"
    ( x )
# 5151 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let _menhir_s = _menhir_env.MenhirLib.EngineTypes.current in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _endpos = _startpos in
        let _v : (Javascript.formal_parameter_list) = 
# 139 "<standard.mly>"
    ( [] )
# 5169 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = x;
          MenhirLib.EngineTypes.startp = _startpos_x_;
          MenhirLib.EngineTypes.endp = _endpos_x_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let x : (Javascript.formal_parameter_list) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_x_ in
        let _v : (Javascript.formal_parameter_list) = 
# 141 "<standard.mly>"
    ( x )
# 5194 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let e : (Parse_info.t * Javascript.expression) = Obj.magic e in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_e_ in
        let _endpos = _endpos_e_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 373 "compiler/lib/js_parser.mly"
     ( e )
# 5219 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _4;
          MenhirLib.EngineTypes.startp = _startpos__4_;
          MenhirLib.EngineTypes.endp = _endpos__4_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = e2;
            MenhirLib.EngineTypes.startp = _startpos_e2_;
            MenhirLib.EngineTypes.endp = _endpos_e2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = _2;
              MenhirLib.EngineTypes.startp = _startpos__2_;
              MenhirLib.EngineTypes.endp = _endpos__2_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _menhir_s;
                MenhirLib.EngineTypes.semv = _1;
                MenhirLib.EngineTypes.startp = _startpos__1_;
                MenhirLib.EngineTypes.endp = _endpos__1_;
                MenhirLib.EngineTypes.next = _menhir_stack;
              };
            };
          };
        } = _menhir_stack in
        let _4 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 5258 "compiler/lib/js_parser.ml"
        ) = Obj.magic _4 in
        let e2 : (Javascript.expression) = Obj.magic e2 in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 5264 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let _1 : (Parse_info.t * Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__4_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 375 "compiler/lib/js_parser.mly"
     ( let (start, e1) = _1 in (start, EAccess (e1,e2)) )
# 5273 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = i;
          MenhirLib.EngineTypes.startp = _startpos_i_;
          MenhirLib.EngineTypes.endp = _endpos_i_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = _1;
              MenhirLib.EngineTypes.startp = _startpos__1_;
              MenhirLib.EngineTypes.endp = _endpos__1_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let i : (string) = Obj.magic i in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 5307 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let _1 : (Parse_info.t * Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_i_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 377 "compiler/lib/js_parser.mly"
     ( let (start, e1) = _1 in (start, EDot(e1,i)) )
# 5316 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = a;
          MenhirLib.EngineTypes.startp = _startpos_a_;
          MenhirLib.EngineTypes.endp = _endpos_a_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = e1;
            MenhirLib.EngineTypes.startp = _startpos_e1_;
            MenhirLib.EngineTypes.endp = _endpos_e1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = pi;
              MenhirLib.EngineTypes.startp = _startpos_pi_;
              MenhirLib.EngineTypes.endp = _endpos_pi_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let a : (Javascript.arguments) = Obj.magic a in
        let e1 : (Parse_info.t * Javascript.expression) = Obj.magic e1 in
        let pi : (
# 61 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 5351 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos_a_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 379 "compiler/lib/js_parser.mly"
     ( (pi, ENew(snd e1, Some a)) )
# 5359 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let e : (Parse_info.t * Javascript.expression) = Obj.magic e in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_e_ in
        let _endpos = _endpos_e_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 476 "compiler/lib/js_parser.mly"
   ( e )
# 5384 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _4;
          MenhirLib.EngineTypes.startp = _startpos__4_;
          MenhirLib.EngineTypes.endp = _endpos__4_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = e2;
            MenhirLib.EngineTypes.startp = _startpos_e2_;
            MenhirLib.EngineTypes.endp = _endpos_e2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = _2;
              MenhirLib.EngineTypes.startp = _startpos__2_;
              MenhirLib.EngineTypes.endp = _endpos__2_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _menhir_s;
                MenhirLib.EngineTypes.semv = _1;
                MenhirLib.EngineTypes.startp = _startpos__1_;
                MenhirLib.EngineTypes.endp = _endpos__1_;
                MenhirLib.EngineTypes.next = _menhir_stack;
              };
            };
          };
        } = _menhir_stack in
        let _4 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 5423 "compiler/lib/js_parser.ml"
        ) = Obj.magic _4 in
        let e2 : (Javascript.expression) = Obj.magic e2 in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 5429 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let _1 : (Parse_info.t * Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__4_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 478 "compiler/lib/js_parser.mly"
   ( let (start, e1) = _1 in (start, EAccess(e1, e2)) )
# 5438 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = i;
          MenhirLib.EngineTypes.startp = _startpos_i_;
          MenhirLib.EngineTypes.endp = _endpos_i_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = _1;
              MenhirLib.EngineTypes.startp = _startpos__1_;
              MenhirLib.EngineTypes.endp = _endpos__1_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let i : (string) = Obj.magic i in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 5472 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let _1 : (Parse_info.t * Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_i_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 480 "compiler/lib/js_parser.mly"
   ( let (start, e1) = _1 in (start, EDot(e1,i)) )
# 5481 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = a;
          MenhirLib.EngineTypes.startp = _startpos_a_;
          MenhirLib.EngineTypes.endp = _endpos_a_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = e;
            MenhirLib.EngineTypes.startp = _startpos_e_;
            MenhirLib.EngineTypes.endp = _endpos_e_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = pi;
              MenhirLib.EngineTypes.startp = _startpos_pi_;
              MenhirLib.EngineTypes.endp = _endpos_pi_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let a : (Javascript.arguments) = Obj.magic a in
        let e : (Parse_info.t * Javascript.expression) = Obj.magic e in
        let pi : (
# 61 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 5516 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos_a_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 482 "compiler/lib/js_parser.mly"
   ( (pi, ENew(snd e,Some a)) )
# 5524 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Parse_info.t * Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 368 "compiler/lib/js_parser.mly"
                        ( _1 )
# 5549 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _2;
          MenhirLib.EngineTypes.startp = _startpos__2_;
          MenhirLib.EngineTypes.endp = _endpos__2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = pi;
            MenhirLib.EngineTypes.startp = _startpos_pi_;
            MenhirLib.EngineTypes.endp = _endpos_pi_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let _2 : (Parse_info.t * Javascript.expression) = Obj.magic _2 in
        let pi : (
# 61 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 5577 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos__2_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 369 "compiler/lib/js_parser.mly"
                           ( (pi, ENew (snd _2,None)) )
# 5585 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Parse_info.t * Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 461 "compiler/lib/js_parser.mly"
                                  ( _1 )
# 5610 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _2;
          MenhirLib.EngineTypes.startp = _startpos__2_;
          MenhirLib.EngineTypes.endp = _endpos__2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = pi;
            MenhirLib.EngineTypes.startp = _startpos_pi_;
            MenhirLib.EngineTypes.endp = _endpos_pi_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let _2 : (Parse_info.t * Javascript.expression) = Obj.magic _2 in
        let pi : (
# 61 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 5638 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos__2_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 462 "compiler/lib/js_parser.mly"
                           ( (pi, ENew (snd _2,None)) )
# 5646 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = pi;
          MenhirLib.EngineTypes.startp = _startpos_pi_;
          MenhirLib.EngineTypes.endp = _endpos_pi_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 5667 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos_pi_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 489 "compiler/lib/js_parser.mly"
             ( (pi, EVar (var pi "null")) )
# 5675 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 47 "compiler/lib/js_parser.mly"
      (string * Parse_info.t)
# 5696 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Parse_info.t * string) = 
# 496 "compiler/lib/js_parser.mly"
            ( let (f, pi) = _1 in (pi, f) )
# 5704 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = y;
          MenhirLib.EngineTypes.startp = _startpos_y_;
          MenhirLib.EngineTypes.endp = _endpos_y_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = x;
              MenhirLib.EngineTypes.startp = _startpos_x_;
              MenhirLib.EngineTypes.endp = _endpos_x_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let y : (Javascript.expression) = Obj.magic y in
        let _2 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 5738 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let x : (Javascript.property_name) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_y_ in
        let _v =
          let pair = 
# 172 "<standard.mly>"
    ( (x, y) )
# 5748 "compiler/lib/js_parser.ml"
           in
          (
# 540 "compiler/lib/js_parser.mly"
                                                                      ( pair )
# 5753 "compiler/lib/js_parser.ml"
           : (Javascript.property_name * Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = block;
          MenhirLib.EngineTypes.startp = _startpos_block_;
          MenhirLib.EngineTypes.endp = _endpos_block_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let block : (unit * Parse_info.t * Parse_info.t) = Obj.magic block in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_block_ in
        let _endpos = _endpos_block_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 535 "compiler/lib/js_parser.mly"
   ( let _pairs, pi_start, _pi_end = block in pi_start, EObj [] )
# 5779 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = block;
          MenhirLib.EngineTypes.startp = _startpos_block_;
          MenhirLib.EngineTypes.endp = _endpos_block_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let block : (Javascript.property_name_and_value_list * Parse_info.t * Parse_info.t) = Obj.magic block in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_block_ in
        let _endpos = _endpos_block_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 537 "compiler/lib/js_parser.mly"
   ( let pairs, pi_start, _pi_end = block in pi_start, EObj pairs )
# 5804 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let _menhir_s = _menhir_env.MenhirLib.EngineTypes.current in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _endpos = _startpos in
        let _v : (Javascript.expression option) = 
# 111 "<standard.mly>"
    ( None )
# 5822 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = x;
          MenhirLib.EngineTypes.startp = _startpos_x_;
          MenhirLib.EngineTypes.endp = _endpos_x_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let x : (Javascript.expression) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_x_ in
        let _v : (Javascript.expression option) = 
# 113 "<standard.mly>"
    ( Some x )
# 5847 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let _menhir_s = _menhir_env.MenhirLib.EngineTypes.current in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _endpos = _startpos in
        let _v : (Javascript.expression option) = 
# 111 "<standard.mly>"
    ( None )
# 5865 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = x;
          MenhirLib.EngineTypes.startp = _startpos_x_;
          MenhirLib.EngineTypes.endp = _endpos_x_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let x : (Javascript.expression) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_x_ in
        let _v : (Javascript.expression option) = 
# 113 "<standard.mly>"
    ( Some x )
# 5890 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let _menhir_s = _menhir_env.MenhirLib.EngineTypes.current in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _endpos = _startpos in
        let _v : (Javascript.statement_list option) = 
# 111 "<standard.mly>"
    ( None )
# 5908 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = x;
          MenhirLib.EngineTypes.startp = _startpos_x_;
          MenhirLib.EngineTypes.endp = _endpos_x_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let x : (Javascript.statement_list) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_x_ in
        let _v : (Javascript.statement_list option) = 
# 113 "<standard.mly>"
    ( Some x )
# 5933 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let _menhir_s = _menhir_env.MenhirLib.EngineTypes.current in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _endpos = _startpos in
        let _v : (Javascript.initialiser option) = 
# 111 "<standard.mly>"
    ( None )
# 5951 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = x;
          MenhirLib.EngineTypes.startp = _startpos_x_;
          MenhirLib.EngineTypes.endp = _endpos_x_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let x : (Javascript.initialiser) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_x_ in
        let _v : (Javascript.initialiser option) = 
# 113 "<standard.mly>"
    ( Some x )
# 5976 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let _menhir_s = _menhir_env.MenhirLib.EngineTypes.current in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _endpos = _startpos in
        let _v : (Javascript.initialiser option) = 
# 111 "<standard.mly>"
    ( None )
# 5994 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = x;
          MenhirLib.EngineTypes.startp = _startpos_x_;
          MenhirLib.EngineTypes.endp = _endpos_x_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let x : (Javascript.initialiser) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_x_ in
        let _v : (Javascript.initialiser option) = 
# 113 "<standard.mly>"
    ( Some x )
# 6019 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let _menhir_s = _menhir_env.MenhirLib.EngineTypes.current in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _endpos = _startpos in
        let _v : ((Javascript.Label.t * Parse_info.t) option) = 
# 111 "<standard.mly>"
    ( None )
# 6037 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = x;
          MenhirLib.EngineTypes.startp = _startpos_x_;
          MenhirLib.EngineTypes.endp = _endpos_x_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let x : (Javascript.Label.t * Parse_info.t) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_x_ in
        let _v : ((Javascript.Label.t * Parse_info.t) option) = 
# 113 "<standard.mly>"
    ( Some x )
# 6062 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let _menhir_s = _menhir_env.MenhirLib.EngineTypes.current in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _endpos = _startpos in
        let _v : ((Javascript.statement_list * Javascript.case_clause list) option) = 
# 111 "<standard.mly>"
    ( None )
# 6080 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = y;
          MenhirLib.EngineTypes.startp = _startpos_y_;
          MenhirLib.EngineTypes.endp = _endpos_y_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = x;
            MenhirLib.EngineTypes.startp = _startpos_x_;
            MenhirLib.EngineTypes.endp = _endpos_x_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let y : (Javascript.case_clause list) = Obj.magic y in
        let x : (Javascript.statement_list) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_y_ in
        let _v =
          let x = 
# 164 "<standard.mly>"
    ( (x, y) )
# 6113 "compiler/lib/js_parser.ml"
           in
          (
# 113 "<standard.mly>"
    ( Some x )
# 6118 "compiler/lib/js_parser.ml"
           : ((Javascript.statement_list * Javascript.case_clause list) option))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let _menhir_s = _menhir_env.MenhirLib.EngineTypes.current in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _endpos = _startpos in
        let _v : (Javascript.ident option) = 
# 111 "<standard.mly>"
    ( None )
# 6137 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = x;
          MenhirLib.EngineTypes.startp = _startpos_x_;
          MenhirLib.EngineTypes.endp = _endpos_x_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let x : (Javascript.ident) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_x_ in
        let _v : (Javascript.ident option) = 
# 113 "<standard.mly>"
    ( Some x )
# 6162 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.expression) = 
# 340 "compiler/lib/js_parser.mly"
                     ( _1 )
# 6187 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 6221 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 631 "compiler/lib/js_parser.mly"
                        ( Lt         )
# 6232 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 6237 "compiler/lib/js_parser.ml"
            
          in
          (
# 344 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 6243 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 6278 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 632 "compiler/lib/js_parser.mly"
                        ( Gt         )
# 6289 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 6294 "compiler/lib/js_parser.ml"
            
          in
          (
# 344 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 6300 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 6335 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 633 "compiler/lib/js_parser.mly"
                        ( Le         )
# 6346 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 6351 "compiler/lib/js_parser.ml"
            
          in
          (
# 344 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 6357 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 6392 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 634 "compiler/lib/js_parser.mly"
                        ( Ge         )
# 6403 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 6408 "compiler/lib/js_parser.ml"
            
          in
          (
# 344 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 6414 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 6449 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 635 "compiler/lib/js_parser.mly"
                        ( InstanceOf )
# 6460 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 6465 "compiler/lib/js_parser.ml"
            
          in
          (
# 344 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 6471 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 6506 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 636 "compiler/lib/js_parser.mly"
                        ( EqEq       )
# 6517 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 6522 "compiler/lib/js_parser.ml"
            
          in
          (
# 344 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 6528 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 6563 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 637 "compiler/lib/js_parser.mly"
                        ( NotEq      )
# 6574 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 6579 "compiler/lib/js_parser.ml"
            
          in
          (
# 344 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 6585 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 6620 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 638 "compiler/lib/js_parser.mly"
                        ( EqEqEq     )
# 6631 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 6636 "compiler/lib/js_parser.ml"
            
          in
          (
# 344 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 6642 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 6677 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 639 "compiler/lib/js_parser.mly"
                        ( NotEqEq    )
# 6688 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 6693 "compiler/lib/js_parser.ml"
            
          in
          (
# 344 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 6699 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 6734 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 640 "compiler/lib/js_parser.mly"
                        ( Band       )
# 6745 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 6750 "compiler/lib/js_parser.ml"
            
          in
          (
# 344 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 6756 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 6791 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 641 "compiler/lib/js_parser.mly"
                        ( Bxor       )
# 6802 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 6807 "compiler/lib/js_parser.ml"
            
          in
          (
# 344 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 6813 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 6848 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 642 "compiler/lib/js_parser.mly"
                        ( Bor        )
# 6859 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 6864 "compiler/lib/js_parser.ml"
            
          in
          (
# 344 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 6870 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 6905 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 643 "compiler/lib/js_parser.mly"
                        ( And        )
# 6916 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 6921 "compiler/lib/js_parser.ml"
            
          in
          (
# 344 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 6927 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 6962 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 644 "compiler/lib/js_parser.mly"
                        ( Or         )
# 6973 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 6978 "compiler/lib/js_parser.ml"
            
          in
          (
# 344 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 6984 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 7019 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 648 "compiler/lib/js_parser.mly"
        ( In )
# 7029 "compiler/lib/js_parser.ml"
           in
          (
# 344 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 7034 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.expression) = 
# 415 "compiler/lib/js_parser.mly"
                     ( _1 )
# 7060 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 7094 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 631 "compiler/lib/js_parser.mly"
                        ( Lt         )
# 7104 "compiler/lib/js_parser.ml"
           in
          (
# 419 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 7109 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 7144 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 632 "compiler/lib/js_parser.mly"
                        ( Gt         )
# 7154 "compiler/lib/js_parser.ml"
           in
          (
# 419 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 7159 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 7194 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 633 "compiler/lib/js_parser.mly"
                        ( Le         )
# 7204 "compiler/lib/js_parser.ml"
           in
          (
# 419 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 7209 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 7244 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 634 "compiler/lib/js_parser.mly"
                        ( Ge         )
# 7254 "compiler/lib/js_parser.ml"
           in
          (
# 419 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 7259 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 7294 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 635 "compiler/lib/js_parser.mly"
                        ( InstanceOf )
# 7304 "compiler/lib/js_parser.ml"
           in
          (
# 419 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 7309 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 7344 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 636 "compiler/lib/js_parser.mly"
                        ( EqEq       )
# 7354 "compiler/lib/js_parser.ml"
           in
          (
# 419 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 7359 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 7394 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 637 "compiler/lib/js_parser.mly"
                        ( NotEq      )
# 7404 "compiler/lib/js_parser.ml"
           in
          (
# 419 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 7409 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 7444 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 638 "compiler/lib/js_parser.mly"
                        ( EqEqEq     )
# 7454 "compiler/lib/js_parser.ml"
           in
          (
# 419 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 7459 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 7494 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 639 "compiler/lib/js_parser.mly"
                        ( NotEqEq    )
# 7504 "compiler/lib/js_parser.ml"
           in
          (
# 419 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 7509 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 7544 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 640 "compiler/lib/js_parser.mly"
                        ( Band       )
# 7554 "compiler/lib/js_parser.ml"
           in
          (
# 419 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 7559 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 7594 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 641 "compiler/lib/js_parser.mly"
                        ( Bxor       )
# 7604 "compiler/lib/js_parser.ml"
           in
          (
# 419 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 7609 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 7644 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 642 "compiler/lib/js_parser.mly"
                        ( Bor        )
# 7654 "compiler/lib/js_parser.ml"
           in
          (
# 419 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 7659 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 7694 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 643 "compiler/lib/js_parser.mly"
                        ( And        )
# 7704 "compiler/lib/js_parser.ml"
           in
          (
# 419 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 7709 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 7744 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 644 "compiler/lib/js_parser.mly"
                        ( Or         )
# 7754 "compiler/lib/js_parser.ml"
           in
          (
# 419 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 7759 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.expression) = 
# 439 "compiler/lib/js_parser.mly"
                                  ( _1 )
# 7785 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 7819 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 631 "compiler/lib/js_parser.mly"
                        ( Lt         )
# 7830 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 7835 "compiler/lib/js_parser.ml"
            
          in
          (
# 443 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 7841 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 7876 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 632 "compiler/lib/js_parser.mly"
                        ( Gt         )
# 7887 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 7892 "compiler/lib/js_parser.ml"
            
          in
          (
# 443 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 7898 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 7933 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 633 "compiler/lib/js_parser.mly"
                        ( Le         )
# 7944 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 7949 "compiler/lib/js_parser.ml"
            
          in
          (
# 443 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 7955 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 7990 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 634 "compiler/lib/js_parser.mly"
                        ( Ge         )
# 8001 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 8006 "compiler/lib/js_parser.ml"
            
          in
          (
# 443 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 8012 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 8047 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 635 "compiler/lib/js_parser.mly"
                        ( InstanceOf )
# 8058 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 8063 "compiler/lib/js_parser.ml"
            
          in
          (
# 443 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 8069 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 8104 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 636 "compiler/lib/js_parser.mly"
                        ( EqEq       )
# 8115 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 8120 "compiler/lib/js_parser.ml"
            
          in
          (
# 443 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 8126 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 8161 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 637 "compiler/lib/js_parser.mly"
                        ( NotEq      )
# 8172 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 8177 "compiler/lib/js_parser.ml"
            
          in
          (
# 443 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 8183 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 8218 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 638 "compiler/lib/js_parser.mly"
                        ( EqEqEq     )
# 8229 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 8234 "compiler/lib/js_parser.ml"
            
          in
          (
# 443 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 8240 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 8275 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 639 "compiler/lib/js_parser.mly"
                        ( NotEqEq    )
# 8286 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 8291 "compiler/lib/js_parser.ml"
            
          in
          (
# 443 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 8297 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 8332 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 640 "compiler/lib/js_parser.mly"
                        ( Band       )
# 8343 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 8348 "compiler/lib/js_parser.ml"
            
          in
          (
# 443 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 8354 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 8389 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 641 "compiler/lib/js_parser.mly"
                        ( Bxor       )
# 8400 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 8405 "compiler/lib/js_parser.ml"
            
          in
          (
# 443 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 8411 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 8446 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 642 "compiler/lib/js_parser.mly"
                        ( Bor        )
# 8457 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 8462 "compiler/lib/js_parser.ml"
            
          in
          (
# 443 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 8468 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 8503 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 643 "compiler/lib/js_parser.mly"
                        ( And        )
# 8514 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 8519 "compiler/lib/js_parser.ml"
            
          in
          (
# 443 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 8525 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 8560 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op =
            let op = 
# 644 "compiler/lib/js_parser.mly"
                        ( Or         )
# 8571 "compiler/lib/js_parser.ml"
             in
            
# 647 "compiler/lib/js_parser.mly"
                                                      ( op )
# 8576 "compiler/lib/js_parser.ml"
            
          in
          (
# 443 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 8582 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 8617 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 648 "compiler/lib/js_parser.mly"
        ( In )
# 8627 "compiler/lib/js_parser.ml"
           in
          (
# 443 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 8632 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 8654 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.unop) = 
# 674 "compiler/lib/js_parser.mly"
             ( IncrA )
# 8662 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 8683 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.unop) = 
# 675 "compiler/lib/js_parser.mly"
             ( DecrA )
# 8691 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.expression) = 
# 348 "compiler/lib/js_parser.mly"
   ( _1 )
# 8716 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = op;
          MenhirLib.EngineTypes.startp = _startpos_op_;
          MenhirLib.EngineTypes.endp = _endpos_op_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = e;
            MenhirLib.EngineTypes.startp = _startpos_e_;
            MenhirLib.EngineTypes.endp = _endpos_e_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let op : (Javascript.unop) = Obj.magic op in
        let e : (Javascript.expression) = Obj.magic e in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_e_ in
        let _endpos = _endpos_op_ in
        let _v : (Javascript.expression) = 
# 351 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 8748 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 8776 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_e_ in
        let _v =
          let op = 
# 661 "compiler/lib/js_parser.mly"
             ( Delete )
# 8785 "compiler/lib/js_parser.ml"
           in
          (
# 351 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 8790 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 8819 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_e_ in
        let _v =
          let op = 
# 662 "compiler/lib/js_parser.mly"
             ( Void   )
# 8828 "compiler/lib/js_parser.ml"
           in
          (
# 351 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 8833 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 8862 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_e_ in
        let _v =
          let op = 
# 663 "compiler/lib/js_parser.mly"
             ( Typeof )
# 8871 "compiler/lib/js_parser.ml"
           in
          (
# 351 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 8876 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 8905 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_e_ in
        let _v =
          let op = 
# 664 "compiler/lib/js_parser.mly"
             ( IncrB  )
# 8914 "compiler/lib/js_parser.ml"
           in
          (
# 351 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 8919 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 8948 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_e_ in
        let _v =
          let op = 
# 665 "compiler/lib/js_parser.mly"
             ( IncrB  )
# 8957 "compiler/lib/js_parser.ml"
           in
          (
# 351 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 8962 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 8991 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_e_ in
        let _v =
          let op = 
# 666 "compiler/lib/js_parser.mly"
             ( DecrB  )
# 9000 "compiler/lib/js_parser.ml"
           in
          (
# 351 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 9005 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 9034 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_e_ in
        let _v =
          let op = 
# 667 "compiler/lib/js_parser.mly"
             ( DecrB  )
# 9043 "compiler/lib/js_parser.ml"
           in
          (
# 351 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 9048 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 9077 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_e_ in
        let _v =
          let op = 
# 668 "compiler/lib/js_parser.mly"
             ( Pl     )
# 9086 "compiler/lib/js_parser.ml"
           in
          (
# 351 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 9091 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 9120 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_e_ in
        let _v =
          let op = 
# 669 "compiler/lib/js_parser.mly"
             ( Neg    )
# 9129 "compiler/lib/js_parser.ml"
           in
          (
# 351 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 9134 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 9163 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_e_ in
        let _v =
          let op = 
# 670 "compiler/lib/js_parser.mly"
             ( Bnot   )
# 9172 "compiler/lib/js_parser.ml"
           in
          (
# 351 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 9177 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 9206 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_e_ in
        let _v =
          let op = 
# 671 "compiler/lib/js_parser.mly"
             ( Not    )
# 9215 "compiler/lib/js_parser.ml"
           in
          (
# 351 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 9220 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 9255 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 651 "compiler/lib/js_parser.mly"
             ( Mul   )
# 9265 "compiler/lib/js_parser.ml"
           in
          (
# 355 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 9270 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 9305 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 652 "compiler/lib/js_parser.mly"
             ( Div   )
# 9315 "compiler/lib/js_parser.ml"
           in
          (
# 355 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 9320 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 9355 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 653 "compiler/lib/js_parser.mly"
             ( Mod   )
# 9365 "compiler/lib/js_parser.ml"
           in
          (
# 355 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 9370 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 9405 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 654 "compiler/lib/js_parser.mly"
             ( Plus  )
# 9415 "compiler/lib/js_parser.ml"
           in
          (
# 355 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 9420 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 9455 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 655 "compiler/lib/js_parser.mly"
             ( Minus )
# 9465 "compiler/lib/js_parser.ml"
           in
          (
# 355 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 9470 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 9505 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 656 "compiler/lib/js_parser.mly"
             ( Lsl   )
# 9515 "compiler/lib/js_parser.ml"
           in
          (
# 355 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 9520 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 9555 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 657 "compiler/lib/js_parser.mly"
             ( Asr   )
# 9565 "compiler/lib/js_parser.ml"
           in
          (
# 355 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 9570 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 9605 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 658 "compiler/lib/js_parser.mly"
             ( Lsr   )
# 9615 "compiler/lib/js_parser.ml"
           in
          (
# 355 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 9620 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.expression) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.expression) = 
# 447 "compiler/lib/js_parser.mly"
   ( _1 )
# 9646 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = op;
          MenhirLib.EngineTypes.startp = _startpos_op_;
          MenhirLib.EngineTypes.endp = _endpos_op_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = e;
            MenhirLib.EngineTypes.startp = _startpos_e_;
            MenhirLib.EngineTypes.endp = _endpos_e_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let op : (Javascript.unop) = Obj.magic op in
        let e : (Javascript.expression) = Obj.magic e in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_e_ in
        let _endpos = _endpos_op_ in
        let _v : (Javascript.expression) = 
# 450 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 9678 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 9706 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_e_ in
        let _v =
          let op = 
# 661 "compiler/lib/js_parser.mly"
             ( Delete )
# 9715 "compiler/lib/js_parser.ml"
           in
          (
# 450 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 9720 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 9749 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_e_ in
        let _v =
          let op = 
# 662 "compiler/lib/js_parser.mly"
             ( Void   )
# 9758 "compiler/lib/js_parser.ml"
           in
          (
# 450 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 9763 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 9792 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_e_ in
        let _v =
          let op = 
# 663 "compiler/lib/js_parser.mly"
             ( Typeof )
# 9801 "compiler/lib/js_parser.ml"
           in
          (
# 450 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 9806 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 9835 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_e_ in
        let _v =
          let op = 
# 664 "compiler/lib/js_parser.mly"
             ( IncrB  )
# 9844 "compiler/lib/js_parser.ml"
           in
          (
# 450 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 9849 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 9878 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_e_ in
        let _v =
          let op = 
# 665 "compiler/lib/js_parser.mly"
             ( IncrB  )
# 9887 "compiler/lib/js_parser.ml"
           in
          (
# 450 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 9892 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 9921 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_e_ in
        let _v =
          let op = 
# 666 "compiler/lib/js_parser.mly"
             ( DecrB  )
# 9930 "compiler/lib/js_parser.ml"
           in
          (
# 450 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 9935 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 9964 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_e_ in
        let _v =
          let op = 
# 667 "compiler/lib/js_parser.mly"
             ( DecrB  )
# 9973 "compiler/lib/js_parser.ml"
           in
          (
# 450 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 9978 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 10007 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_e_ in
        let _v =
          let op = 
# 668 "compiler/lib/js_parser.mly"
             ( Pl     )
# 10016 "compiler/lib/js_parser.ml"
           in
          (
# 450 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 10021 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 10050 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_e_ in
        let _v =
          let op = 
# 669 "compiler/lib/js_parser.mly"
             ( Neg    )
# 10059 "compiler/lib/js_parser.ml"
           in
          (
# 450 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 10064 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 10093 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_e_ in
        let _v =
          let op = 
# 670 "compiler/lib/js_parser.mly"
             ( Bnot   )
# 10102 "compiler/lib/js_parser.ml"
           in
          (
# 450 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 10107 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let e : (Javascript.expression) = Obj.magic e in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 10136 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos_e_ in
        let _v =
          let op = 
# 671 "compiler/lib/js_parser.mly"
             ( Not    )
# 10145 "compiler/lib/js_parser.ml"
           in
          (
# 450 "compiler/lib/js_parser.mly"
   ( EUn (op, e) )
# 10150 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 10185 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 651 "compiler/lib/js_parser.mly"
             ( Mul   )
# 10195 "compiler/lib/js_parser.ml"
           in
          (
# 454 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 10200 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 10235 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 652 "compiler/lib/js_parser.mly"
             ( Div   )
# 10245 "compiler/lib/js_parser.ml"
           in
          (
# 454 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 10250 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 10285 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 653 "compiler/lib/js_parser.mly"
             ( Mod   )
# 10295 "compiler/lib/js_parser.ml"
           in
          (
# 454 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 10300 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 10335 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 654 "compiler/lib/js_parser.mly"
             ( Plus  )
# 10345 "compiler/lib/js_parser.ml"
           in
          (
# 454 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 10350 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 10385 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 655 "compiler/lib/js_parser.mly"
             ( Minus )
# 10395 "compiler/lib/js_parser.ml"
           in
          (
# 454 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 10400 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 10435 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 656 "compiler/lib/js_parser.mly"
             ( Lsl   )
# 10445 "compiler/lib/js_parser.ml"
           in
          (
# 454 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 10450 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 10485 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 657 "compiler/lib/js_parser.mly"
             ( Asr   )
# 10495 "compiler/lib/js_parser.ml"
           in
          (
# 454 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 10500 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = right;
          MenhirLib.EngineTypes.startp = _startpos_right_;
          MenhirLib.EngineTypes.endp = _endpos_right_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _1;
            MenhirLib.EngineTypes.startp = _startpos__1_;
            MenhirLib.EngineTypes.endp = _endpos__1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = left;
              MenhirLib.EngineTypes.startp = _startpos_left_;
              MenhirLib.EngineTypes.endp = _endpos_left_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let right : (Javascript.expression) = Obj.magic right in
        let _1 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 10535 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let left : (Javascript.expression) = Obj.magic left in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_left_ in
        let _endpos = _endpos_right_ in
        let _v =
          let op = 
# 658 "compiler/lib/js_parser.mly"
             ( Lsr   )
# 10545 "compiler/lib/js_parser.ml"
           in
          (
# 454 "compiler/lib/js_parser.mly"
   ( EBin (op, left, right) )
# 10550 "compiler/lib/js_parser.ml"
           : (Javascript.expression))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let e : (Parse_info.t * Javascript.expression) = Obj.magic e in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_e_ in
        let _endpos = _endpos_e_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 384 "compiler/lib/js_parser.mly"
                         ( e )
# 10576 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let e : (Parse_info.t * Javascript.expression) = Obj.magic e in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_e_ in
        let _endpos = _endpos_e_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 384 "compiler/lib/js_parser.mly"
                         ( e )
# 10601 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = e;
          MenhirLib.EngineTypes.startp = _startpos_e_;
          MenhirLib.EngineTypes.endp = _endpos_e_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let e : (Parse_info.t * Javascript.expression) = Obj.magic e in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_e_ in
        let _endpos = _endpos_e_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 384 "compiler/lib/js_parser.mly"
                         ( e )
# 10626 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = pi;
          MenhirLib.EngineTypes.startp = _startpos_pi_;
          MenhirLib.EngineTypes.endp = _endpos_pi_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 10647 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos_pi_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 387 "compiler/lib/js_parser.mly"
                     ( (pi, EVar (var pi "this")) )
# 10655 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.ident * Parse_info.t) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 388 "compiler/lib/js_parser.mly"
                     ( let (i, pi) = _1 in (pi, EVar i) )
# 10680 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = n;
          MenhirLib.EngineTypes.startp = _startpos_n_;
          MenhirLib.EngineTypes.endp = _endpos_n_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let n : (Parse_info.t * Javascript.expression) = Obj.magic n in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_n_ in
        let _endpos = _endpos_n_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 389 "compiler/lib/js_parser.mly"
                     ( n )
# 10705 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = b;
          MenhirLib.EngineTypes.startp = _startpos_b_;
          MenhirLib.EngineTypes.endp = _endpos_b_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let b : (Parse_info.t * Javascript.expression) = Obj.magic b in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_b_ in
        let _endpos = _endpos_b_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 390 "compiler/lib/js_parser.mly"
                     ( b )
# 10730 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Parse_info.t * string) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 391 "compiler/lib/js_parser.mly"
                     ( let (start, n) = _1 in (start, ENum (Num.of_string_unsafe n)) )
# 10755 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 49 "compiler/lib/js_parser.mly"
      (string * Parse_info.t * int)
# 10776 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 392 "compiler/lib/js_parser.mly"
                     ( let (s, start, _len) = _1 in (start, EStr (s, `Utf8)) )
# 10784 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = r;
          MenhirLib.EngineTypes.startp = _startpos_r_;
          MenhirLib.EngineTypes.endp = _endpos_r_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let r : (Parse_info.t * Javascript.expression) = Obj.magic r in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_r_ in
        let _endpos = _endpos_r_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 393 "compiler/lib/js_parser.mly"
                                  ( r )
# 10809 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = a;
          MenhirLib.EngineTypes.startp = _startpos_a_;
          MenhirLib.EngineTypes.endp = _endpos_a_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let a : (Parse_info.t * Javascript.expression) = Obj.magic a in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_a_ in
        let _endpos = _endpos_a_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 394 "compiler/lib/js_parser.mly"
                                  ( a )
# 10834 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _3;
          MenhirLib.EngineTypes.startp = _startpos__3_;
          MenhirLib.EngineTypes.endp = _endpos__3_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = e;
            MenhirLib.EngineTypes.startp = _startpos_e_;
            MenhirLib.EngineTypes.endp = _endpos_e_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = pi;
              MenhirLib.EngineTypes.startp = _startpos_pi_;
              MenhirLib.EngineTypes.endp = _endpos_pi_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let _3 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 10867 "compiler/lib/js_parser.ml"
        ) = Obj.magic _3 in
        let e : (Javascript.expression) = Obj.magic e in
        let pi : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 10873 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos__3_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 395 "compiler/lib/js_parser.mly"
                                     ( (pi, e) )
# 10881 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _2;
          MenhirLib.EngineTypes.startp = _startpos__2_;
          MenhirLib.EngineTypes.endp = _endpos__2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = l;
            MenhirLib.EngineTypes.startp = _startpos_l_;
            MenhirLib.EngineTypes.endp = _endpos_l_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let _2 : (
# 106 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 10908 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let l : (Javascript.program_with_annots) = Obj.magic l in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_l_ in
        let _endpos = _endpos__2_ in
        let _v : (Javascript.program_with_annots) = 
# 144 "compiler/lib/js_parser.mly"
                                    ( l )
# 10917 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = i;
          MenhirLib.EngineTypes.startp = _startpos_i_;
          MenhirLib.EngineTypes.endp = _endpos_i_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let i : (string) = Obj.magic i in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_i_ in
        let _endpos = _endpos_i_ in
        let _v : (Javascript.property_name) = 
# 607 "compiler/lib/js_parser.mly"
                      ( PNI i )
# 10942 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = s;
          MenhirLib.EngineTypes.startp = _startpos_s_;
          MenhirLib.EngineTypes.endp = _endpos_s_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let s : (
# 49 "compiler/lib/js_parser.mly"
      (string * Parse_info.t * int)
# 10963 "compiler/lib/js_parser.ml"
        ) = Obj.magic s in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_s_ in
        let _endpos = _endpos_s_ in
        let _v : (Javascript.property_name) = 
# 608 "compiler/lib/js_parser.mly"
                      (
    let s, _info, _len = s in PNS s )
# 10972 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = n;
          MenhirLib.EngineTypes.startp = _startpos_n_;
          MenhirLib.EngineTypes.endp = _endpos_n_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let n : (Parse_info.t * string) = Obj.magic n in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_n_ in
        let _endpos = _endpos_n_ in
        let _v : (Javascript.property_name) = 
# 610 "compiler/lib/js_parser.mly"
                      ( PNN (Num.of_string_unsafe (snd n)) )
# 10997 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (
# 50 "compiler/lib/js_parser.mly"
      (string * Parse_info.t)
# 11018 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Parse_info.t * Javascript.expression) = 
# 499 "compiler/lib/js_parser.mly"
           (
   let (s, pi) = _1 in
   let len = String.length s in
   let regexp, option =
     if s.[len - 1] = '/'
     then String.sub s 1 (len - 2),None
     else
       let i = String.rindex s '/' in
       String.sub s 1 (i - 1),Some (String.sub s (i+1) (len - i - 1))
   in
   (pi, ERegexp (regexp, option)) )
# 11036 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _2;
          MenhirLib.EngineTypes.startp = _startpos__2_;
          MenhirLib.EngineTypes.endp = _endpos__2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = pi;
            MenhirLib.EngineTypes.startp = _startpos_pi_;
            MenhirLib.EngineTypes.endp = _endpos_pi_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let _2 : (Javascript.expression option) = Obj.magic _2 in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 11064 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos__2_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 259 "compiler/lib/js_parser.mly"
                           ( (Return_statement _2, Pi pi) )
# 11072 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = x;
          MenhirLib.EngineTypes.startp = _startpos_x_;
          MenhirLib.EngineTypes.endp = _endpos_x_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let x : (Javascript.expression * Javascript.spread) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_x_ in
        let _v : (Javascript.arguments) = 
# 238 "<standard.mly>"
    ( [ x ] )
# 11097 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = xs;
          MenhirLib.EngineTypes.startp = _startpos_xs_;
          MenhirLib.EngineTypes.endp = _endpos_xs_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = x;
              MenhirLib.EngineTypes.startp = _startpos_x_;
              MenhirLib.EngineTypes.endp = _endpos_x_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let xs : (Javascript.arguments) = Obj.magic xs in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 11131 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let x : (Javascript.expression * Javascript.spread) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_xs_ in
        let _v : (Javascript.arguments) = 
# 240 "<standard.mly>"
    ( x :: xs )
# 11140 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = y;
          MenhirLib.EngineTypes.startp = _startpos_y_;
          MenhirLib.EngineTypes.endp = _endpos_y_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = x;
            MenhirLib.EngineTypes.startp = _startpos_x_;
            MenhirLib.EngineTypes.endp = _endpos_x_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let y : (Javascript.initialiser option) = Obj.magic y in
        let x : (Javascript.ident) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_y_ in
        let _v =
          let x = 
# 164 "<standard.mly>"
    ( (x, y) )
# 11173 "compiler/lib/js_parser.ml"
           in
          (
# 238 "<standard.mly>"
    ( [ x ] )
# 11178 "compiler/lib/js_parser.ml"
           : (Javascript.variable_declaration list))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = xs;
          MenhirLib.EngineTypes.startp = _startpos_xs_;
          MenhirLib.EngineTypes.endp = _endpos_xs_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = y;
              MenhirLib.EngineTypes.startp = _startpos_y_;
              MenhirLib.EngineTypes.endp = _endpos_y_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _menhir_s;
                MenhirLib.EngineTypes.semv = x;
                MenhirLib.EngineTypes.startp = _startpos_x_;
                MenhirLib.EngineTypes.endp = _endpos_x_;
                MenhirLib.EngineTypes.next = _menhir_stack;
              };
            };
          };
        } = _menhir_stack in
        let xs : (Javascript.variable_declaration list) = Obj.magic xs in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 11219 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let y : (Javascript.initialiser option) = Obj.magic y in
        let x : (Javascript.ident) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_xs_ in
        let _v =
          let x = 
# 164 "<standard.mly>"
    ( (x, y) )
# 11230 "compiler/lib/js_parser.ml"
           in
          (
# 240 "<standard.mly>"
    ( x :: xs )
# 11235 "compiler/lib/js_parser.ml"
           : (Javascript.variable_declaration list))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = y;
          MenhirLib.EngineTypes.startp = _startpos_y_;
          MenhirLib.EngineTypes.endp = _endpos_y_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = x;
            MenhirLib.EngineTypes.startp = _startpos_x_;
            MenhirLib.EngineTypes.endp = _endpos_x_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let y : (Javascript.initialiser option) = Obj.magic y in
        let x : (Javascript.ident) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_y_ in
        let _v =
          let x = 
# 164 "<standard.mly>"
    ( (x, y) )
# 11269 "compiler/lib/js_parser.ml"
           in
          (
# 238 "<standard.mly>"
    ( [ x ] )
# 11274 "compiler/lib/js_parser.ml"
           : (Javascript.variable_declaration list))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = xs;
          MenhirLib.EngineTypes.startp = _startpos_xs_;
          MenhirLib.EngineTypes.endp = _endpos_xs_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = y;
              MenhirLib.EngineTypes.startp = _startpos_y_;
              MenhirLib.EngineTypes.endp = _endpos_y_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _menhir_s;
                MenhirLib.EngineTypes.semv = x;
                MenhirLib.EngineTypes.startp = _startpos_x_;
                MenhirLib.EngineTypes.endp = _endpos_x_;
                MenhirLib.EngineTypes.next = _menhir_stack;
              };
            };
          };
        } = _menhir_stack in
        let xs : (Javascript.variable_declaration list) = Obj.magic xs in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 11315 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let y : (Javascript.initialiser option) = Obj.magic y in
        let x : (Javascript.ident) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_xs_ in
        let _v =
          let x = 
# 164 "<standard.mly>"
    ( (x, y) )
# 11326 "compiler/lib/js_parser.ml"
           in
          (
# 240 "<standard.mly>"
    ( x :: xs )
# 11331 "compiler/lib/js_parser.ml"
           : (Javascript.variable_declaration list))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = x;
          MenhirLib.EngineTypes.startp = _startpos_x_;
          MenhirLib.EngineTypes.endp = _endpos_x_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let x : (Javascript.ident) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_x_ in
        let _v : (Javascript.formal_parameter_list) = 
# 238 "<standard.mly>"
    ( [ x ] )
# 11357 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = xs;
          MenhirLib.EngineTypes.startp = _startpos_xs_;
          MenhirLib.EngineTypes.endp = _endpos_xs_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = x;
              MenhirLib.EngineTypes.startp = _startpos_x_;
              MenhirLib.EngineTypes.endp = _endpos_x_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let xs : (Javascript.formal_parameter_list) = Obj.magic xs in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 11391 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let x : (Javascript.ident) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_xs_ in
        let _v : (Javascript.formal_parameter_list) = 
# 240 "<standard.mly>"
    ( x :: xs )
# 11400 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = x;
          MenhirLib.EngineTypes.startp = _startpos_x_;
          MenhirLib.EngineTypes.endp = _endpos_x_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let x : (Javascript.property_name * Javascript.expression) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_x_ in
        let _v : (Javascript.property_name_and_value_list) = 
# 700 "compiler/lib/js_parser.mly"
       ( [x] )
# 11425 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _2;
          MenhirLib.EngineTypes.startp = _startpos__2_;
          MenhirLib.EngineTypes.endp = _endpos__2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = x;
            MenhirLib.EngineTypes.startp = _startpos_x_;
            MenhirLib.EngineTypes.endp = _endpos_x_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 11452 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let x : (Javascript.property_name * Javascript.expression) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos__2_ in
        let _v : (Javascript.property_name_and_value_list) = 
# 701 "compiler/lib/js_parser.mly"
                 ( [x] )
# 11461 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = xs;
          MenhirLib.EngineTypes.startp = _startpos_xs_;
          MenhirLib.EngineTypes.endp = _endpos_xs_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = x;
              MenhirLib.EngineTypes.startp = _startpos_x_;
              MenhirLib.EngineTypes.endp = _endpos_x_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let xs : (Javascript.property_name_and_value_list) = Obj.magic xs in
        let _2 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 11495 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let x : (Javascript.property_name * Javascript.expression) = Obj.magic x in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_x_ in
        let _endpos = _endpos_xs_ in
        let _v : (Javascript.property_name_and_value_list) = 
# 702 "compiler/lib/js_parser.mly"
                                                               ( x :: xs )
# 11504 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.statement * Javascript.location) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.source_element * Javascript.location) = 
# 157 "compiler/lib/js_parser.mly"
   ( let statement, pi = _1 in Statement statement, pi )
# 11529 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = _1;
          MenhirLib.EngineTypes.startp = _startpos__1_;
          MenhirLib.EngineTypes.endp = _endpos__1_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let _1 : (Javascript.function_declaration * Javascript.location) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : (Javascript.source_element * Javascript.location) = 
# 159 "compiler/lib/js_parser.mly"
   ( let declaration, pi = _1 in Function_declaration declaration, pi )
# 11554 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = s;
          MenhirLib.EngineTypes.startp = _startpos_s_;
          MenhirLib.EngineTypes.endp = _endpos_s_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = annots;
            MenhirLib.EngineTypes.startp = _startpos_annots_;
            MenhirLib.EngineTypes.endp = _endpos_annots_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let s : (Javascript.source_element * Javascript.location) = Obj.magic s in
        let annots : (Js_token.Annot.t list) = Obj.magic annots in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_annots_ in
        let _endpos = _endpos_s_ in
        let _v : ((Javascript.source_element * Javascript.location) * Js_token.Annot.t list) = 
# 153 "compiler/lib/js_parser.mly"
                                  (s,annots)
# 11586 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _2;
          MenhirLib.EngineTypes.startp = _startpos__2_;
          MenhirLib.EngineTypes.endp = _endpos__2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = e;
            MenhirLib.EngineTypes.startp = _startpos_e_;
            MenhirLib.EngineTypes.endp = _endpos_e_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let _2 : (
# 106 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 11613 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let e : (Javascript.expression) = Obj.magic e in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_e_ in
        let _endpos = _endpos__2_ in
        let _v : (Javascript.expression) = 
# 147 "compiler/lib/js_parser.mly"
                    ( e )
# 11622 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = s;
          MenhirLib.EngineTypes.startp = _startpos_s_;
          MenhirLib.EngineTypes.endp = _endpos_s_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let s : (Javascript.statement * Javascript.location) = Obj.magic s in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_s_ in
        let _endpos = _endpos_s_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 190 "compiler/lib/js_parser.mly"
                       ( s )
# 11647 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _2;
          MenhirLib.EngineTypes.startp = _startpos__2_;
          MenhirLib.EngineTypes.endp = _endpos__2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = s;
            MenhirLib.EngineTypes.startp = _startpos_s_;
            MenhirLib.EngineTypes.endp = _endpos_s_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let _2 : (Parse_info.t) = Obj.magic _2 in
        let s : (Javascript.statement * Javascript.location) = Obj.magic s in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_s_ in
        let _endpos = _endpos__2_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 191 "compiler/lib/js_parser.mly"
                                                                  ( s )
# 11679 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = s;
          MenhirLib.EngineTypes.startp = _startpos_s_;
          MenhirLib.EngineTypes.endp = _endpos_s_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let s : (Javascript.statement * Javascript.location) = Obj.magic s in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_s_ in
        let _endpos = _endpos_s_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 187 "compiler/lib/js_parser.mly"
                        ( s )
# 11704 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = s;
          MenhirLib.EngineTypes.startp = _startpos_s_;
          MenhirLib.EngineTypes.endp = _endpos_s_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let s : (Javascript.statement * Javascript.location) = Obj.magic s in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_s_ in
        let _endpos = _endpos_s_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 187 "compiler/lib/js_parser.mly"
                        ( s )
# 11729 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = s;
          MenhirLib.EngineTypes.startp = _startpos_s_;
          MenhirLib.EngineTypes.endp = _endpos_s_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let s : (Javascript.statement * Javascript.location) = Obj.magic s in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_s_ in
        let _endpos = _endpos_s_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 187 "compiler/lib/js_parser.mly"
                        ( s )
# 11754 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = s;
          MenhirLib.EngineTypes.startp = _startpos_s_;
          MenhirLib.EngineTypes.endp = _endpos_s_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let s : (Javascript.statement * Javascript.location) = Obj.magic s in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_s_ in
        let _endpos = _endpos_s_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 187 "compiler/lib/js_parser.mly"
                        ( s )
# 11779 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = s;
          MenhirLib.EngineTypes.startp = _startpos_s_;
          MenhirLib.EngineTypes.endp = _endpos_s_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let s : (Javascript.statement * Javascript.location) = Obj.magic s in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_s_ in
        let _endpos = _endpos_s_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 187 "compiler/lib/js_parser.mly"
                        ( s )
# 11804 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = s;
          MenhirLib.EngineTypes.startp = _startpos_s_;
          MenhirLib.EngineTypes.endp = _endpos_s_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let s : (Javascript.statement * Javascript.location) = Obj.magic s in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_s_ in
        let _endpos = _endpos_s_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 187 "compiler/lib/js_parser.mly"
                        ( s )
# 11829 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = s;
          MenhirLib.EngineTypes.startp = _startpos_s_;
          MenhirLib.EngineTypes.endp = _endpos_s_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let s : (Javascript.statement * Javascript.location) = Obj.magic s in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_s_ in
        let _endpos = _endpos_s_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 187 "compiler/lib/js_parser.mly"
                        ( s )
# 11854 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = s;
          MenhirLib.EngineTypes.startp = _startpos_s_;
          MenhirLib.EngineTypes.endp = _endpos_s_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let s : (Javascript.statement * Javascript.location) = Obj.magic s in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_s_ in
        let _endpos = _endpos_s_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 187 "compiler/lib/js_parser.mly"
                        ( s )
# 11879 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = block;
          MenhirLib.EngineTypes.startp = _startpos_block_;
          MenhirLib.EngineTypes.endp = _endpos_block_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let block : (Javascript.statement_list * Parse_info.t * Parse_info.t) = Obj.magic block in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_block_ in
        let _endpos = _endpos_block_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 167 "compiler/lib/js_parser.mly"
   ( let statements, pi_start, _pi_end = block in
     Block statements, Pi pi_start )
# 11905 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = s;
          MenhirLib.EngineTypes.startp = _startpos_s_;
          MenhirLib.EngineTypes.endp = _endpos_s_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let s : (Javascript.statement * Javascript.location) = Obj.magic s in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_s_ in
        let _endpos = _endpos_s_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 177 "compiler/lib/js_parser.mly"
                     ( s )
# 11930 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = s;
          MenhirLib.EngineTypes.startp = _startpos_s_;
          MenhirLib.EngineTypes.endp = _endpos_s_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let s : (Javascript.statement * Javascript.location) = Obj.magic s in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_s_ in
        let _endpos = _endpos_s_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 177 "compiler/lib/js_parser.mly"
                     ( s )
# 11955 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = s;
          MenhirLib.EngineTypes.startp = _startpos_s_;
          MenhirLib.EngineTypes.endp = _endpos_s_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let s : (Javascript.statement * Javascript.location) = Obj.magic s in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_s_ in
        let _endpos = _endpos_s_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 177 "compiler/lib/js_parser.mly"
                     ( s )
# 11980 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = s;
          MenhirLib.EngineTypes.startp = _startpos_s_;
          MenhirLib.EngineTypes.endp = _endpos_s_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let s : (Javascript.statement * Javascript.location) = Obj.magic s in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_s_ in
        let _endpos = _endpos_s_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 177 "compiler/lib/js_parser.mly"
                     ( s )
# 12005 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = s;
          MenhirLib.EngineTypes.startp = _startpos_s_;
          MenhirLib.EngineTypes.endp = _endpos_s_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let s : (Javascript.statement * Javascript.location) = Obj.magic s in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_s_ in
        let _endpos = _endpos_s_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 177 "compiler/lib/js_parser.mly"
                     ( s )
# 12030 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = s;
          MenhirLib.EngineTypes.startp = _startpos_s_;
          MenhirLib.EngineTypes.endp = _endpos_s_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let s : (Javascript.statement * Javascript.location) = Obj.magic s in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_s_ in
        let _endpos = _endpos_s_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 177 "compiler/lib/js_parser.mly"
                     ( s )
# 12055 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = s;
          MenhirLib.EngineTypes.startp = _startpos_s_;
          MenhirLib.EngineTypes.endp = _endpos_s_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let s : (Javascript.statement * Javascript.location) = Obj.magic s in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_s_ in
        let _endpos = _endpos_s_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 177 "compiler/lib/js_parser.mly"
                     ( s )
# 12080 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = s;
          MenhirLib.EngineTypes.startp = _startpos_s_;
          MenhirLib.EngineTypes.endp = _endpos_s_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let s : (Javascript.statement * Javascript.location) = Obj.magic s in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_s_ in
        let _endpos = _endpos_s_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 177 "compiler/lib/js_parser.mly"
                     ( s )
# 12105 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = s;
          MenhirLib.EngineTypes.startp = _startpos_s_;
          MenhirLib.EngineTypes.endp = _endpos_s_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let s : (Javascript.statement * Javascript.location) = Obj.magic s in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_s_ in
        let _endpos = _endpos_s_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 177 "compiler/lib/js_parser.mly"
                     ( s )
# 12130 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _5;
          MenhirLib.EngineTypes.startp = _startpos__5_;
          MenhirLib.EngineTypes.endp = _endpos__5_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = y;
            MenhirLib.EngineTypes.startp = _startpos_y_;
            MenhirLib.EngineTypes.endp = _endpos_y_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = x;
              MenhirLib.EngineTypes.startp = _startpos_x_;
              MenhirLib.EngineTypes.endp = _endpos_x_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _;
                MenhirLib.EngineTypes.semv = _3;
                MenhirLib.EngineTypes.startp = _startpos__3_;
                MenhirLib.EngineTypes.endp = _endpos__3_;
                MenhirLib.EngineTypes.next = {
                  MenhirLib.EngineTypes.state = _;
                  MenhirLib.EngineTypes.semv = _3_inlined1;
                  MenhirLib.EngineTypes.startp = _startpos__3_inlined1_;
                  MenhirLib.EngineTypes.endp = _endpos__3_inlined1_;
                  MenhirLib.EngineTypes.next = {
                    MenhirLib.EngineTypes.state = _;
                    MenhirLib.EngineTypes.semv = item;
                    MenhirLib.EngineTypes.startp = _startpos_item_;
                    MenhirLib.EngineTypes.endp = _endpos_item_;
                    MenhirLib.EngineTypes.next = {
                      MenhirLib.EngineTypes.state = _;
                      MenhirLib.EngineTypes.semv = _1;
                      MenhirLib.EngineTypes.startp = _startpos__1_;
                      MenhirLib.EngineTypes.endp = _endpos__1_;
                      MenhirLib.EngineTypes.next = {
                        MenhirLib.EngineTypes.state = _menhir_s;
                        MenhirLib.EngineTypes.semv = pi;
                        MenhirLib.EngineTypes.startp = _startpos_pi_;
                        MenhirLib.EngineTypes.endp = _endpos_pi_;
                        MenhirLib.EngineTypes.next = _menhir_stack;
                      };
                    };
                  };
                };
              };
            };
          };
        } = _menhir_stack in
        let _5 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 12193 "compiler/lib/js_parser.ml"
        ) = Obj.magic _5 in
        let y : ((Javascript.statement_list * Javascript.case_clause list) option) = Obj.magic y in
        let x : (Javascript.case_clause list) = Obj.magic x in
        let _3 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 12200 "compiler/lib/js_parser.ml"
        ) = Obj.magic _3 in
        let _3_inlined1 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 12205 "compiler/lib/js_parser.ml"
        ) = Obj.magic _3_inlined1 in
        let item : (Javascript.expression) = Obj.magic item in
        let _1 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 12211 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 12216 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos__5_ in
        let _v =
          let pair = 
# 164 "<standard.mly>"
    ( (x, y) )
# 12225 "compiler/lib/js_parser.ml"
           in
          let subject = 
# 697 "compiler/lib/js_parser.mly"
                                                         ( item )
# 12230 "compiler/lib/js_parser.ml"
           in
          (
# 267 "compiler/lib/js_parser.mly"
   ( let switch = match pair with
       | cases, None ->
         Switch_statement (subject, cases, None, [])
       | cases, Some (default, more_cases) ->
         Switch_statement (subject, cases, Some default, more_cases)
      in switch, Pi pi )
# 12240 "compiler/lib/js_parser.ml"
           : (Javascript.statement * Javascript.location))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = alternative;
          MenhirLib.EngineTypes.startp = _startpos_alternative_;
          MenhirLib.EngineTypes.endp = _endpos_alternative_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _4;
            MenhirLib.EngineTypes.startp = _startpos__4_;
            MenhirLib.EngineTypes.endp = _endpos__4_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = consequence;
              MenhirLib.EngineTypes.startp = _startpos_consequence_;
              MenhirLib.EngineTypes.endp = _endpos_consequence_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _;
                MenhirLib.EngineTypes.semv = _2;
                MenhirLib.EngineTypes.startp = _startpos__2_;
                MenhirLib.EngineTypes.endp = _endpos__2_;
                MenhirLib.EngineTypes.next = {
                  MenhirLib.EngineTypes.state = _menhir_s;
                  MenhirLib.EngineTypes.semv = condition;
                  MenhirLib.EngineTypes.startp = _startpos_condition_;
                  MenhirLib.EngineTypes.endp = _endpos_condition_;
                  MenhirLib.EngineTypes.next = _menhir_stack;
                };
              };
            };
          };
        } = _menhir_stack in
        let alternative : (Javascript.expression) = Obj.magic alternative in
        let _4 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 12287 "compiler/lib/js_parser.ml"
        ) = Obj.magic _4 in
        let consequence : (Javascript.expression) = Obj.magic consequence in
        let _2 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 12293 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let condition : (Javascript.expression) = Obj.magic condition in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_condition_ in
        let _endpos = _endpos_alternative_ in
        let _v : (Javascript.expression) = 
# 337 "compiler/lib/js_parser.mly"
   ( ECond (condition, consequence, alternative) )
# 12302 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = alternative;
          MenhirLib.EngineTypes.startp = _startpos_alternative_;
          MenhirLib.EngineTypes.endp = _endpos_alternative_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _4;
            MenhirLib.EngineTypes.startp = _startpos__4_;
            MenhirLib.EngineTypes.endp = _endpos__4_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = consequence;
              MenhirLib.EngineTypes.startp = _startpos_consequence_;
              MenhirLib.EngineTypes.endp = _endpos_consequence_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _;
                MenhirLib.EngineTypes.semv = _2;
                MenhirLib.EngineTypes.startp = _startpos__2_;
                MenhirLib.EngineTypes.endp = _endpos__2_;
                MenhirLib.EngineTypes.next = {
                  MenhirLib.EngineTypes.state = _menhir_s;
                  MenhirLib.EngineTypes.semv = condition;
                  MenhirLib.EngineTypes.startp = _startpos_condition_;
                  MenhirLib.EngineTypes.endp = _endpos_condition_;
                  MenhirLib.EngineTypes.next = _menhir_stack;
                };
              };
            };
          };
        } = _menhir_stack in
        let alternative : (Javascript.expression) = Obj.magic alternative in
        let _4 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 12348 "compiler/lib/js_parser.ml"
        ) = Obj.magic _4 in
        let consequence : (Javascript.expression) = Obj.magic consequence in
        let _2 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 12354 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let condition : (Javascript.expression) = Obj.magic condition in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_condition_ in
        let _endpos = _endpos_alternative_ in
        let _v : (Javascript.expression) = 
# 337 "compiler/lib/js_parser.mly"
   ( ECond (condition, consequence, alternative) )
# 12363 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = alternative;
          MenhirLib.EngineTypes.startp = _startpos_alternative_;
          MenhirLib.EngineTypes.endp = _endpos_alternative_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _4;
            MenhirLib.EngineTypes.startp = _startpos__4_;
            MenhirLib.EngineTypes.endp = _endpos__4_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = consequence;
              MenhirLib.EngineTypes.startp = _startpos_consequence_;
              MenhirLib.EngineTypes.endp = _endpos_consequence_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _;
                MenhirLib.EngineTypes.semv = _2;
                MenhirLib.EngineTypes.startp = _startpos__2_;
                MenhirLib.EngineTypes.endp = _endpos__2_;
                MenhirLib.EngineTypes.next = {
                  MenhirLib.EngineTypes.state = _menhir_s;
                  MenhirLib.EngineTypes.semv = condition;
                  MenhirLib.EngineTypes.startp = _startpos_condition_;
                  MenhirLib.EngineTypes.endp = _endpos_condition_;
                  MenhirLib.EngineTypes.next = _menhir_stack;
                };
              };
            };
          };
        } = _menhir_stack in
        let alternative : (Javascript.expression) = Obj.magic alternative in
        let _4 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 12409 "compiler/lib/js_parser.ml"
        ) = Obj.magic _4 in
        let consequence : (Javascript.expression) = Obj.magic consequence in
        let _2 : (
# 79 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 12415 "compiler/lib/js_parser.ml"
        ) = Obj.magic _2 in
        let condition : (Javascript.expression) = Obj.magic condition in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_condition_ in
        let _endpos = _endpos_alternative_ in
        let _v : (Javascript.expression) = 
# 337 "compiler/lib/js_parser.mly"
   ( ECond (condition, consequence, alternative) )
# 12424 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _2;
          MenhirLib.EngineTypes.startp = _startpos__2_;
          MenhirLib.EngineTypes.endp = _endpos__2_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = pi;
            MenhirLib.EngineTypes.startp = _startpos_pi_;
            MenhirLib.EngineTypes.endp = _endpos_pi_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let _2 : (Javascript.expression) = Obj.magic _2 in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 12452 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos__2_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 275 "compiler/lib/js_parser.mly"
                         ( (Throw_statement _2, Pi pi) )
# 12460 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _4;
          MenhirLib.EngineTypes.startp = _startpos__4_;
          MenhirLib.EngineTypes.endp = _endpos__4_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _3;
            MenhirLib.EngineTypes.startp = _startpos__3_;
            MenhirLib.EngineTypes.endp = _endpos__3_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = _2;
              MenhirLib.EngineTypes.startp = _startpos__2_;
              MenhirLib.EngineTypes.endp = _endpos__2_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _menhir_s;
                MenhirLib.EngineTypes.semv = pi;
                MenhirLib.EngineTypes.startp = _startpos_pi_;
                MenhirLib.EngineTypes.endp = _endpos_pi_;
                MenhirLib.EngineTypes.next = _menhir_stack;
              };
            };
          };
        } = _menhir_stack in
        let _4 : (Javascript.statement_list option) = Obj.magic _4 in
        let _3 : (Javascript.ident * Javascript.statement_list) = Obj.magic _3 in
        let _2 : (Javascript.statement_list) = Obj.magic _2 in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 12502 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos__4_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 278 "compiler/lib/js_parser.mly"
                                 ( (Try_statement (_2, Some _3, _4), Pi pi) )
# 12510 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _3;
          MenhirLib.EngineTypes.startp = _startpos__3_;
          MenhirLib.EngineTypes.endp = _endpos__3_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _2;
            MenhirLib.EngineTypes.startp = _startpos__2_;
            MenhirLib.EngineTypes.endp = _endpos__2_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _menhir_s;
              MenhirLib.EngineTypes.semv = pi;
              MenhirLib.EngineTypes.startp = _startpos_pi_;
              MenhirLib.EngineTypes.endp = _endpos_pi_;
              MenhirLib.EngineTypes.next = _menhir_stack;
            };
          };
        } = _menhir_stack in
        let _3 : (Javascript.statement_list) = Obj.magic _3 in
        let _2 : (Javascript.statement_list) = Obj.magic _2 in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 12545 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos__3_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 279 "compiler/lib/js_parser.mly"
                                ( (Try_statement (_2, None, Some _3), Pi pi) )
# 12553 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = i;
          MenhirLib.EngineTypes.startp = _startpos_i_;
          MenhirLib.EngineTypes.endp = _endpos_i_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let i : (Javascript.ident * Parse_info.t) = Obj.magic i in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_i_ in
        let _endpos = _endpos_i_ in
        let _v : (Javascript.ident) = 
# 598 "compiler/lib/js_parser.mly"
                       ( fst i )
# 12578 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = list;
          MenhirLib.EngineTypes.startp = _startpos_list_;
          MenhirLib.EngineTypes.endp = _endpos_list_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _menhir_s;
            MenhirLib.EngineTypes.semv = pi;
            MenhirLib.EngineTypes.startp = _startpos_pi_;
            MenhirLib.EngineTypes.endp = _endpos_pi_;
            MenhirLib.EngineTypes.next = _menhir_stack;
          };
        } = _menhir_stack in
        let list : (Javascript.variable_declaration list) = Obj.magic list in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 12606 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos_list_ in
        let _v : (Javascript.statement * Javascript.location) = 
# 202 "compiler/lib/js_parser.mly"
   ( Variable_statement list, Pi pi )
# 12614 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = i;
          MenhirLib.EngineTypes.startp = _startpos_i_;
          MenhirLib.EngineTypes.endp = _endpos_i_;
          MenhirLib.EngineTypes.next = _menhir_stack;
        } = _menhir_stack in
        let i : (
# 48 "compiler/lib/js_parser.mly"
      (string * Parse_info.t)
# 12635 "compiler/lib/js_parser.ml"
        ) = Obj.magic i in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_i_ in
        let _endpos = _endpos_i_ in
        let _v : (Javascript.ident * Parse_info.t) = 
# 601 "compiler/lib/js_parser.mly"
                  ( let name, pi = i in var pi name, pi )
# 12643 "compiler/lib/js_parser.ml"
         in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = body;
          MenhirLib.EngineTypes.startp = _startpos_body_;
          MenhirLib.EngineTypes.endp = _endpos_body_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _3;
            MenhirLib.EngineTypes.startp = _startpos__3_;
            MenhirLib.EngineTypes.endp = _endpos__3_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = item;
              MenhirLib.EngineTypes.startp = _startpos_item_;
              MenhirLib.EngineTypes.endp = _endpos_item_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _;
                MenhirLib.EngineTypes.semv = _1;
                MenhirLib.EngineTypes.startp = _startpos__1_;
                MenhirLib.EngineTypes.endp = _endpos__1_;
                MenhirLib.EngineTypes.next = {
                  MenhirLib.EngineTypes.state = _menhir_s;
                  MenhirLib.EngineTypes.semv = pi;
                  MenhirLib.EngineTypes.startp = _startpos_pi_;
                  MenhirLib.EngineTypes.endp = _endpos_pi_;
                  MenhirLib.EngineTypes.next = _menhir_stack;
                };
              };
            };
          };
        } = _menhir_stack in
        let body : (Javascript.statement * Javascript.location) = Obj.magic body in
        let _3 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 12689 "compiler/lib/js_parser.ml"
        ) = Obj.magic _3 in
        let item : (Javascript.expression) = Obj.magic item in
        let _1 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 12695 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let pi : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 12700 "compiler/lib/js_parser.ml"
        ) = Obj.magic pi in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos_pi_ in
        let _endpos = _endpos_body_ in
        let _v =
          let condition = 
# 697 "compiler/lib/js_parser.mly"
                                                         ( item )
# 12709 "compiler/lib/js_parser.ml"
           in
          (
# 228 "compiler/lib/js_parser.mly"
     ( (While_statement (condition, body), Pi pi) )
# 12714 "compiler/lib/js_parser.ml"
           : (Javascript.statement * Javascript.location))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
      (fun _menhir_env ->
        let _menhir_stack = _menhir_env.MenhirLib.EngineTypes.stack in
        let {
          MenhirLib.EngineTypes.state = _;
          MenhirLib.EngineTypes.semv = _3;
          MenhirLib.EngineTypes.startp = _startpos__3_;
          MenhirLib.EngineTypes.endp = _endpos__3_;
          MenhirLib.EngineTypes.next = {
            MenhirLib.EngineTypes.state = _;
            MenhirLib.EngineTypes.semv = _3_inlined1;
            MenhirLib.EngineTypes.startp = _startpos__3_inlined1_;
            MenhirLib.EngineTypes.endp = _endpos__3_inlined1_;
            MenhirLib.EngineTypes.next = {
              MenhirLib.EngineTypes.state = _;
              MenhirLib.EngineTypes.semv = item;
              MenhirLib.EngineTypes.startp = _startpos_item_;
              MenhirLib.EngineTypes.endp = _endpos_item_;
              MenhirLib.EngineTypes.next = {
                MenhirLib.EngineTypes.state = _;
                MenhirLib.EngineTypes.semv = _1_inlined1;
                MenhirLib.EngineTypes.startp = _startpos__1_inlined1_;
                MenhirLib.EngineTypes.endp = _endpos__1_inlined1_;
                MenhirLib.EngineTypes.next = {
                  MenhirLib.EngineTypes.state = _menhir_s;
                  MenhirLib.EngineTypes.semv = _1;
                  MenhirLib.EngineTypes.startp = _startpos__1_;
                  MenhirLib.EngineTypes.endp = _endpos__1_;
                  MenhirLib.EngineTypes.next = _menhir_stack;
                };
              };
            };
          };
        } = _menhir_stack in
        let _3 : (Javascript.statement * Javascript.location) = Obj.magic _3 in
        let _3_inlined1 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 12761 "compiler/lib/js_parser.ml"
        ) = Obj.magic _3_inlined1 in
        let item : (Javascript.expression) = Obj.magic item in
        let _1_inlined1 : (
# 64 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 12767 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1_inlined1 in
        let _1 : (
# 53 "compiler/lib/js_parser.mly"
       (Parse_info.t)
# 12772 "compiler/lib/js_parser.ml"
        ) = Obj.magic _1 in
        let _endpos__0_ = _menhir_stack.MenhirLib.EngineTypes.endp in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v =
          let _2 = 
# 697 "compiler/lib/js_parser.mly"
                                                         ( item )
# 12781 "compiler/lib/js_parser.ml"
           in
          (
# 262 "compiler/lib/js_parser.mly"
                                              ( assert false )
# 12786 "compiler/lib/js_parser.ml"
           : (Javascript.statement * Javascript.location))
        in
        {
          MenhirLib.EngineTypes.state = _menhir_s;
          MenhirLib.EngineTypes.semv = Obj.repr _v;
          MenhirLib.EngineTypes.startp = _startpos;
          MenhirLib.EngineTypes.endp = _endpos;
          MenhirLib.EngineTypes.next = _menhir_stack;
        });
    |]
  
  and trace =
    None
  
end

module MenhirInterpreter = struct
  
  module ET = MenhirLib.TableInterpreter.MakeEngineTable (Tables)
  
  module TI = MenhirLib.Engine.Make (ET)
  
  include TI
  
end

let standalone_expression =
  fun lexer lexbuf ->
    (Obj.magic (MenhirInterpreter.entry `Legacy 538 lexer lexbuf) : (Javascript.expression))

and program =
  fun lexer lexbuf ->
    (Obj.magic (MenhirInterpreter.entry `Legacy 0 lexer lexbuf) : (Javascript.program_with_annots))

module Incremental = struct
  
  let standalone_expression =
    fun initial_position ->
      (Obj.magic (MenhirInterpreter.start 538 initial_position) : (Javascript.expression) MenhirInterpreter.checkpoint)
  
  and program =
    fun initial_position ->
      (Obj.magic (MenhirInterpreter.start 0 initial_position) : (Javascript.program_with_annots) MenhirInterpreter.checkpoint)
  
end
